package Contig;	  # -*- perl -*-

use Util;
use HashDB;
use Carp;
use Messages;

use Exporter ();

@ISA    = (Exporter);
@EXPORT = qw(
	     dna_seq_ofC
	     dna_seq_of_regionC
	     downstream_regionC
	     length_contigC
	     which_contigsC
	     upstream_regionC
);

my ( %orgs_loaded, %contig );

sub dna_seq_of_regionC {
    my ($genome,$loc,$upstream,$downstream) = @_;

    my (@locs,@seqs,$seq,$contig,$beg,$end,$b1,$e1,$seq1,$seq2);
    my ($contig_seq,$contig_len,$failed);

    @locs = split(/\s*,\s*/,$loc);
    @seqs = ();

    for ($i=0; ($i <= $#locs); $i++)
    {
	$loc = $locs[$i];

	if ($loc =~ /^(\S+)_(\d+)_(\d+)$/)
	{
	    $contig     = $1; 
	    $beg        = $2; 
	    $end        = $3;

	    $contig_seq = &dna_seq_ofC($genome,$contig);
	    $contig_len = length($$contig_seq);

#	    warn "beg=$beg end=$end len=$contig_len";    

	    if ($beg <= $end)
	    {
		($b1,$e1) = &adjust_bounds_plus (
						 $beg,$end, ($i == 0),
						 $upstream, ($i == $#locs),
						 $downstream, $contig_len
						);

		if ( not defined $b1 or not defined $e1 ) 
		{
		    return undef;
		} 
		elsif ( $b1 > $e1 )
		{
		    return \"";
                }
                else 
                {
	            $seq = substr( $$contig_seq, $b1-1, $e1+1-$b1 );
                }
	    }
	    else
	    {
		($b1,$e1) = &adjust_bounds_minus (
                                                  $beg,$end, ($i == 0),
                                                  $upstream, ($i == $#locs),
                                                  $downstream, $contig_len
                                                 );

		if ( not defined $b1 or not defined $e1 ) 
		{
		    return undef;
		} 
		elsif ( $e1 > $b1 )
		{
		    return \"";
                }
		else 
		{
		    $seq = substr( $$contig_seq, $e1-1, $b1+1-$e1 );
		    $seq = ${ &rev_comp(\$seq) };
	        }
	    }
	    push(@seqs,$seq);
        }
	else
	{
	    print STDERR "Contig::dna_seq_of_regionC -- BAD ERROR IN location IN "
		, "$genome, $loc, $upstream, $downstream\n";
	}
    }
    
    if (! $failed)
    {
	$seq = join("",@seqs);
    }
    else
    {
	$seq = "";
    }
    return \$seq;
}

sub downstream_regionC
{
    # Niels Dec 15, 98. 
    
    my ( $orgid,      # Organism ID
	 $locstr,     # Example: CEAH6_16211_16860, ... 
	 $downstream  # Maximum length of output sequence
       ) = @_;

    my @locations = split ",", $locstr;

    my ( $contig_id, $contig_seq, $contig_len );
    my ( $beg, $end );
    my ( $offset, $length );

    my $down_seq;

    if ( $locations[-1] =~ /^(\S+)_(\d+)_(\d+)$/ and $downstream >= 0 ) 
    {
	$contig_id  = $1;

	$beg = $2;
	$end = $3;

	$contig_seq = ${ &dna_seq_ofC( $orgid, $contig_id ) };
	$contig_len = length $contig_seq;

	if ( $beg <= $end )
	{
	    $offset = $end;
	    $length = $downstream;

	    if ( $offset + $length > $contig_len )
	    {
		$length = $contig_len - $offset;
	    }

	    $down_seq = substr $contig_seq, $offset, $length;
	}
	else 
	{
	    $offset = $end - $downstream - 1;
	    $length = $downstream;

	    if ( $offset < 0 ) 
	    {
		$length = $length + $offset;
		$offset = 0;
	    }

	    $down_seq = substr $contig_seq, $offset, $length;
	    $down_seq = ${ &rev_comp( \$down_seq ) };
	}
    }
    else 
    {
	confess "downstream_regionC error - check arguments: \n" 
	       ."\$orgid    = $orgid\n"
	       ."\$locstr   = $locstr\n"
	       ."\$downstream = $downstream\n";
    }
    return $down_seq;
}

sub dna_seq_ofC 
{
    my ( $orgID, $contig_name ) = @_;
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# First time for a genome causes $organisms/OrgID/contigs to be loaded
# so that $contig{Rid} contains a pointer to the sequence of the contig.
# $contigID = "orgID,contig_name" (contig_name == first line of FASTA file)
#----------------------------------------------------------------------------
    my ( @lines, $contigID, $tmp, @tmp, $seq, $old_eor );
#============================================================================

    $contigID  = "$orgID\t$contig_name";

    unless ( $ref = $orgs_loaded )
    {
	$ref = ($orgs_loaded = &opendbH("contigs"));
    }

    if ( $tmp = $contig{ $contigID } )
    {
	return $tmp;
    }
    elsif (($tmp = &accessdbH($ref,"$contigID")) &&
	   open(TMP,"<$WIT2_Config::organisms/$orgID/contigs"))
    {
	($seek,$ln) = split(/\t/,$tmp);
	seek(TMP,$seek,0);
	$old_eor = $/;
	$/ = "\n";
	if ((defined($_ = <TMP>) && ( $_ =~ /^>(\S+)/ )) &&
	    ($1 eq $contig_name))
	{
	    @lines = ();
	    while (defined($_ = <TMP>) && ($_ !~ /^>/) )
	    {
		$_ =~ s/\s//g;
		push( @lines, $_ );
	    }
	    $seq = [];
	    push( @$seq,join( "", @lines ) );
	    $contig{ $contigID } = \( $seq->[0] );
        }
        close( TMP );
        $/ = $old_eor;
    }

    if ( $tmp = $contig{ $contigID } )
    {
#	warn "found $contigID: $tmp\n\n";
	return $tmp;
    } else {
#	warn "no contig $contigID\n\n";
	return \"";
    }
}

sub adjust_bounds_plus {
  my($beg,$end,$upC,$upstream,$downC,$downstream,$contig_len) = @_;
  my($b1,$e1);

  if ($upC && ($beg > $upstream) )
  {
      $b1 = $beg - $upstream;
  }
  elsif (! $upC)
  {
      $b1 = $beg;
  }
  else
  {
      $b1 = undef;
  }
  
  if ($downC && ($end <= ($contig_len + $downstream)))
  {
      $e1 = $end + $downstream;
  }
  elsif (! $downC)
  {
      $e1 = $end;
  }
  else
  {
      $e1 = undef;
  }
  return ($b1,$e1);
}

sub adjust_bounds_minus {
  my($beg,$end,$upC,$upstream,$downC,$downstream,$contig_len) = @_;
  my($b1,$e1);

  if ($downC && ($end > $downstream) )
  {
      $e1 = $end - $downstream;
  }
  elsif (! $downC)
  {
      $e1 = $end;
  }
  else
  {
      $e1 = $end;
  }
  
  if ($upC && ($beg <= ($contig_len + $upstream)))
  {
      $b1 = $beg + $upstream;
  }
  elsif (! $upC)
  {
      $b1 = $beg;
  }
  else
  {
      $b1 = $beg;
  }
  return ($b1,$e1);
}

sub length_contigC {
    my($org,$contig) = @_;
    my($seq,$contigID,$tmp,$ref,$ln);

    $contigID  = "$org\t$contig";
    if ( $seq = $contig{ $contigID } )
    {
	return length($$seq);
    }
    else
    {
	unless ( $ref = $orgs_loaded )
	{
	    $ref = ($orgs_loaded = &opendbH("contigs"));
	}
	if ($tmp = &accessdbH($ref,"$contigID")) 
	{
	    ($tmp,$ln) = split(/\t/,$tmp);
	    return $ln;
	}
	else
	{
	    return undef;
	}
    }
}

sub which_contigsC {
    my($genome) = @_;
    my(@contigs,$x);

    @contigs = `grep \"^>\" $WIT2_Config::organisms/$genome/contigs`;
    foreach $x (@contigs)
    {
	if ($x =~ /^>(\S+)/) { $x = $1; }
    }
    return \@contigs;
}

sub upstream_regionC
{
    # Niels, 15 Dec 98

    my ( $orgid,    # Organism ID
	 $locstr,   # Example: CEAH6_16211_16860, ... 
	 $upstream  # Maximum length of output sequence
       ) = @_;

    my @locations = split ",", $locstr;

    my ( $contig_id, $contig_seq, $contig_len );
    my ( $beg, $end );
    my ( $offset, $length );

    my $up_seq;

    if ( $locations[0] =~ /^(\S+)_(\d+)_(\d+)$/ and $upstream >= 0 ) 
    {
	$contig_id  = $1;

	$beg = $2;
	$end = $3;

	$contig_seq = $ { &dna_seq_ofC( $orgid, $contig_id ) };
	$contig_len = length $contig_seq;

	if ( $beg <= $end )
	{
	    $offset = $beg - $upstream - 1;
	    $length = $upstream;

	    if ( $offset < 0 ) 
	    {
		$length = $length + $offset;
		$offset = 0;
	    }

	    $up_seq = substr $contig_seq, $offset, $length;
	}
	else 
	{
	    $offset = $beg;
	    $length = $upstream;

	    if ( $offset + $length > $contig_len )
	    {
		$length = $contig_len - $offset;
	    }

	    $up_seq = substr $contig_seq, $offset, $length;
	    $up_seq = ${ &rev_comp( \$up_seq ) };
	}
    }
    else 
    {
	confess "upstream_regionC error - check arguments: \n" 
	       ."\$orgid    = $orgid\n"
	       ."\$locstr   = $locstr\n"
	       ."\$upstream = $upstream\n";
    }
    return $up_seq;
}

1;
#####  End of Contigs.pm  ##############################################

