package DA;			# -*- perl -*-

use Exporter ();
use Util;

@ISA    = ( Exporter );
@EXPORT = qw(
	     T_to_DA
	     DA_to_T
	     DA_pri
	     DA_nil
	     DA_one
	     DA_const
	     DA_var
	     DA_cop
	     DA_peek
	     DA_poke
	     DA_cadd
	     DA_cmul
	     DA_add
	     DA_sub
	     DA_mul
	     DA_div
	     DA_exp
	     DA_log
	     DA_fun
	     );

sub T_to_DA
{
    my ( $Tx ) = @_;
    my ( $k, $k_fact, $Dz );
    
    $Dz = &DA_cop( $Tx );
    
    $k_fact = 1;
    for ( $k = 1; $k <= $#{$Dz}; $k++, $k_fact *= $k )
    { 
	$Dz->[$k] /= $k_fact;
#	print STDERR "DA::T_to_DA --- $k,\t$k_fact,\t$Dz->[$k]\n";
    }
    
    return $Dz;
}

sub DA_to_T
{
    my ( $Dx ) = @_;
    my ( $k, $k_fact, $Tz );
    
    $Tz = &DA_cop( $Dx );
    
    $k_fact = 1;
    for ( $k = 1; $k <= $#{$Tz}; $k++, $k_fact *= $k )
    {
	$Tz->[$k] *= $k_fact;
#	print STDERR "DA::DA_to_T --- $k,\t$k_fact,\t$Tz->[$k]\n";
    }
    
    return $Tz;
}

sub DA_pri
{
    my ( $Dx ) = @_;
    my ( $Tz );
    
    $Tz = &DA_to_T( $Dx );
    
    return join( ", ", @$Tz );
}

sub DA_nil
{
    my ( $n ) = @_;
    
    return &DA_const( $n, 0.0 );
}

sub DA_one
{
    my ( $n ) = @_;
    
    return &DA_const( $n, 1.0 );
}

sub DA_const
{
    my ( $n, $value ) = @_;
    my ( $k, $Dz );
    
    return &DA_var( $n, 0, $value );
}

sub DA_var
{
    my ( $n, $m, $value ) = @_;
    my ( $k, $Dz );
    
    $Dz = [];
    for ( $k = 0; $k <= $n; $k++ ) { $Dz->[$k] = 0; }
    if (defined($value)) { $Dz->[$m] = $value; } else { $Dz->[$m] = 1; }
    
    return $Dz;
}

sub DA_cop
{
    my ( $Dx ) = @_;
    my ( $k, $Dz );
    
    $Dz = [];
    for ( $k = 0; $k <= $#{$Dx}; $k++ ) { $Dz->[$k] = $Dx->[$k]; }
    
    return $Dz;
}

sub DA_peek
{
    my ( $Dx, $k ) = @_;
    my ( $x_k );
    
    if ( $k <= $#{$Dx} ) { $x_k = $Dx->[$k]; } else { $x_k = undef; }

    return $x_k;
}

sub DA_poke
{
    my ( $Dx, $k, $value ) = @_;
    my ( $Dz );
    
    if ( $k <= $#{$Dx} ) 
    {
	$Dz = &DA_cop($Dx);
	if (defined($value)) { $Dz->[$k] = $value; } else { $Dz->[$k] = 0.0; }
    }
    
    return $Dz;
}

sub DA_cadd
{
    my ( $Dx, $y ) = @_;
    my ( $k, $Dz );
    
    if (defined($y))
    {
	$Dz = &DA_cop( $Dx );
	$Dz->[0] += $y;
    }
    else
    {
	$Dz = [ undef ];
    }
    
    return $Dz;
}

sub DA_cmul
{
    my ( $Dx, $y ) = @_;
    my ( $k, $Dz );
    
    if (defined($y))
    {
	$Dz = &DA_cop( $Dx );
#	print STDERR "DA::DA_cmul --- y = $y, Dx = ", join( ", ", @$Dx), "\n";
#	print STDERR "                        Dz = ", join( ", ", @$Dz), "\n";
	for ( $k = 0; $k <= $#{$Dx}; $k++ ) { $Dz->[$k] *= $y; }
#	print STDERR "                        Dz = ", join( ", ", @$Dz), "\n";
    }
    else
    {
	$Dz = [ undef ];
    }
    
    return $Dz;
}

sub DA_add
{
    my ( $Dx, $Dy ) = @_;
    my ( $n,  $k, $Dz );
    
    $n  = &min($#{$Dx}, $#{$Dy});
    $Dz = &DA_nil( $n );
    for ($k = 0; $k <= $n; $k++ )
    {
	$Dz->[$k] = $Dx->[$k] + $Dy->[$k];
    }
    
    return $Dz;
}

sub DA_sub
{
    my ( $Dx, $Dy ) = @_;
    my ( $n,  $k, $Dz );
    
    $n = &min($#{$Dx}, $#{$Dy});
    $Dz = &DA_nil( $n );
    for ($k = 0; $k <= $n; $k++ )
    {
	$Dz->[$k] = $Dx->[$k] - $Dy->[$k];
    }
    
    return $Dz;
}

sub DA_mul
{
    my ( $Dx, $Dy ) = @_;
    my ( $n,  $j, $k, $z_k, $Dz );
    
    $n = &min($#{$Dx}, $#{$Dy});
    $Dz = &DA_nil( $n );
    for ($k = 0; $k <= $n; $k++ )
    {
	$z_k = 0;
	for ($j = 0; $j <= $k; $j++ )
	{
	    $z_k += $Dx->[$j] * $Dy->[$k-$j];
	}
	$Dz->[$k] = $z_k;
    }
    
    return $Dz;
}

sub DA_div
{
    my ( $Dx, $Dy ) = @_;
    my ( $n,  $j, $k, $y_0, $d_0, $Du, $dv, $Dw, $Dz );
    
    $n = &min( $#{$Dx}, $#{$Dy} );
    
    $y_0 = $Dy->[0];
    $d_0 = 1.0/$y_0;
    $Du  = &DA_cmul( $Dx, $d_0 );
#   print STDERR "DA::DA_div --- Du = ", join( ", ", @$Du ), "\n";
    
    $d_0 = -$d_0;
    $dv  = &DA_cmul( $Dt = &DA_poke($Dy, 0, 0.0), $d_0 );
#   print STDERR "DA::DA_div --- Dt = ", join( ", ", @$Dt ), "\n";
#   print STDERR "DA::DA_div --- dv = ", join( ", ", @$dv ), "\n";

    $Dw  = &DA_cadd( $dv, 1.0 );
#   print STDERR "DA::DA_div --- Dw   = ", join( ", ", @$Dw ), "\n";
    for ( $k = 1; $k < $n; $k++ )
    {
	$Dw = &DA_cadd( &DA_mul( $Dw, $dv ), 1.0 );
#   print STDERR "DA::DA_div --- Dw $k = ", join( ", ", @$Dw ), "\n";
    }
    
    $Dz = &DA_mul( $Du, $Dw );
    
    return $Dz;
}

sub DA_exp
{
    my ( $Dx ) = @_;
    my ( $n, $x_0, $du, $z_0, $fact, $coeff, $Dz );
    
    $n = $#{$Dx};
    $x_0 = $Dx->[0];

    $z_0 = exp( $x_0 );

    $fact  = 1;
    $coeff = [0];
    for ($k=1; $k <= $n; $k++) 
    { 
	$fact *= $k;
	push( @$coeff, (1.0/$fact) );
    }
#   print STDERR "DA:DA_exp --- coeff = ", join( ", ", @$coeff ), "\n";
    
    $du = &DA_poke($Dx, 0, 0.0);
    $Dz = &DA_cadd( &DA_fun( $du, $coeff ), 1.0 );
    $Dz = &DA_cmul( $Dz, $z_0 );
    
    return $Dz;
}

sub DA_log
{
    my ( $Dx ) = @_;
    my ( $n, $x_0, $du, $z_0, $coeff, $Dz );
    
    $n = $#{$Dx};
    $x_0 = $Dx->[0];
    return [ undef ] if ($x_0 <= 0);

    $z_0 = log( $x_0 );

    $coeff = [0];
    for ($k=1; $k <= $n; $k++) { push( @$coeff, -(((-1.0)**$k)/$k) ); }
#   print STDERR "DA:DA_log --- coeff = ", join( ", ", @$coeff ), "\n";
    
    $du = &DA_cmul( &DA_poke($Dx, 0, 0.0), (1.0/$x_0) );
    $Dz = &DA_fun( $du, $coeff );
    $Dz = &DA_cadd( $Dz, $z_0 );
#    print STDERR "DA:DA_log --- Dz = ", join( ", ", @$Dz ), "\n";

    return $Dz;
}

sub DA_fun
{
    my ( $Dx, $coeff ) = @_;
    my ( $dx, $n, $k, $Dz );
    
    $n = &min( $#{$Dx}, $#{$coeff} );
#   print STDERR "DA:DA_fun --- n = $n\n";

    $dx = &DA_poke( $Dx, 0, 0.0 );
    $Dz = &DA_cmul( $dx, $coeff->[$n] );
    for ( $k = ($n-1); $k > 0; $k-- )
    {
	$Dz = &DA_mul( $dx, &DA_cadd($Dz, $coeff->[$k]) );
#	print STDERR "DA:DA_fun --- k = $k, Dz = ", join( ", ", @$Dz ), "\n";
    }
    
    return $Dz;
}

1;
#######  End package DA  ###############################################
