package FR;			# -*- perl -*-

use HashDB;
use Html;
use PW;
use Carp;
use Messages;

require Exporter;
@ISA = (Exporter);
@EXPORT = qw(
	        ec_name
	        functions_connect_toR
		function_to_pwsR
		function_to_seqsR
	        function_to_spsR
	        higher_level_functional_rolesR
	        kegg_enzyme_entryR
	        lower_level_functional_rolesR
	        role_nameR
);


sub ec_name {
    my($ec) = @_;

    my($x);
    if (! $ec_name_ref)
    {
	$ec_name_ref = &opendbH("ec_name");
    }
    
    if ($x = &accessdbH($ec_name_ref,$ec))
    {
	return $x;
    }
    return "";
}

sub function_to_pwsR {
    my($role,$org,$user) = @_;
    my($pws);

    $pws = [];
    &function_to_pws1($role,$org,$user,$pws);
    return $pws;
}

sub function_to_pws1 {
    my($role,$org,$user,$pws) = @_;
    my($x,@pws1);

    if (! $func_pw_ref)
    {
	$func_pw_ref = &opendbH("func_pw");
    }

    if ($x = &accessdbH($func_pw_ref,$role))
    {
	@pws1 = split(/\n/,$x);
    }
    else
    {
	@pws1 = ();
    }

    if ($user && (-s "$WIT2_Config::global/UserModels/$user/pw_func.table"))
    {
	@tmp = `grep \"\t$role\" $WIT2_Config::global/UserModels/$user/pw_func.table` ;
	foreach $_ (map { $_ =~ /^(\S+)/; $1; } @tmp)
	{
	    push(@pws1,$_);
	}
    }
    
    foreach $pw (@pws1)
    {
	if ((! $org) || (! defined($user)) || &asserted_pathwayP($org,$user,$pw))
	{
	    push(@$pws,$pw);
	}
    }
    $higher = &higher_level_functional_rolesR($role);
    foreach $role1 (@$higher)
    {
	&function_to_pws1($role1,$org,$user,$pws);
    }
}

sub kegg_enzyme_entryR {
    my($ec) = @_;
    my($ref,$fh,$old_eor,$kegg_entry,$seek);

    if (! $opened_kegg_enz)
    {
	open(KEGG,"<$WIT2_Config::global/kegg.enzymes") &&
	    ($ref = &opendbH("kegg_enz_index")) 
	    || wit_die "could not open kegg_enz_index";
	$opened_kegg_enz = [$ref,\*KEGG];
    }

    if (defined(($seek = &accessdbH($opened_kegg_enz->[0],$ec))))
    {
	$fh = $opened_kegg_enz->[1];
	seek($fh,$seek,0);
	
	$old_eor = $/;
	$/ = "\n///\n";
	$kegg_entry = <$fh>;
	$/ = $old_eor;
	return \$kegg_entry;
    }
    return \"";
}

sub function_to_spsR 
{
    my($role) = @_;
    my($ans,$hits,$x);
    
    $ans = [];
    $hits = {};
    &function_to_sps_recursive($role,$hits);
    foreach $x (keys(%{$hits}))
    {
#	print STDERR "FR::function_to_spsR: $x\t@{$hits->{$x}}\n";
	push( @$ans, [$x, $hits->{$x}] );
    }
    
    if (@$ans > 0)
    {
	return $ans;
    }
    else
    {
	return "";
    }
}

sub function_to_sps_recursive {
    my($role,$hits) = @_;
    my($lower, $lower_role);

#    print STDERR "recording hits for $role\n";
    &record_hits_sps($role,$hits);
    $lower    = &lower_level_functional_rolesR($role);
    foreach $lower_role (@$lower)
    {
	&record_hits_sps($lower_role,$hits);
#	&function_to_sps_recursive($lower_role,$hits);
    }
}

sub record_hits_sps {
    my($role,$hits) = @_;
    my(@hits,$x,$y);

    if (! $func_sps_ref)
    {
	$func_sps_ref = &opendbH("func_sp");
    }

    if ($x = &accessdbH($func_sps_ref,$role))
    {
#	print STDERR "got $x\n";
	@hits = split(/\t/,$x);
	foreach $x (@hits)
	{
	    if (! ($y = $hits->{$x}))
	    {
		$y = ($hits->{$x} = []);
	    }
	    push(@$y,$role);
	}
    }
}

#
# probably should implement a version such that org="" -> get seqs from
# all orgs
#
sub function_to_seqsR {
    my($role,$org,$user) = @_;
    my($ans,$hits,$x);

    if (defined($x = $function_to_seqs{"$role\t$org\t$user"})) { return $x; }

    $ans = [];
    $hits = {};
    &function_to_seqs_recursive($role,$org,$user,$hits);
    foreach $x (keys(%{$hits}))
    {
#	print STDERR "function_to_seqsR: $x @{$hits->{$x}}\n";
	push(@$ans,[$x,$hits->{$x}]);
    }

    if (@$ans > 0)
    {
	$function_to_seqs{"$role\t$org\t$user"} = $ans;
	return $ans;
    }
    else
    {
	$function_to_seqs{"$role\t$org\t$user"} = "";
	return "";
    }
}

sub function_to_seqs_recursive {
    my($role,$org,$user,$hits) = @_;
    my($lower, $lower_role);

    &record_hits($role,$org,$user,$hits);
    $lower    = &lower_level_functional_rolesR($role);

    foreach $lower_role (@$lower)
    {
#	&function_to_seqs_recursive($lower_role,$org,$user,$hits);
	&record_hits($lower_role,$org,$user,$hits);
    }
}

sub record_hits {
    my($role,$org,$user,$hits) = @_;
    my($all_hits,$x,$y);

    $all_hits = &basic_function_to_seqs($role,$org,$user);
#    print STDERR "recording hits for $role\n";
    foreach $x (@$all_hits)
    {
	if (! ($y = $hits->{$x}))
	{
	    $y = ($hits->{$x} = []);
	}
#	print STDERR "recording $x -> $role\n";
	push(@$y,$role);
    }
}

sub basic_function_to_seqs {
    my($role,$org,$user) = @_;
    my(@tmp,$x,@rids,$rids,$funcs);

    @tmp = ($role =~ /R[A-Z]{2}\d{5}/g);
    if (@tmp > 0) { return [@tmp]; }

    if ($user && 
	($x = &user_function_to_seqsR($role,$org,$user)) &&
	($#{$x} >= 0))
    {
	return $x;
    }
    
    if (! $opened_func_rids)
    {
	($opened_func_rids = &opendbH("function_to_rids"))
		|| wit_die "could not open function_to_rids";
    }

    if ($rids = &accessdbH($opened_func_rids,$role))
    {
#	print STDERR "rids=$rids\n";
	@rids = split(/\t/,$rids);
	@rids = grep(/^R$org/,@rids);
	$x = [];
	foreach $rid (@rids)
	{
	    $_ = &ORF::func_ofO($rid,$user);
	    if ($_ eq $role)
	    {
		push(@$x,$rid);
	    }
	}
	return $x;
    }
    else
    {
	return [];
    }
}

sub user_function_to_seqsR {
    my($role,$org,$user) = @_;

    if (! ($func_to_orfs = $loaded_assignments{"$org\t$user"}))
    {
	$func_to_orfs = {};
	$loaded_assignments{"$org\t$user"} = $func_to_orfs;

	if (!(-s "$WIT2_Config::organisms/$org/UserModels/$user/assigned_functions"))
	{
	    return [];
	}

	@assignments = `cat $WIT2_Config::organisms/$org/UserModels/$user/assigned_functions`;
	chop @assignments;
	foreach $assignment (@assignments)
	{
	    ($orf,$func) = split(/\t/,$assignment);
	    defined($func) || ($func = "");

	    if (! ($x = $func_to_orfs->{$func}))
	    {
		$x = ($func_to_orfs->{$func} = []);
	    }
	    push(@$x,$orf);
	}
    }

    if ($x = $func_to_orfs->{$role})
    {
	return $x;
    }
    else
    {
	return [];
    }
}

sub role_nameR {
    my($role,$set_link,$org,$user) = @_;
    my($ec,$x);

    if ($role =~ /^(\d+\.\d+\.\d+\.\d+)$/)
    {
	$ec = $1;
	if ($set_link && $org && $user)
	{
	    $role = ${&set_role_link($role,$org,$user)};
        }

	if (! $ec_name_ref)
	{
	    $ec_name_ref = &opendbH("ec_name");
	}

	if ($x = &accessdbH($ec_name_ref,$ec))
	{
	    return \"$role - $x";
	}
	else
	{
	    return \$role;
	}
    }
    else
    {
	if ($set_link && $org && $user)
	{
	    $role = ${&set_role_link($role,$org,$user)};
        }
	return \$role;
    }
}

sub lower_level_functional_rolesR {
    my($role) = @_;
    my(@roles,$x);

    if (! $lower_level_ref )
    {
	$lower_level_ref = &opendbH("func_to_lower");
    }

    if ($x = &accessdbH($lower_level_ref,$role))
    {
	@roles = split(/\t/,$x);
	return \@roles;
    }
    else
    {
	return [];
    }
}

sub higher_level_functional_rolesR {
    my($role) = @_;
    my(@roles,$x);

    if (! $higher_level_ref )
    {
	$higher_level_ref = &opendbH("func_to_higher");
    }

    if ($x = &accessdbH($higher_level_ref,$role))
    {
	@roles = split(/\t/,$x);
	return \@roles;
    }
    else
    {
	return [];
    }
}

sub functions_connect_toR {
    my($funcL,$org,$user) = @_;
    my($pws,$func,@pws);
    
    @pws = ();
    foreach $func (@$funcL)
    {
	$pws = &function_to_pwsR($func,$org,$user);
	push(@pws,@$pws);
    }
    return \@pws;
}

1;

