package Feature;		# -*- perl -*-

use Contig;
use Util;
use HashDB;
use Carp;
use Messages;

require Exporter;
@ISA = (Exporter);
@EXPORT = qw(
	     aliasesF
	     all_featuresF
	     dna_seq_ofF
	     downstreamF
	     genome_ofF
	     get_cannonicalF
	     get_locationF
	     get_parsed_locationF
	     features_in_regionF
	     known_featureF
	     locationF
	     set_cannonicalF
	     set_locationF
	     typeF
	     upstreamF
);

# 
# The first time that you access any of these for a genome,
# cache the $organisms/Org/Features/<feature>/tbl entries.
#     Keep $cached{"$genome,$type"} set to know what genome/features
#      have been cached.
# 
$organisms = $main::organisms;

sub all_featuresF {
    my($org,$type) = @_;
    my(@ids);

    ($org && $type) || confess "org=$org type=$type";
    if (-s "$WIT2_Config::organisms/$org/Features/$type/tbl")
    {
	@ids = `cut -f1 $WIT2_Config::organisms/$org/Features/$type/tbl`;
	chomp @ids;
	return \@ids;
    }
    else
    {
	return [];
    }
}

sub locationF 
{
    my($id) = @_;
    my($x,$loc,$genome);

    if ($id =~ /^R([A-Z]{2})\d{5}$/)
    {
	$genome = $1;
	&load_feature_tbl($1,"orf");
    }
    elsif ($id =~ /^([A-Z]{2})INTE\d{3}$/)
    {
	$genome = $1;
	&load_feature_tbl($1,"intein");
    }
    elsif ($id =~ /^([A-Z]{2})INTR\d{3}$/)
    {
	$genome = $1;
	&load_feature_tbl($1,"intron");
    }
    elsif ($id =~ /^([A-Z]{2})RNA\d{3}$/)
    {
	$genome = $1;
	&load_feature_tbl($1,"rna");
    }
    elsif ($id =~ /^([A-Z]{2})REP\S+$/)
    {
	$genome = $1;
	&load_feature_tbl($1,"repeat");
    }
    
    if ($x = &accessdbH($loaded_feature_tbls,$id))
    {
	($loc) = split(/\t/,$x);
	if ($loc)
	{
	    return $loc;
	}
    }
    return "";
}

sub aliasesF 
{
    my($id) = @_;
    my($x,$loc,@rest);

    &load_needed_feature_tbls($id);
    if ($x = &accessdbH($loaded_feature_tbls,$id))
    {
	($loc,@rest) = split(/\t/,$x);
	if (@rest > 0)
	{
	    return \@rest;
	}
	else
	{
	    return [];
	}
    }
    return "";
}

sub dna_seq_ofF {
    my($fid,$upstream,$downstream) = @_;
    my($loc,$genome);

    $loc     = &get_locationF($fid);
    $genome  = &genome_ofF($fid);

    return &dna_seq_of_regionC($genome,$loc,$upstream,$downstream);
}

sub upstreamF {
    my($id,$size) = @_;
# &upstreamF(RMJ00001,100) returns a 100-character string
# of characters that directly preced the orf.
#

}

sub load_needed_feature_tbls {
    my($id) = @_;

    defined($id) || confess "bad ID";

    if ($id =~ /^R([A-Z]{2})\d{5}$/)
    {
	&load_feature_tbl($1,"orf");
    }
    elsif ($id =~ /^([A-Z]{2})INTE\d{3}$/)
    {
	&load_feature_tbl($1,"intein");
    }
    elsif ($id =~ /^([A-Z]{2})INTR\d{3}$/)
    {
	&load_feature_tbl($1,"intron");
    }
    elsif ($id =~ /^([A-Z]{2})RNA\d{3}$/)
    {
	&load_feature_tbl($1,"rna");
    }
    elsif ($id =~ /^([A-Z]{2})REP\S+$/)
    {
	&load_feature_tbl($1,"repeat");
    }
}    

sub known_featureF {
    my($id) = @_;
    my($x);

    &load_needed_feature_tbls($id);
    if ($x = &accessdbH($loaded_feature_tbls,$id))
    {
	return 1;
    }
    else
    {
	return 0;
    }
}

sub load_feature_tbl {
    my($genome,$feature) = @_;
    my(@tbl,$x,$old_eor);

    if (! $loaded_feature_tbls)
    {
	($loaded_feature_tbls = &opendbH("tbls")) || wit_die "could not open tbls.db";
    }
}



sub genome_ofF {
    my($id) = @_;

    if ( $id =~ /^R([A-Z]+)\d{5}$/    )   { return $1; }
    if ( $id =~ /^([A-Z]+)RNA\d{3}$/  )   { return $1; }
    if ( $id =~ /^([A-Z]+)INTR\d{3}$/ )   { return $1; }
    if ( $id =~ /^([A-Z]+)INTE\d{3}$/ )   { return $1; }
    if ( $id =~ /^([A-Z]+)REP\S+$/    )   { return $1; }
    
    return "unknown";
}



sub features_in_regionF {
    my($org,$contig,$beg,$end) = @_;
    my(@tbl,@features,$line,$fid,$loc,@locs,$i,$keep,$min,$max,$b1,$e1);
    
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# This routine is a bit weird in that it returns not only the
# features in the specified region, but also the minimum and maximum 
# coordinates of features on the contig.
#-----------------------------------------------------------------------
    
    if ($beg > $end)  { ($beg, $end) = ($end, $beg); }
    
    @tbl       = `cat $WIT2_Config::organisms/$org/Features/*/tbl`;
    @features  = (); $min = 100000000; $max = 0;
    foreach $line (@tbl)
    {
	($fid,$loc) = split(/\t/,$line);
#	print STDERR "Feature::features_in_regionF -- fid=$fid loc=$loc line=$line\n";
	@locs       = split(/\s*,\s*/,$loc);
	for ($i=0, $keep=0; ($i <= $#locs); $i++)
	{
	    if (&loc_overlaps($contig,$beg,$end,$loc))
	    {
		$keep = 1;
	    }

	    if ($loc =~ /^(\S+)_(\d+)_(\d+)$/)
	    {
#		print STDERR "Feature::features_in_regionF -- $loc, $1, $2, $3\n";
		if ($1 eq $contig)
		{
		    $b1 = $2; $e1 = $3;
		    $min = &min($min,&min($b1,$e1));
		    $max = &max($max,&max($b1,$e1));
		}
	    }
	}
	
	if ($keep)
	{
#	    print STDERR "Feature::features_in_regionF -- pushing $fid\n";
	    push(@features,$fid);
	}
    }
    
    if ($min == 100000000) { $min = ""; }
    if ($max == 0)         { $max = ""; }
    
#    print STDERR "Feature::features_in_regionF -- returning min=$min, max=$max\n"
#	, "\tfeatures = ", join( ", ", @features), "\n";

    return (\@features,$min,$max);
}

sub loc_overlaps {
    my($contig,$beg,$end,$loc) = @_;
    
    foreach $_ (split(/,/, $loc))
    {
	if (($_ =~ /^(\S+)_(\d+)_(\d+)$/) && 
	    ($contig eq $1) && 
	    &overlaps($beg,$end,$2,$3))
	{
	    return 1;
	}
    }
    return 0;
}
	
sub typeF {
    my($id) = @_;

    if ($id =~ /^R([A-Z]{2})\d{5}$/)
    {
	return("orf");
    }
    elsif ($id =~ /^([A-Z]{2})INTE\d{3}$/)
    {
	return("intein");
    }
    elsif ($id =~ /^([A-Z]{2})INTR\d{3}$/)
    {
	return("intron");
    }
    elsif ($id =~ /^([A-Z]{2})RNA\d{3}$/)
    {
	return("rna");
    }
    elsif ($id =~ /^([A-Z]{2})REP\S+$/)
    {
	return("repeat");
    }
    return "unknown";
}



sub get_cannonicalF {
    my($fid) = @_;

    return &get_field($fid,1);
}

sub get_locationF {
    my($fid) = @_;

    return &get_field($fid,0);
}

sub get_parsed_locationF {
    my ( $fid ) = @_;
    my ( $loc );
    my ( $contig, $beg, $end );
    
    $loc = &get_field($fid,0);
    if ($loc =~ /^(\S+)_(\d+)_(\d+)$/)
    {
	$contig = $1; $beg = $2; $end = $3;
    } else {
	$contig = ""; $beg = ""; $end = "";
    }
    
    return ($contig, $beg, $end);
}

sub get_field {
    my($fid,$field) = @_;
    my($x);

    &load_needed_feature_tbls($fid);
    if ($x = &accessdbH($loaded_feature_tbls,$fid))
    {
	@f = split(/\t/,$x);
	if (@f > $field)
	{
	    return $f[$field];
	}
    }
    return "";
}



sub set_cannonicalF {
    my($fid,$name) = @_;

    return &set_field($fid,1,$name);
}

sub set_locationF {
    my($fid,$loc) = @_;

    return &set_field($fid,0,$loc);
}

sub set_field {
    my($fid,$field,$value) = @_;
    my($org,$type,$i,@fields,$rest);

    $org = &genome_of($fid);
    if (($type = &typeF($fid)) && (-s "$WIT2_Config::organisms/$org/Features/$type/tbl"))
    {
	@tbl = `cat $WIT2_Config::organisms/$org/Features/$type/tbl`;
	for ($i=0; ($i <= $#tbl) && ($tbl[$i] !~ /^$fid/); $i++) {}
	if ($i <= $#tbl)
	{
	    chop $tbl[$i];
	    @fields = split(/\t/,$tbl[$i]);
	    shift @fields;
	    $fields[$field] = $value;
	    $rest = join("\t",@fields);
	    $tbl[$i] = "$fid\t$rest" . "\n";
	    &write_file("$WIT2_Config::organisms/$org/Features/$type/tbl",\@tbl);
	    &load_feature_tbl($org,$type);
	    &replacedbH($loaded_feature_tbls,$fid,$rest);
	    &sync_dbH($loaded_feature_tbls);
	    return 1;
	}
	else
	{
	    return 0;
	}
    }
    else
    {
	return 0;
    }
}

1;
#####  End of Feature.pm  ##############################################

