package FuncCoupling;		# -*- perl -*-

require Exporter;
@ISA = (Exporter);
@EXPORT = qw(
	functionally_coupled
	functionally_coupledH
	functional_cluster
	functional_clusterH
	fc_matrix_to_html
	fill_matrix
	show_matrix
);

use Taxonomy;
use HashDB;
use Carp;
use ORF;
use Html;
use Util;
use Sim;
use Messages;

# functionally_coupled($rid) -> [[Rid2,Sc,Rid3,Rid4],...]
sub functionally_coupled {
    my($rid) = @_;
    
    return &functionally_coupled_guts($rid,"pcbbhs");
}

sub functionally_coupledH {
    my($rid) = @_;
    my($bi,$tri);
    my(@tmp);

    if ($bi = &functionally_coupled_guts($rid,"pchs"))
    {
	if ($tri = &functionally_coupledT($rid))
	{
	    foreach $_ (@$tri)
	    {
		push(@$bi,$_);
	    }
	}
	@tmp = sort { ($a->[0] cmp $b->[0]) or 
		      ($b->[1] <=> $a->[1]) or
		      ($a->[2] cmp $b->[2]) or
			  ($a->[3] cmp $b->[3]) } @$bi;
	return \@tmp;
    }
    return &functionally_coupledT($rid);
}
    

sub functionally_coupled_guts {
    my($rid,$table) = @_;
    my($hits,$rid2,$rid3,$rid4,$sc);

#    print STDERR "functionally_coupled: rid=$rid\n";
    $hits = [];
    if (! $opened_fc_ref{$table})
    {
	$opened_fc_ref{$table} = &opendbH($table);
    }

    if ($_ = &accessdbH($opened_fc_ref{$table},$rid))
    {
	foreach $_ (split(/\t/,$_))
	{
	    ($rid2,$rid3,$rid4,$sc) = split(/,/,$_);
#	    print STDERR "$rid/$rid2 -> $rid3/$rid4 $sc\n";
	    push(@$hits,[$rid2,$sc,$rid3,$rid4]);
	}
    }

    if (@$hits > 0)
    {
	return $hits;
    }
    else
    {
	return "";
    }
}

# functional_cluster($rid,$cutoff) -> [[setr1sorg1,setr1sorg2,...setrnsorgN],[setr2sorg1,setr2sorg2...],...]
# functional_cluster($rid,$cutoff) -> [[[r1sorg1,r2sorg1,...rnsorg1],[r1sorg2,r2sorg2...]],...]
sub functional_cluster {
    my($rid,$cutoff) = @_;

    return &functional_cluster_guts($rid,$cutoff,\&functionally_coupled);
}

sub functional_clusterH {
    my($rid,$cutoff) = @_;

    return &functional_cluster_guts($rid,$cutoff,\&functionally_coupledH);
}

sub functional_cluster_guts {
    my($rid,$cutoff,$functionally_coupled) = @_;
    my($same,$seen,@to_expand,$i,$j,$rid1,$rid2,$rid3,$rid4,$sc,$genomes);
    my($x,$y,$z,%coupled,%by,@coupled);

    if (! defined($cutoff)) { $cutoff = 1.0e-5; }

    $same = {};
    $seen = {};
    @to_expand = ();
    &seen($seen,\@to_expand,$rid);

    for ($i=0; ($i < @to_expand) && ($i < 200); $i++)
    {
#	print STDERR "processing to_expand $i = $to_expand[$i]\n";
	$rid1 = $to_expand[$i];
	if ($x = &{$functionally_coupled}($rid1))
	{
	    undef %coupled; undef %by;
	    foreach $y (@$x)
	    {
		$coupled{$y->[0]} += $y->[1];
#		print STDERR "$y->[0] $y->[1]\n";
		if (! ($z = $by{$y->[0]}))
		{
		    $z = ($by{$y->[0]} = []);
		}
		if (@$y == 4)
		{
		    push(@$z,[$y->[2],$y->[3]]);
		}
		else
		{
#		    push(@$z,[$y->[2],$y->[4]]);
		}
	    }
	
	    @coupled = sort { $coupled{$b} <=> $coupled{$a} } keys(%coupled);
#	    print STDERR "$coupled[0] $coupled{$coupled[0]}\n";
	    for ($j=0; ($j <= $#coupled) && ($coupled{$coupled[$j]} >= $cutoff); $j++) {}
	    if ($j <= $#coupled)
	    {
		splice(@coupled,$j);
	    }

	    foreach $rid2 (@coupled)
	    {
		&seen($seen,\@to_expand,$rid2);
		$sc = $coupled{$rid2};
		$z  = $by{$rid2};
		foreach $pair (@$z)
		{
		    ($rid3,$rid4) = @$pair;
		    &seen($seen,\@to_expand,$rid3);
		    &seen($seen,\@to_expand,$rid4);
		    &add_conn($same,$rid1,$rid3);
		    &add_conn($same,$rid2,$rid4);
#		    print STDERR "coupled $coupled{$rid2} $rid1 $rid3 $rid2 $rid4\n";
		}
	    }
	}
    }

    $genomes = &genome_list($seen);
#    print STDERR "genomes=@$genomes\n";
    $matrix = &cluster_rids($genomes,$same);
    return &sort_rows($matrix);
}

sub sort_rows {
    my($matrix) = @_;
    my(%loc,$i,$rid,$cell,@m);

    for ($i=0; ($i < @$matrix); $i++)
    {
	foreach $cell (@{$matrix->[$i]})
	{
	    foreach $rid (@$cell)
	    {
		$loc = &get_locationO($rid);
		if ((! $loc{$i}) || ($loc{$i} gt $loc)) { $loc{$i} = $loc; }
	    }
	}
    }
    
    @m = ();
    for ($i=0; ($i < @$matrix); $i++)
    {
	push(@m,[$loc{$i},$matrix->[$i]]);
    }
    @m = sort { $a->[0] cmp $b->[0] } @m;
    @m = map { $_->[1] } @m;
    return \@m;
}	    

sub cluster_rids {
    my($genomes,$links) = @_;
    my(@clusters,$x,%seen,@cluster,$i,$y,$z);

    @clusters = ();
    foreach $x (keys(%$links))
    {
	if (! $seen{$x})
	{
	    $seen{$x} = 1;
	    @cluster = ($x);
	    for ($i=0; ($i < @cluster); $i++)
	    {
		$y = $links->{$cluster[$i]};
		foreach $z (@$y)
		{
		    if (! $seen{$z})
		    {
			push(@cluster,$z);
			$seen{$z} = 1;
		    }
		}
	    }
	    @cluster = sort @cluster;
	    push(@clusters,&format_cluster($genomes,\@cluster));
	}
    }
    return \@clusters;
}

sub format_cluster {
    my($genomes,$cluster) = @_;
    my(@formatted,$org,$rids);

    @formatted = ();
    foreach $org (@$genomes)
    {
	$rids = [];
	foreach $_ (@$cluster)
	{
	    if (substr($_,1,2) eq $org)
	    {
		push(@$rids,$_);
	    }
	}
	push(@formatted,$rids);
    }
    return \@formatted;
}

sub genome_list {
    my($seen) = @_;
    my(%genomes,@x);

    foreach $_ (keys(%$seen))
    {
	$genomes{substr($_,1,2)} = 1;
    }

    @x = ();
    foreach $_ (&genomes)
    {
	if ($genomes{$_})
	{
	    push(@x,$_);
	}
    }
    return \@x;
}

sub seen {
    my($seen,$to_expand,$rid) = @_;

    if (! $seen->{$rid})
    {
#	print STDERR "first seen: $rid\n";
	defined($rid) || confess "rid=$rid";
	$seen->{$rid} = 1;
	push(@$to_expand,$rid);
    }
}

sub add_conn {
    my($table,$rid1,$rid2) = @_;

    &add_link($table,$rid1,$rid2);
    &add_link($table,$rid2,$rid1);
}

sub add_link {
    my($table,$key,$data) = @_;
    my($x,$i);

    if (! ($x = $table->{$key}))
    {
	$x = ($table->{$key} = []);
    }

    for ($i=0; ($i <= $#{$x}) && ($x->[$i] gt $data); $i++) {}
    if (($i > $#{$x}) || ($x->[$i] lt $data))
    {
	splice(@$x,$i,0,$data);
    }
}

# fc_matrix_to_html($matrix) -> pointer to HTML string
sub fc_matrix_to_html {
    my($matrix,$user,$fill,$mapids,$rows,$hdr) = @_;
    my($color_of,$col_hdrs,$i,$j,$row_hdrs,$table,$row,%funcs,$cell);
    my($rid,$func,%rid_func,@tmp,$n,$entry,$link);

    $color_of = {};
    &set_colors($matrix,$color_of);

    $col_hdrs = [];
    for ($i=0; ($i <= $#{$matrix->[0]}); $i++)
    {
	for ($j=0; ($#{$matrix->[$j]->[$i]} < 0); $j++) {}
	push(@$col_hdrs,substr($matrix->[$j]->[$i]->[0],1,2));
    }

    $row_hdrs = [];
    $table    = [];
    foreach $row (@$matrix)
    {
	undef %funcs; undef %rid_func;
	foreach $cell (@$row)
	{
	    foreach $rid (@$cell)
	    {
		if (!($func = &func_ofO($rid,$user)))
		{
		    $func = "unknown";
		}

		$funcs{$func}++;
		$rid_func{$rid} = $func;
	    }
	}

	@funcs = sort { $funcs{$b} <=> $funcs{$a} } keys(%funcs);
	if (@funcs > 1)
	{
	    @tmp = ();
	    for ($i=0; ($i < @funcs); $i++)
	    {
		$n = $i+1;
		$tmp[$i] = "$n. " . $funcs[$i];
		$funcs{$funcs[$i]} = "$n. ";
	    }
	}
	else
	{
	    @tmp = @funcs;
	    $funcs{$funcs[0]} = "";
	}

	push(@$row_hdrs,join("<br>",@tmp));

	if ($fill) { &fill_row($row,$col_hdrs); }

	$entry = [];
	foreach $cell (@$row)
	{
	    @tmp = ();
	    $color = "#CCCCCC";
	    foreach $rid (@$cell)
	    {
		if ($_ = $color_of->{$rid}) { $color = $_; }
		$f = $rid_func{$rid};
		if ((! ($n = $funcs{$f})) || $rows)
		{
		    $n = "";
		}

		if ($mapids)
		{
		    $link = $mapids->{$rid};
		    if ($link eq "")
		    {
			$link = "<font color=\"$color\">-</FONT>";
		    }
		}
		else
		{
		    $link = &region_link($rid,$user);
		}
		push(@tmp,"$n$link");
	    }
	    if ($color ne "black")
	    {
		push(@$entry,join("<br>",@tmp) . "\tbgcolor=$color");
	    }
	    else
	    {
		push(@$entry,join("<br>",@tmp));
	    }
	}
	push(@$table,$entry);
    }
    
    if ($rows)   { $row_hdrs = $rows; }
    if (! $hdr)  { $hdr = "Functional Couplings"; }
    return &matrix_to_html($row_hdrs,$col_hdrs,$table,1,$hdr,"left","nowrap");
}

sub fill_row {
    my($row,$orgs) = @_;
    my($i,%orthos,$cell,@orthos,$rid);

    for ($i=0; ($i <= $#{$row}); $i++)
    {
	if ($#{$row->[$i]} < 0)
	{
	    undef %orthos;
	    foreach $cell (@$row)
	    {
		if (@$cell == 1)
		{
		    @orthos = &ortholog_ridS($cell->[0],1.0e-7);
		    foreach $rid (@orthos)
		    {
			if (substr($rid,1,2) eq $orgs->[$i]) 
			{
			    $orthos{$rid} = 1;
			}
		    }
		}
	    }
	    @orthos = sort keys(%orthos);
	    if (@orthos > 0)
	    {
		$row->[$i] = [@orthos];
	    }
	}
    }
}


sub region_link {
    my($rid,$user) = @_;
    my($org,$loc,$contig,$beg,$end,$beg1,$end1);

    if (! defined($user)) { $user = ""; }

    $loc = &get_locationO($rid);
    $org = &genome_ofO($rid);
	
    if ($loc =~ /^([^,]+)_(\d+)_(\d+)$/)
    {
	$contig = $1;
	$beg    = $2;
	$end    = $3;
	$beg1   = &min($beg,$end) - 10000;
	$end1   = &max($beg,$end) + 10000;
	return "<a href=contig.cgi?org=$org&user=$user&contig=$contig&beg=$beg1&end=$end1>$rid</a>";
    }
    return $rid;
}


sub set_colors {
    my($matrix,$color_of) = @_;
    my($col,@rids,$i,$j,$links,@colors,$clusters,$cluster,$rid);
    my(%used,%row_of,$key,$color,%rows);

    @colors = ("#50FFFF", "#FFFF00", "#FF9FE5", "#7E9FFFC", "#FF9900", "MediumTurquoise", "tan",
               "YellowGreen", "LightSeaGreen", "PaleVioletRed", "yellow", "orange", "beige", 
	       "pink", "LightCoral","gold","RosyBrown","green", "VioletRed", "lavender", 
	       "cornsilk", "LightCoral", "MediumSpringGreen", "LightGoldenrodYellow", 
	       "PaleTurquoise","red", "peru", "CornflowerBlue", "salmon","orchid", 
	       "PeachPuff", "GreenYellow", "tomato",  "PaleGreen", "MediumTurquoise",
	       "violet", "AliceBlue", "LightYellow", "magenta", "red", "SeaGreen", 
	       "MediumOrchid", "seashell", "thistle", "LimeGreen", "khaki", "chartreuse",
	       "OliveDrab", "SandyBrown", "MediumPurple", "ForestGreen", 
	       "chocolate", "DarkOrange", "LightYellow",  "LightSteelBlue",
	       "firebrick", "plum",
	       "#50FFFF", "#FFFF00", "#FF9FE5", "#7E9FFFC", "#FF9900", "MediumTurquoise", "tan",
               "YellowGreen", "LightSeaGreen", "PaleVioletRed", "yellow", "orange", "beige", 
	       "pink", "LightCoral","gold","RosyBrown","green", "VioletRed", "lavender", 
	       "cornsilk", "LightCoral", "MediumSpringGreen", "LightGoldenrodYellow", 
	       "PaleTurquoise","red", "peru", "CornflowerBlue", "salmon","orchid", 
	       "PeachPuff", "GreenYellow", "tomato",  "PaleGreen", "MediumTurquoise",
	       "violet", "AliceBlue", "LightYellow", "magenta", "red", "SeaGreen", 
	       "MediumOrchid", "seashell", "thistle", "LimeGreen", "khaki", "chartreuse",
	       "OliveDrab", "SandyBrown", "MediumPurple", "ForestGreen", 
	       "chocolate", "DarkOrange", "LightYellow",  "LightSteelBlue",
	       "firebrick", "plum"
	       );

    for ($col=0; ($col <= $#{$matrix->[0]}); $col++)
    {
	@rids = ();
	for ($j=0; ($j <= $#{$matrix}); $j++)
	{
	    foreach $rid (@{$matrix->[$j]->[$col]})
	    {
		$row_of{$rid} = $j;
		push(@rids,$rid);
	    }
	}

	$links = {};
	for ($i=0; ($i < @rids); $i++)
	{
	    for ($j=$i+1; ($j < @rids); $j++)
	    {
		if (&close_rids($rids[$i],$rids[$j]))
		{
		    &add_conn($links,$rids[$i],$rids[$j]);
#		    print STDERR "linking $rids[$i] $rids[$j]\n";
		}
	    }
	}

	$clusters = &basic_cluster($links);
	for ($i=0; ($i <= $#{$clusters}); $i++)
	{
	    $cluster = $clusters->[$i];
	    undef %rows;
	    foreach $rid (@$cluster)
	    {
		$rows{$row_of{$rid}} = 1;
	    }
	    $key = join(",",sort(keys(%rows)));
	    if (!($color = $used{$key}))
	    {
		$color = $colors[keys(%used)];
		$used{$key} = $color;
	    }

	    foreach $rid (@$cluster)
	    {
#		print STDERR "color $rid $colors[$i]\n";
		$color_of->{$rid} = $color;
	    }
	}
    }
}

sub basic_cluster {
    my($links) = @_;
    my(@clusters,$x,%seen,@cluster,$i,$y,$z);

    @clusters = ();
    foreach $x (keys(%$links))
    {
	if (! $seen{$x})
	{
	    $seen{$x} = 1;
	    @cluster = ($x);
	    for ($i=0; ($i < @cluster); $i++)
	    {
		$y = $links->{$cluster[$i]};
		foreach $z (@$y)
		{
		    if (! $seen{$z})
		    {
			push(@cluster,$z);
			$seen{$z} = 1;
		    }
		}
	    }
	    push(@clusters,[@cluster]);
	}
    }
    return \@clusters;
}

sub close_rids {
    my($x,$y) = @_;
    my($loc1,$contig1,$loc2);

    if (($loc1 = &get_locationO($x)) && ($loc2 = &get_locationO($y)))
    {
	if ($loc1 =~ /^(\S+)_(\d+)_(\d+)$/)
	{
	    $contig1 = $1; $loc1 = int(($2 + $3)/2);
	    
	    if (($loc2 =~ /^(\S+)_(\d+)_(\d+)$/) && ($1 eq $contig1))
	    {
		if (abs($loc1 - (($2 + $3)/2)) < 10000)
		{
		    return 1;
		}
	    }
	}
    }
    return 0;
}

sub functionally_coupledT {
    my($rid) = @_;
    my($ref,$fh,$seek,$to,$sc,$rid1,$rid2,$rid3,$rid4,$rid5,$rid6);
    my($p1,$p2);

    if (! $opened_tri)
    {
	open(TMP,"<$WIT2_Config::global/triangulated_fc.table") &&
	    ($ref = &opendbH("tri_fc_index")) 
	    || wit_die "could not open tri_fc_index";
	$opened_tri = [$ref,\*TMP];
    }

    if (defined(($seek = &accessdbH($opened_tri->[0],$rid))))
    {
	$fh = $opened_tri->[1];
	seek($fh,$seek,0);
	
	$to = [];
	while (defined($_ = <$fh>) && ($_ =~ /^$rid/))
	{
	    chop;
	    ($rid1,$rid2,$sc,$p1,$p2) = split(/\t/,$_);
	    ($rid3,$rid4) = split(/,/,$p1);
	    ($rid5,$rid6) = split(/,/,$p2);
	    push(@$to,[$rid2,$sc,$rid3,$rid4,$rid5,$rid6]);
	}
	return $to;
    }
    else
    {
	return "";
    }
}

	
sub fill_matrix {
    my($matrix,$cutoff) = @_;
    my(@orgs,$i,$j,$org,$row,@rids,$cell,%orthos,$rid);

    @orgs = ();
    for ($i=0; ($i <= $#{$matrix->[0]}); $i++)
    {
	for ($j=0,$org=""; ($j < @$matrix) && (! $org); $j++)
	{
	    if ($matrix->[$j]->[$i]->[0])
	    {
		$org = substr($matrix->[$j]->[$i]->[0],1,2);
	    }
	}
	$org || confess "what happened?";
	push(@orgs,$org);
    }

    for ($i=0; ($i < @$matrix); $i++)
    {
	$row = $matrix->[$i];
	@rids = ();
	foreach $cell (@$row)
	{
	    foreach $_ (@$cell)
	    {
		push(@rids,$_);
	    }
	}
	    
	for ($j=0; ($j < @$row); $j++)
	{
	    $cell = $row->[$j];
	    if (@$cell == 0)
	    {
		undef %orthos;
		foreach $rid (@rids)
		{
		    @orthos = &ortholog_ridS($rid,$cutoff);
		    foreach $_ (@orthos)
		    {
			if (substr($_,1,2) eq $orgs[$j])
			{
			    $orthos{$_} = 1;
			}
		    }
		}
		$row->[$j] = [sort keys(%orthos)];
	    }
	}
    }
    return $matrix;
}

sub show_matrix {
    my($matrix) = @_;
    my($i,$j,$row,$org,$rids);

    print "displaying coupling matrix\n";
    for ($i=0; ($i < @$matrix); $i++)
    {
	$row = $matrix->[$i];
	for ($j=0; ($j < @$row); $j++)
	{
	    $org  = $row->[$j];
	    $rids = join(",",@$org);
	    print "[$i,$j] $rids\n";
	}
    }
    print "---------------\n\n";
}
	

1;
