package URLExpander;
use URI::URL;
use HTML::Parser;
@ISA = ("HTML::Parser");

sub new {
  my $type = shift;
  my $self = HTML::Parser->new;
  $self->{'BASE'} = shift;
  bless $self, $type;
}

sub base {
  my $self = shift;
  if (@_) { $self->{'BASE'} = shift }
  return $self->{'BASE'};
}

sub expand {
  my ($self, $buffer) = @_;
  $self->parse($buffer);
  return $self->{'OUTPUT'};
}

sub start {
  # $attr is reference to a HASH, $attrseq is reference to an ARRAY
  my($self, $tag, $attr, $attrseq, $origtext) = @_;

  if ( $attr->{'href'} && ( $attr->{'href'} !~ /^\#/ ) ) {
    $url = URI::URL->new( $attr->{'href'}, $self->{'BASE'} );
    $attr->{'href'} = $url->abs->as_string;
    $self->{'OUTPUT'} .= "<$tag";
    foreach ( keys %$attr ) {
#      print qq( $_="$attr->{$_}");
      $self->{'OUTPUT'} .= qq( $_="$attr->{$_}");
    }
    $self->{'OUTPUT'} .= ">";
  }
  elsif ( $attr->{'src'} ) {
    $url = URI::URL->new( $attr->{'src'}, $self->{'BASE'} );
    $attr->{'src'} = $url->abs->as_string;
    $self->{'OUTPUT'} .= "<$tag";
    foreach ( keys %$attr ) {
      $self->{'OUTPUT'} .= qq( $_="$attr->{$_}");
    }
    $self->{'OUTPUT'} .= ">";
  }
  elsif ( $attr->{'action'} ) {
    $url = URI::URL->new( $attr->{'action'}, $self->{'BASE'} );
    $attr->{'action'} = $url->abs->as_string;
    $self->{'OUTPUT'} .= "<$tag";
    foreach ( keys %$attr ) {
      $self->{'OUTPUT'} .= qq( $_="$attr->{$_}");
    }
    $self->{'OUTPUT'} .= ">";
  }
  else {
    $self->{'OUTPUT'} .= $origtext;
  }
}

sub text
{
  my($self, $text) = @_;
  $self->{'OUTPUT'} .= $text;
}

sub end
{
  my($self, $tag, $origtext) = @_;
  $self->{'OUTPUT'} .= $origtext;
}

1;

package Html;		# -*- perl -*-

use Carp;

use LWP::UserAgent;
use LWP::Simple;
use HTTP::Request::Common;
use HTML::Parser;
use HTML::Element;
use URI::URL;
use Util;

use WIT2_Config;
use Messages;

use Exporter ();
@ISA    = ( Exporter );
@EXPORT = qw(
	     add_attrib_col
	     add_attrib_row
	     add_table_col
	     add_table_row
	     aliases_with_links
	     blast_url
	     blast_type
	     data_dump
	     display_menu
	     entrez_url
	     entrez_type
	     enzyme_link
	     enz_forms_link
	     explore_possible_connections
	     fasta_sequence_of_gids
	     filter_special_chars
	     forward_request
	     get_a_form
	     get_hidden_values
	     get_html
	     get_medline_for_ec
	     get_prodom_analysis
	     gi_entries
	     html_encode
	     html_table_contig
	     html_table_default
	     html_table_prot
	     html_table
	     invoke_bcm
	     invoke_cog
	     invoke_emotif
	     invoke_workbench
	     invoke_pfscan
	     invoke_protscale
	     invoke_psort
	     invoke_tmpred
	     kegg_enzyme_link
	     kegg_map_link
	     matrix_to_html
	     ncbi_blast
	     prot_link
	     psi_blast
	     put_in_links
	     put_in_real_sp_links
	     pw_link
	     red_blue_ramp
	     results_headbar
	     rid_number 
	     save_prodom_analysis
	     select_menu
	     set_role_link
	     set_role_links
	     set_attrib_col
	     set_attrib_row
	     set_table_col
	     set_table_row
	     topnavbar
	     url_decode
	     url_encode
);


sub add_attrib_col {

# Dummy routine for now. 

    my ( $newpos, $tableP, $newcolP, $maskP ) = @_;

    &add_table_col( $newpos, $tableP, $newcolP, $maskP );
}

sub add_attrib_row {

# Dummy routine for now. 

    my ( $newpos, $tableP, $newcolP, $maskP ) = @_;

    &add_table_row( $newpos, $tableP, $newcolP, $maskP );
}


sub add_table_col {

# Inserts the values of @{ $newcolP } before position $newpos in 
# @{ $tableP }, so that the new column has position $newpos after
# the insertion. If in a given row the column does not exist, then
# it will. If $newcolP is a reference to an array, that array will 
# be inserted as is. If it is anything else, then this will be
# inserted in all rows of the $newpos column; however, if $maskP 
# is given as a reference to a list of 0 or 1's, then this mask 
# is used to decide which rows get a value. If $maskP is a number,
# then that number of rows are set. The routine works directly with 
# the incoming data, there is no return value; if you dont like 
# this, supply a copy of the table instead. 

    use strict;
    use Storable qw( dclone );

    my ( $newpos, $tableP, $newcolP, $maskP ) = @_;

    # ----------------------- CHECK ARGUMENTS ----------------------------
    
    if ( defined $maskP ) 
    {
	if ( ref $maskP and ref $maskP ne "ARRAY" ) 
	{
	    wit_die "Input mask is not an array";
	} 
	elsif ( not ref $maskP ) 
	{
	    if ( $maskP !~ /^\d+$/ )
	    {
		wit_die "Input mask is not a number";
	    }
	}
    }

    if ( not defined $newcolP ) { wit_die "No input column given"; }
    if ( not defined $newpos )  { wit_die "No input column number given"; }
    if ( not defined $tableP )  { wit_die "No input table given"; }

    # --------------------- COMPOSE COLUMN TO INSERT ---------------------

    my ( $row );
    my ( @insert_col );

    if ( ref $newcolP eq "ARRAY" )
    {
	@insert_col = @{ $newcolP };
    }
    elsif ( defined $maskP )
    {
	if ( ref $maskP eq "ARRAY" )
	{
	    for ( $row = 0; $row < @{ $maskP }; $row++ )
	    {
		if ( $maskP->[$row] ) 
		{
		    $insert_col[$row] = $newcolP;
		}
	    }
	}
	elsif ( not ref $maskP )
	{
	    for ( $row = 0; $row < $maskP; $row++ )
	    {
		push @insert_col, $newcolP;
	    }
	}
    }
    else 
    {
	for ( $row = 0; $row < @{ $tableP }; $row++ )
	{
	    push @insert_col, $newcolP;
	}
    }

    # -------------------- INSERT COLUMN -----------------------------

    my $tab_row;
    my $elem;

    for ( $row = 0; $row < @insert_col; $row++ )
    {
	if ( defined $insert_col[$row] )
	{
	    if ( ref $insert_col[$row] )
	    {
		$insert_col[$row] = dclone $insert_col[$row];
	    }

	    if ( defined $tableP->[$row] )
	    {
		if ( ref $tableP->[$row] ne "ARRAY" )
		{
		    $elem = $tableP->[$row];
		    undef $tableP->[$row];
		    $tableP->[$row]->[0] = $elem;
		}

		if ( $newpos > @{ $tableP->[$row] } )
		{
		    $tableP->[$row]->[$newpos] = $insert_col[$row];
		} 
		else 
		{
		    splice @{ $tableP->[$row] }, $newpos, 0, $insert_col[$row];
		}
	    }
	    else 
	    {
		$tableP->[$row]->[$newpos] = $insert_col[$row];
	    }
	}
    }
}

sub add_table_row {

# Inserts the values of @{ $newrowP } into @{ $tableP } at position
# $newpos, so that this new row has index $newpos after the insertion. 
# If the third argument ($newrowP) is a reference to an array, then that
# array will be inserted. If it is anything else, then this will be 
# inserted in all columns in the $newpos row; however, if $maskP is 
# given as a reference to a list of 0 or 1's, then this mask is used 
# to decide which columns are set. If given as a number, then all 
# columns are set. The routine works directly with the incoming data, 
# there is no return value; if you dont like this, supply a copy of 
# the table instead. 

    use strict;
    use Storable qw( dclone );

    my ( $newpos, $tableP, $newrowP, $maskP ) = @_;

    # ----------------------- CHECK ARGUMENTS ----------------------------
    
    if ( defined $maskP ) 
    {
	if ( ref $maskP and ref $maskP ne "ARRAY" ) 
	{
	    wit_die "Input mask is not an array";
	} 
	elsif ( not ref $maskP ) 
	{
	    if ( $maskP !~ /^\d+$/ )
	    {
		wit_die "Input mask is not a number";
	    }
	}
    }

    if ( not defined $newrowP ) { wit_die "No input row given"; }
    if ( not defined $newpos )  { wit_die "No input row number given"; }
    if ( not defined $tableP )  { wit_die "No input table given"; }


    # ---------------------- FIND MAX COLUMN INDEX -----------------------

    my $tab_row;
    my $maxcol = 1;

    foreach $tab_row ( @{ $tableP } )
    {
	if ( ref $tab_row eq "ARRAY" and @{ $tab_row } > $maxcol )
	{
	    $maxcol = @{ $tab_row };
	}
    }

    # ---------------------- COMPOSE ROW TO BE INSERTED ------------------

    my ( $row, $col );
    my ( @insert_row );

    if ( ref $newrowP eq "ARRAY" )
    {
	@insert_row = @{ $newrowP };
    }
    elsif ( defined $maskP )
    {
	if ( ref $maskP eq "ARRAY" )
	{
	    if ( ref $newrowP )
	    {
		for ( $col = 0; $col < @{ $maskP }; $col++ )
		{
		    $insert_row[$col] = dclone $newrowP if $maskP->[$col];
		}
	    } 
	    else 
	    {
		for ( $col = 0; $col < @{ $maskP }; $col++ )
		{
		    $insert_row[$col] = $newrowP if $maskP->[$col];
		}
	    }
	}
	elsif ( not ref $maskP )
	{
	    $maskP = $maxcol if $maskP < $maxcol;

	    if ( ref $newrowP )
	    {
		for ( $col = 0; $col < $maskP; $col++ )
		{
		    push @insert_row, dclone $newrowP;
		}
	    }
	    else 
	    {
		for ( $col = 0; $col < $maskP; $col++ )
		{
		    push @insert_row, $newrowP;
		}
	    }
	}
    }
    else 
    {
	for ( $col = 0; $col < $maxcol; $col++ )
	{
	    push @insert_row, $newrowP;
	}
    }

    # ---------------------- INSERT ROW -------------------------------

    if ( $newpos >= @{ $tableP } )
    {
	for ( $row = @{ $tableP }; $row < $newpos; $row++ )
	{
	    $tableP->[$row] = [ map { undef } ( 1 .. $maxcol ) ];
	}

	$tableP->[$newpos] = [ @insert_row ];
    } 
    else 
    {
	splice @{ $tableP }, $newpos, 0, [ @insert_row ];
    }
}

sub aliases_with_links {
    my ( $id, $user ) = @_;
    my ( $otherL, $other, @tmp, $tmp, $tmp1 );

    (defined($id) && defined($user)) || confess "prot=$id user=$user";

    $otherL = &ORF::aliasesO($id);
    $other  = "";
    @tmp    = ();
    while ( $tmp = shift(@$otherL) )
    {
	if (($tmp =~ /^sp\|/) || ($tmp =~ /^tr\|/) || ($tmp =~ /^tn\|/))
	{ 
	    $tmp1 = &prot_link( $tmp, $user );
	    push( @tmp, $tmp1 ); 
	} else {
	    if ( length($tmp) <= 12) { push( @tmp, $tmp ); }
	}
    }
    return ( $other = join("; ", @tmp) );
}

$blast_url   = "http://www.ncbi.nlm.nih.gov/cgi-bin/BLAST/nph-newblast?Jform=1";
$blast_type  = "post";
$ncbi_base   = "http://www.ncbi.nlm.nih.gov";

$entrez_url  = "http://www.ncbi.nlm.nih.gov/htbin-post/Entrez/query";
$entrez_type = "post";



sub ncbi_blast 
{
    my($prot) = @_;
    my($seq,@output);

    $seq = &ORF::prot_seq_ofO($prot);

    if ($$seq)
    {
	@output = &get_html($blast_url,
			    $blast_type,
			    [["PROGRAM","blastp"],
			     ["DATALIB","nr"],
			     ["INPUT_TYPE","Sequence in FASTA format"],
			     ["SEQUENCE",">$prot\n$$seq"],
			     ["EXPECT","10"],
			     ["CUTOFF","default"],
			     ["MATRIX","default"],
			     ["STRAND","both"],
			     ["FILTER","default"],
			     ["DESCRIPTIONS","100"],
			     ["ALIGNMENTS","100"],
			     ["OVERVIEW","is_set"],
			     ["GENETIC_CODE","standard"],
			     ["GAPPED_ALIGNMENT","is_set"]
			    ]
			    , $ncbi_base);
	
#	&patchup_document(\@output,$ncbi_base);
	$seq = join("",@output);
	return \$seq;
    }
    else
    {
	return \"Content-type: text/html\n\n<h1>Could not get seq of $prot</h1>\n";
    }
}

sub psi_blast 
{
    my($cgiQuery,$prot) = @_;
    my($seq,@output);

    $seq = &ORF::prot_seq_ofO($prot);

    if ($$seq)
    {
	&forward_request($cgiQuery,
			 "http://www.ncbi.nlm.nih.gov/cgi-bin/BLAST/nph-psi_blast",
			 [["PROGRAM","blastp"],
			  ["DATALIB","nr"],
			  ["GAPPED_ALIGNMENT","is_set"],
			  ["INPUT_TYPE"," Sequence in FASTA format "],
			  ["SEQUENCE",">$prot\n$$seq"],
			  ["EXPECT","10"],
			  ["FILTER","default"],
			  ["NCBI_GI","is_set"],
			  ["GRAPHIC_OVERVIEW","is_set"],
			  ["DESCRIPTIONS","100"],
			  ["ALIGNMENTS","100"],
			  ["E_THRESH","0.001"],
			  ]
			 );

#			  ["MAT_PARAM","BLOSOM62         11      1"]
    }
    else
    {
	return \"Content-type: text/html\n\n<h1>Could not get seq of $prot</h1>\n";
    }
}

sub patchup_document {
    my($doc,$base) = @_;
    my($i);

    foreach $line (@$doc)
    {
	$line =~ s/(href\s*=\s*\")\//$1$base\//ig;
	$line =~ s/(src\s*=\s*\")\//$1$base\//ig;
    }

    splice(@$doc,0,0,("Content-type: text/html\n","\n","<head>\n"));
}

sub get_a_form {
    my($html,$n) = @_;
    my(@out) = ();
    my(@html) = @$html;
    my($line);

    while (($line = shift @html) && (($line !~ /\<form\b/i) || ($n-- > 1))) {}
    if ($line)
    {
	push(@out,$line);
	while (($line = shift @html) && ($line !~ /\<\/form\>/i))
	{
	    push(@out,$line);
	}
	if ($line)
	{
	    push(@out,$line);
	}
    }
    return @out;
}

sub get_hidden_values {
    my($html) = @_;
    my(@out) = ();
    my(@html) = @$html;
    my($line);

    while (($line = shift @html) && ($line !~ /\<INPUT\b/i)) {}
    while ($line)
    {
	@lines = ();
	while ($line && ($line !~ /\>/)) 
	{ 
	    push(@lines,$line); 
	    $line = shift @html; 
	}
	if ($line)
	{
	    push(@lines,$line);
	    chop @lines;
	    $big_line = join(" ",@lines);
	    if ((($big_line =~ /\bhidden\b/i) || ($big_line =~ /\bsubmit\b/i)) &&
		((($big_line =~ /\bname\=\"([^\"]+)\"/i) && ($name=$1)) ||
		 (($big_line =~ /\bname\=([A-Za-z0-9_]+)/i) && ($name=$1))) &&
		((($big_line =~ /\bvalue\=\"([^\"]+)\"/i) && ($value=$1)) ||
		 (($big_line =~ /\bvalue\=([A-Za-z0-9_]+)/i) && ($value=$1))))
	    {
		push(@out,[$name,$value]);
	    }
	}
	while (($line = shift @html) && ($line !~ /\<INPUT\b/i)) {}
    }
    return @out;
}

sub forward_request
{
    my( $cgiQuery, $url, $kv_pairs) = @_;
    my( @args, $x);
        
    @args = ();
    foreach $x (@$kv_pairs)
    {
	$x->[0] = &url_encode( $x->[0] );
	$x->[1] = &url_encode( $x->[1] );
	push( @args, "$x->[0]=$x->[1]" );
    }
	
    if (@args > 0)
    {
	$url .= "?" . join("&",@args);
    }
    print $cgiQuery->redirect($url);

#    open LOG, ">/tmp/test2.log";
#    print LOG $cgiQuery->redirect($url);
#    close LOG;
 
}

sub get_html 
{
    my( $url, $type, $kv_pairs, $base ) = @_;
    my( $encoded, $ua, $args, @args, $out, @output, $x, $h);
    
    if (defined($base))
    {
	local($BASE) = $base;
    }
    
    local %link_elements =
	(
	 'a'    => 'href',
	 'img'  => 'src',
	 'form' => 'action',
	 'link' => 'href',
	 );
    
    if ($type =~/post/i)
    {
	$args = [];
	foreach $x (@$kv_pairs)
	{
#	    $x->[0] = &url_encode( $x->[0] );
#	    $x->[1] = &url_encode( $x->[1] );
	    push(@$args, ( $x->[0], $x->[1]) );
	}
	
	$ua = new LWP::UserAgent;
	my $request  = POST $url, $args;
	my $response = $ua->request($request);
	if ($response->is_success) 
	{
	    $out = $response->content;
	} 
	else 
	{
	    $out = "<H1>Error: " . $response->code . "</H1>" . $response->message;
	}
    }
    else
    {
	@args = ();
	foreach $x (@$kv_pairs)
	{
	    $x->[0] = &url_encode( $x->[0] );
	    $x->[1] = &url_encode( $x->[1] );
	    push( @args, "$x->[0]=$x->[1]" );
	}
	
	if (@args > 0)
	{
	    $url .= "?" . join("&",@args);
	}
	
	$out = &LWP::Simple::get($url);
    }

    if ($base)
    {
      $h = URLExpander->new( $base );
      $out = $h->expand( $out );
#	$h = HTML::Parse::parse_html( $out );
#	$h = HTML::Parser->new();
#	$h->parse( $out );
#	$h->traverse(\&expand_urls, 1);
#	$out = $h->as_HTML;
    }
    $out =  "Content-type: text/html\n\n" . $out;
    
    @output = split("\n",$out);
    foreach $out (@output) { $out .= "\n"; }
    return @output;
}

sub expand_urls
{
    my($e, $start) = @_;
    return 1 unless $start;

    my $attr = $link_elements{$e->tag};    
#    my $tag  = $e->tag;
#    my $attr_type = (ref $attr);
#    print STDERR "\$e=$e, \$e\-\>tag=$tag, \$attr=$attr\n";

    return 1 unless defined $attr;

    my $url = $e->attr($attr);
    return 1 unless defined $url;
#    print STDERR "\$url=$url\n";

    return 1 if ($url =~ m/^\#/);
    $e->attr($attr, url($url, $BASE)->abs->as_string);
}


sub html_encode {
    my($arg) = @_;
    
    $arg =~ s/\(/&#40;/g;
    $arg =~ s/\)/&#41;/g;
    $arg =~ s/\'/&#39;/g;
    $arg =~ s/\`/&#96;/g;
    $arg =~ s/\+/&#43;/g;
    $arg =~ s/\:/&#58;/g;
    return $arg;
}

sub data_dump {

#  IN: 1) a reference to any data structure, 
#      2) optional structure display name
# OUT: Between <PRE> tags, a readable display of the structure

    my ( $dataP, $name ) = @_;

    my @output;

    use Data::Dumper;
	       
    push @output, "<PRE>\n";

    if ( $name ) 
    {
	push @output, Data::Dumper->Dump( [$dataP], ["*$name"] );
    } 
    else 
    {
	push @output, Data::Dumper->Dump( [$dataP], ["*data"] );
    }
	
    push @output, "</PRE>\n";

    return @output;
}

sub display_menu {

    ( $headers,       # Headers; string or ref to list. 
      $items,         # Menu options; ref to list, or list of lists.
      $attribs        # Optional hash of rendition attributes.
    ) = @_;

    my ( $elem, $item, $type );
    my ( $headmax, $itemmax, $imax, $i );
    my ( $key, $row, $col );

    my %att;          # Table attributes
    my @output;       # HTML output 
   
    my %def_attribs = (		       
		       "LINE_COLOR"    => "#000000",
		       "LINE_WIDTH"    => 2,
		       "MENU_WIDTH"    => "100%",
		       "MENU_BGCOLOR"  => "#AAAAAA",
		       "MENU_PADDING"  => 0,
		       "MENU_SPACING"  => 0,
		       "MENU_BORDER"   => 0,
		       "HEAD_BGCOLOR"  => "#336666",
		       "HEAD_FGCOLOR"  => "#FFFFFF",
		       "HEAD_JUSTIFY"  => "left",
		       "HEAD_PADDING"  => 5,
		       "HEAD_SPACING"  => 0,
		       "HEAD_BORDER"   => 2,
		       "ITEM_ATTRIBS"  => "columns",
		       "ITEM_BGCOLOR"  => "#EEEEEE",
		       "ITEM_FGCOLOR"  => "#000000",
		       "ITEM_JUSTIFY"  => "middle",
		       "ITEM_PADDING"  => 2,
		       "ITEM_SPACING"  => 2,
		       "ITEM_BORDER"   => 0,
		      );

    if ( $headers ) 
    {
	if ( ref $headers eq "ARRAY" ) {
	    $headmax = @{ $headers };
	} elsif ( $headers ) {
	    $type = ref $headers;
	    confess "Headers has type $type, should be ARRAY";
	}
    } 
    else
    {
	$headmax = 0;
    }

    if ( $items ) 
    {
	if ( ref $items eq "ARRAY" ) {
	    $itemmax = @{ $items };
	} elsif ( $items ) {
	    $type = ref $items;
	    confess "Items has type $type, should be ARRAY";
	}
    } 
    else 
    {
	$itemmax = 0;
    }

    if ( $attribs ) 
    {
	if ( ref $attribs ne "HASH" ) {
	    $type = ref $attribs;
	    confess "Attribs has type $type, should be HASH";
	}
    }

    # Let incoming settings, if any, override the above defaults, 

    foreach $key ( keys %def_attribs ) 
    {
	if ( $key =~ /^HEAD/ ) 
	{
	    $imax = $headmax;
	} 
	elsif ( $key =~ /^ITEM/ ) 
	{
	    $imax = $itemmax;
	} 
	else 
	{
	    $imax = 0;
	}
	
	if ( $attribs and ref $attribs->{ $key } eq "ARRAY" ) 
	{
	    for ( $i = 0; $i <= $imax; $i++ ) {
		$att{$key}[$i] = $attribs->{$key}->[$i];
	    }
	} 
	elsif ( $attribs and defined $attribs->{ $key } )
	{
	    for ( $i = 0; $i <= $imax; $i++ ) {
		$att{$key}[$i] = $attribs->{$key};
	    }
	}
	else
	{
	    for ( $i = 0; $i <= $imax; $i++ ) {
		$att{$key}[$i] = $def_attribs{$key};
	    }
	}
    }
    
    if ( $att{'MENU_WIDTH'}[0] ) 
    { 
	$menu_widstr = qq (WIDTH="$att{'MENU_WIDTH'}[0]") 
    }

    if ( $att{'LINE_WIDTH'}[0] > 0 ) 
    {
	push @output, qq (
<TABLE BGCOLOR="$att{'LINE_COLOR'}[0]" CELLSPACING=0 CELLPADDING="$att{'LINE_WIDTH'}[0]" $menu_widstr>
<TR><TD>
);
	$menu_widstr = qq (WIDTH="100%");
    }

    push @output, qq 
(<TABLE BGCOLOR="$att{'MENU_BGCOLOR'}[0]" CELLPADDING="$att{'MENU_PADDING'}[0]" CELLSPACING=0 BORDER=0 $menu_widstr>
<TR><TD>
<TABLE BGCOLOR="$att{'MENU_BGCOLOR'}[0]" CELLPADDING=0 CELLSPACING="$att{'MENU_SPACING'}[0]" BORDER="$att{'MENU_BORDER'}[0]" WIDTH="100%">
);
    
    if ( $headers ) 
    {
	push @output, qq (
<TR><TD>
<TABLE CELLSPACING="$att{'HEAD_SPACING'}[0]" 
       CELLPADDING="$att{'HEAD_PADDING'}[0]" 
       BGCOLOR="$att{'HEAD_BGCOLOR'}[0]" 
       BORDER="$att{'HEAD_BORDER'}[0]" 
       WIDTH="100%">
);
	for ( $row = 0; $row < @{ $headers }; $row++ ) 
	{
	    $item = $headers->[$row];

	    push @output, qq (<TR><TD ALIGN="$att{'HEAD_JUSTIFY'}[$row]" NOWRAP>)
	                 .qq (<STRONG><FONT COLOR="$att{'HEAD_FGCOLOR'}[$row]">)
			 .qq ($item</FONT></STRONG></TD></TR>);
	}
	push @output, qq (</TABLE></TD></TR>);
    }
    
    if ( $items ) {
	
	push @output, qq (
<TR><TD>
<TABLE CELLPADDING="$att{'ITEM_PADDING'}[0]"
       CELLSPACING="$att{'ITEM_SPACING'}[0]"
           BGCOLOR="$att{'ITEM_BGCOLOR'}[0]" 
	    BORDER="$att{'ITEM_BORDER'}[0]" 
             WIDTH="100%">
);

	for ( $row = 0; $row < @{ $items }; $row++ ) 
	{
	    push @output, "<TR>\n";

	    $elem = $items->[$row];
	    
	    if ( ref $elem eq "ARRAY" ) 
	    {
		if ( $att{'ITEM_ATTRIBS'}[0] =~ /^colu/i ) 
		{
		    for ( $col = 0; $col < @{ $elem }; $col++ ) 
		    {
			$item = $elem->[$col];
			
			push @output, qq (<TD ALIGN="$att{'ITEM_JUSTIFY'}[$col]">)
		                     .qq (<FONT COLOR="$att{'ITEM_FGCOLOR'}[$col]">$item</FONT></TD>\n);
		    }
		}
		else 
		{
		    for ( $col = 0; $col < @{ $elem }; $col++ ) 
		    {
			$item = $elem->[$col];
			
			push @output, qq (<TD ALIGN="$att{'ITEM_JUSTIFY'}[$row]">)
		                     .qq (<FONT COLOR="$att{'ITEM_FGCOLOR'}[$row]">$item</FONT></TD>\n);
		    }
		}   
	    } 
	    else 
	    {
		push @output, qq (<TD ALIGN="$att{'ITEM_JUSTIFY'}[$row]">)
		             .qq (<FONT COLOR="$att{'ITEM_FGCOLOR'}[$row]">$elem</FONT></TD>\n);
	    }
	    push @output, "</TR>\n";
	}
	
	push @output, qq (
</TABLE>
</TD></TR>
);
    }

    push @output, qq (
</TABLE>
</TD></TR></TABLE>
);

    if ( $att{'LINE_WIDTH'}[0] > 0 ) 
    {
	push @output, "</TD></TR></TABLE>\n";
    }

    return @output;
}



sub matrix_to_html
{
    my ( $row_lblP, $col_lblP, $matrixP, $tbl_att, $caption, $align, $nowrap ) = @_;
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# Use of ``attribute'' fields to control formating:  
#
# $tbl_att --- if defined and a string, table is opened using '<TABLE $tbl_att>'.
#              (For backward compatibility, if ($tbl_att == 1), table is opened
#              using '<TABLE BORDER>'; however, the prefered syntax is to call
#              using an explicit $tbl_att = "BORDER" argument.)
#
#   $align --- if defined, sets the default alignment for cells and headers;
#              if undefined, cells and headers default to 'ALIGN="CENTER"'.
#
#  $nowrap --- for backward compatibility, if defined, forces a 'NOWRAP' on all elements;
#              DEPRECATED in favor of the element attribute fields.
#              WARNING: $nowrap and the element attribute fields should NEVER
#              be used simultaneously !!!  Unpredictable results may occur !!!
#
# Arguments $caption, and each element of @$row_lblP, @$col_lblP, and @@$matrixP
# have optional tab-delimited ``attribute'' fields that control table formatting.
#
# $caption --- if $caption  is of the form "Cap_String\tCap_Att", the caption 
#              is generated as '<CAPTION Cap_Att> cap_string </CAPTION>' --- 
#              e.g., to produce a centered caption 'Centered Caption', call 
#              using $caption = qq(Centered Caption\tALIGN="CENTER").
#
# row and column attributes --- elements of @$col_lblP and @$row_lblP may be of form:
#         "Header_String\tHeader_Att\tData_Att". 
#     'Header_Att' applies only to that row or column header, while 'Data_Att' applies 
#     to every element in the row or column. For example, to NOWRAP and RIGHT-justify 
#     the row-header, while CENTERing each data-cell and coloring its background 
#     'pale goldenrod', the easiest method is:
#     
#   $row_label .= qq(\tNOWRAP ALIGN="RIGHT"\tALIGN="CENTER" BGCOLOR="pale goldenrod");
#   push( @$row_lblP, $row_label );
#
# cell attributes --- row- and column-attributes can be overridden in a cell by appending
#      a tab-delimited 'cell attribute'; e.g., a particular cell can be LEFT-justified and
#      colored 'pale green' as follows:
#
#   $cell_data .= qq(\tALIGN="LEFT" BGCOLOR="pale green");
#
# The priority of attributes in decreasing order are as follows: cell, col, row --- 
#      and if none of these are defined, then it defaults to "$align$nowrap".
#-----------------------------------------------------------------------------------------
    my ( $cap_att, $row_att, $row_lbl_att, $col_att, $col_lbl_att, $cell_att );
    my ( $line, $i, $j, $row, $col, $column, $x, $pad );
    my ( @output );
    
    @output = ();
    if ( defined($tbl_att) && $tbl_att)
    {
	if ($tbl_att eq "1")
	{ 
# ... NOTE: this is a DEPRECATED feature, inserted for backward compatibility; 
# preferred usage is an explicit $tbl_att of "BORDER".
	    push(@output, "<TABLE BORDER>\n");
	} else {
	    push(@output,"<TABLE $tbl_att>\n");
	}
    } 
    else 
    {
	push(@output,"<TABLE>\n");
    }
    
    if ( (! defined($align)) || (! $align) )
    { 
	$align = qq( ALIGN="CENTER"); 
    }
    else
    {
	$align = qq( ALIGN="$align"); 
    }
    
# ... NOTE: this is a DEPRECATED feature, inserted for backward compatibility; 
# 'NOWRAP' should be a $row_att, $col_att, or $cell_att attribute.
    if ( defined($nowrap) || $nowrap )  { $nowrap = " NOWRAP"; } else { $nowrap = ""; }
    
    if ( $caption )
    {
	if (index($caption,"\t") >= 0)
	{
	    ( $caption, $cap_att ) = split( /\t/, $caption );
	}
	else
	{
	    $cap_att = "";
	}
	push(@output,"<CAPTION $cap_att><B> $caption </B></CAPTION>\n");
    }
    
    $line = "";
    foreach $column (@$col_lblP)
    {
	( $col, $col_lbl_att, $col_att ) = split( /\t/, $column );
	
	if (! defined($col_lbl_att)) { $col_lbl_att = "$align$nowrap"; }
        if (! defined($col_att))     { $col_att = ""; }
	
	$line .= " <TH $col_lbl_att> $col </TH> ";
    }
    $line = "<TR> <TH> </TH> " . $line . " </TR>";
    push(@output, "$line\n");
    
    for ($i=0; $i < @$row_lblP; $i++)
    {
	( $row, $row_lbl_att, $row_att ) = split( /\t/, $row_lblP->[$i] );
	if (! defined($row_lbl_att)) { $row_lbl_att = ""; }
        if (! defined($row_att))     { $row_att = "";     }
        if (! defined($row))         { $row = "";     }
	
	if ( $row_lbl_att )
	{
	    push(@output,"<TR> <TH $row_lbl_att$nowrap> $row </TH>\n");
	}
	else
	{
	    if ( $align )
	    {
		push(@output,"<TR> <TH $align$nowrap> $row </TH>\n");
	    }
	    else
	    {
		push(@output, qq(<TR> <TH ALIGN="RIGHT"$nowrap> $row </TH>\n) );
	    }
	}
	
	for ($j=0; $j < @$col_lblP; $j++)
	{
	    ( $col, $col_lbl_att, $col_att ) = split( /\t/, $col_lblP->[$j] );
	    if (! defined($col_lbl_att)) { $col_lbl_att = ""; }
	    if (! defined($col_att))     { $col_att = "";     }

	    if ( defined($x = $matrixP->[$i]->[$j]) )
	    {
		( $x, $cell_att ) = split( /\t/, $x );
		if (! defined($cell_att))  { $cell_att = "";  }
		
		if (! $cell_att )
		{
		    if ( $col_att ) { 
			$cell_att = $col_att; 
		    } else {
			if ( $row_att ) { 
			    $cell_att = $row_att; 
			} else{
			    $cell_att = "$align$nowrap"; 
			}
		    }
		}
		
		if (defined($x) && ($x ne ""))
		{
		    $pad = " " x abs(8 - length($x));
		    $x   = " <TD $cell_att> $pad$x </TD>\n";
		} else {
		    $x   = " <TD ALIGN=\"CENTER\">      --- </TD>\n";
		}
		push( @output, $x);
	    }
	}
	push(@output," </TR>\n");
    }
    push(@output,"</TABLE>\n");
    $i = join("",@output);
    return \$i;
}

sub html_table_contig 
{
# Returns a table specific to the textual contig browser. 

    my ( $row_lblP, 
	 $col_lblP, 
	 $matrixP, 
	 $attribsP ) = @_;

    my $row;          
    my $label;
    my $key;
    my ($bold_beg, $bold_end);

    my %att;          # Table attributes
    my @output;       # HTML output 

    # Some of these are overridden below, 

    my %def_attribs = (
		       "TABLE_BORDER"   => 0,
		       "CHEAD_BGCOLOR"  => "#555555",
		       "CHEAD_FGCOLOR"  => "#FFFFFF",
		       "CHEAD_PADDING"  => 2,
		       "CHEAD_WRAP"     => "NOWRAP",
                       "RHEAD_BGCOLOR"  => "#DDDDDD",
                       "RHEAD_FGCOLOR"  => "#000000",
		       "RHEAD_WRAP"     => "NOWRAP",
                       "ELEMS_BGCOLOR"  => "#DDDDDD",
                       "ELEMS_FGCOLOR"  => "#000000",
		       "ELEMS_PADDING"  => 0,
		       "ELEMS_SPACING"  => 2,
		       "ELEMS_WRAP"     => "",
		      );

    # Let incoming settings, if any, override the defaults,

    foreach $key ( %def_attribs ) 
    {
	if ( $attribsP->{ $key } 
	  or $attribsP->{ lc $key } ) 
	{
	    $att{$key} = $attribsP->{$key};
	} 
	else
 	{
	    $att{$key} = $def_attribs{$key};
	}
    }

    @output = qq (
<TABLE CELLPADDING="$att{'ELEMS_PADDING'}" 
       CELLSPACING="$att{'ELEMS_SPACING'}" 
       BORDER="$att{'TABLE_BORDER'}" 
       BGCOLOR="$att{'CHEAD_BGCOLOR'}">
);

    push @output, "<TR>";

    for ($i = 0; $i < (@{$col_lblP} - 1); $i++)   
    {
	$label = $col_lblP->[$i];

	push @output, qq (
<TD><TABLE  
       WIDTH="100%" 
       CELLPADDING="$att{'CHEAD_PADDING'}">
    <TR><TD ALIGN="$att{'CHEAD_JUSTIFY'}" $att{'CHEAD_WRAP'} BGCOLOR="$att{'CHEAD_BGCOLOR'}">
          <STRONG><FONT COLOR="$att{'CHEAD_FGCOLOR'}">$label</FONT></STRONG></TD>
    </TR>
   </TABLE>
</TD>
);
    }

    # Function Description, align left, 

    $label = $col_lblP->[ $#{$col_lblP} ];

    push @output, qq (
<TD><TABLE WIDTH="100%" CELLPADDING="$att{'CHEAD_PADDING'}">
       <TR><TD ALIGN="left" $att{'CHEAD_WRAP'} BGCOLOR="$att{'CHEAD_BGCOLOR'}">
          <STRONG><FONT COLOR="$att{'CHEAD_FGCOLOR'}">$label</FONT></STRONG></TD>
       </TR>
   </TABLE>
</TD>
);
    push @output, "</TR>\n";

    $colspan = @{$col_lblP};

    # Print the table, 

    for ( $row=0; $row < @{ $matrixP }; $row++ )
    {
        # Optional row headers, 

	push @output, "<TR>";

	if ( $row_lblP->[$row] =~ /\w/ ) 
	{
	    push @output, qq (
<TD ALIGN="center" $att{'RHEAD_JUSTIFY'} BGCOLOR="$att{'RHEAD_BGCOLOR'}">
<FONT COLOR="$att{'RHEAD_FGCOLOR'}">$row_lblP->[$row]</FONT></TD>
);
	    # Matrix elements, 

	    $function = pop @{ $matrixP->[$row] };

	    # All elements except the last, align center, 

	    foreach $elem ( @{ $matrixP->[$row] } ) {
		push @output,  
qq (<TD ALIGN="center" BGCOLOR="$att{'ELEMS_BGCOLOR'}"><FONT COLOR="$att{'ELEMS_FGCOLOR'}">$elem</FONT></TD>\n);
	    }

	    # Function description, align left, 

	    push @output, 
qq (<TD ALIGN="left" BGCOLOR="$att{'ELEMS_BGCOLOR'}"><FONT COLOR="$att{'ELEMS_FGCOLOR'}">$function</FONT></TD>\n);
	} 
	else 
	{
	    push @output, qq (<TR><TD COLSPAN=$colspan></TD></TR>\n);
	    push @output, qq (<TR><TD COLSPAN=$colspan></TD>\n);

	}
	push @output, "</TR>\n";
    }

    push @output, "<TR>";

    for ($i = 0; $i < (@{$col_lblP} - 1); $i++)   
    {
	$label = $col_lblP->[$i];

	push @output, qq (
<TD><TABLE  
       WIDTH="100%" 
       CELLPADDING="$att{'CHEAD_PADDING'}">
    <TR><TD ALIGN="$att{'CHEAD_JUSTIFY'}" $att{'CHEAD_WRAP'} BGCOLOR="$att{'CHEAD_BGCOLOR'}">
          <STRONG><FONT COLOR="$att{'CHEAD_FGCOLOR'}">$label</FONT></STRONG></TD>
    </TR>
   </TABLE>
</TD>
);
    }

    # Function Description, align left, 

    $label = $col_lblP->[ $#{$col_lblP} ];

    push @output, qq (
<TD><TABLE WIDTH="100%" CELLPADDING="$att{'CHEAD_PADDING'}">
       <TR><TD ALIGN="left" $att{'CHEAD_WRAP'} BGCOLOR="$att{'CHEAD_BGCOLOR'}">
          <STRONG><FONT COLOR="$att{'CHEAD_FGCOLOR'}">$label</FONT></STRONG></TD>
       </TR>
   </TABLE>
</TD>
);
    push @output, "</TR>\n";

    push @output, "</TABLE>\n";

    return @output;
} 

sub html_table_default
{
# Table template used to display most of WIT's tables. 

    my ( $row_lblP,       # Optional ref to leftmost column list
	 $col_lblP,       # Optional ref to top row list 
	 $matrixP,        # Table elements
	 $attribsP ) = @_;

# matrixP  - is a list of lists, giving a single or multi-column 
#            table, depending how you supply it. There may be an 
#            uneven number of elements per list, and HTML code 
#            as part of the elements is ok. 
# 
# attribsP - a hash of table layout attributes, see defaults 
#            below. Any element can be a simple string, or a 
#            reference to a list, in which case element in left
#            column, upper row and matrix columns are controlled
#            individually. Not possible to control each element
#            of matrixP, only column by column.
# 
    my ( $row, $col );
    my ( $rows, $cols );
    my ( $i, $imax );
    my ( $rowmax, $colmax, $matcolmax );

    my ( $label, $item );
    my ( $bbeg, $bend );

    my $key;
    my %att;          # Table attributes
    my @output;       # HTML output 

    my %def_attribs = (
		       "TABLE_BORDER"   => 1,
		       "TABLE_ATTRIBS"  => "columns",
		       "CHEAD_BGCOLOR"  => "#555555",
		       "CHEAD_FGCOLOR"  => "#FFFFFF",
		       "CHEAD_JUSTIFY"  => "center",
		       "CHEAD_PADDING"  => 2,
		       "CHEAD_WRAP"     => "NOWRAP",
		       "CHEAD_BOLD"     => 1,
                       "RHEAD_BGCOLOR"  => "#CCCCCC",
                       "RHEAD_FGCOLOR"  => "#000000",
		       "RHEAD_JUSTIFY"  => "center",
		       "RHEAD_WRAP"     => "NOWRAP",
		       "RHEAD_BOLD"     => 0,
                       "ELEMS_BGCOLOR"  => "#CCCCCC",
                       "ELEMS_FGCOLOR"  => "#000000",
		       "ELEMS_JUSTIFY"  => "center",
		       "ELEMS_PADDING"  => 1,
		       "ELEMS_WRAP"     => "",
		       "ELEMS_BOLD"     => 0,
		      );

    # Let incoming settings, if any, override the defaults,

    $rowmax = @{ $row_lblP } - 1;
    $colmax = @{ $col_lblP } - 1;
    $matrowmax = @{ $matrixP } - 1;

    $matcolmax = 0;

    foreach $row ( @{ $matrixP } ) 
    {
	$matcolmax = @{ $row } if @{ $row } > $matcolmax;
    }

    $matcolmax -= 1; 

    foreach $key ( keys %def_attribs ) 
    {
	if ( $key =~ /^CHEAD/ ) 
	{
	    $imax = $colmax;
	} 
	elsif ( $key =~ /^RHEAD/ ) 
	{
	    $imax = $rowmax;
	} 
	elsif ( $key =~ /^ELEMS/ ) 
	{
	    $imax = $matcolmax;
	}
	else 
	{
	    $imax = 0;
	}

	if ( ref $attribsP->{ $key } eq "ARRAY" ) 
	{
	    for ( $i = 0; $i <= $imax; $i++ ) {
		$att{$key}[$i] = $attribsP->{$key}->[$i];
	    }
	} 
	elsif ( $attribsP->{ $key } )
	{
	    for ( $i = 0; $i <= $imax; $i++ ) {
		$att{$key}[$i] = $attribsP->{$key};
	    }
	}
	else
	{
	    for ( $i = 0; $i <= $imax; $i++ ) {
		$att{$key}[$i] = $def_attribs{$key};
	    }
	}
    }

    if ( $att{'TABLE_BORDER'}[0] =~ /NOBORDER/i )
    {
	@output = qq (<TABLE CELLPADDING="$att{'ELEMS_PADDING'}[0]" CELLSPACING=0)
	        . qq ( BGCOLOR="$att{'CHEAD_BGCOLOR'}[0]">\n);
    }
    else
    {
	@output = qq (<TABLE CELLPADDING="$att{'ELEMS_PADDING'}[0]" CELLSPACING=0)
	        . qq ( BORDER="$att{'TABLE_BORDER'}[0]" BGCOLOR="$att{'CHEAD_BGCOLOR'}[0]">\n);
    }

#    use Data::Dumper;
#    push @output, "<PRE>";
#    push @output, Data::Dumper->Dump([ \%att ], [ "*att"]);
#    push @output, "colmax, rowmax = $colmax, $rowmax\n";
#    push @output, "matcolmax, matrowmax = $matcolmax, $matrowmax\n";
#    push @output, "</PRE>";


# Optional column headers, ie the top table row, 

    if ( @{ $col_lblP } ) 
    {
	push @output, "<TR>";

	for ( $col = 0; $col <= $colmax; $col++ ) 
	{
	    $item = $col_lblP->[$col];

	    if ( $att{'CHEAD_BOLD'}[$col] ) 
	    {
		$bbeg = "<STRONG>";
		$bend = "</STRONG>";
	    } 
	    else 
	    {
		$bbeg = "";
		$bend = "";
	    }

	    push @output, qq (<TD><TABLE WIDTH="100%" CELLPADDING="$att{'CHEAD_PADDING'}[$col]">)
	                . qq (<TR><TD ALIGN="$att{'CHEAD_JUSTIFY'}[$col]" BGCOLOR="$att{'CHEAD_BGCOLOR'}[$col]")
                        . qq ( $att{'CHEAD_WRAP'}[$col]>$bbeg<FONT COLOR="$att{'CHEAD_FGCOLOR'}[$col]">$item</FONT>$bend</TD>\n)
			. qq (    </TR>\n  </TABLE>\n</TD>\n);
	}
	push @output, "</TR>\n";
    }

    # Rest of table, 

    for ( $row=0; $row <= $matrowmax; $row++ )
    {
	push @output, "<TR>";

	if ( defined $row_lblP->[$row] ) 
	{
	    $item = $row_lblP->[$row];

	    if ( $att{'RHEAD_BOLD'}[$row] ) 
	    {
		$bbeg = "<STRONG>";
		$bend = "</STRONG>";
	    } 
	    else 
	    {
		$bbeg = "";
		$bend = "";
	    }

	    push @output, qq (<TD ALIGN="$att{'RHEAD_JUSTIFY'}[$row]")
	                . qq ( BGCOLOR="$att{'RHEAD_BGCOLOR'}[$row]" $att{'RHEAD_WRAP'}[$row]>)
                        . qq ($bbeg<FONT COLOR="$att{'RHEAD_FGCOLOR'}[$row]">$item</FONT>$bend</TD>\n);
	}

        # Matrix elements, apply attributes to either rows or columns,

	if ( $att{'TABLE_ATTRIBS'}[0] =~ /^col/i ) 
	{
	    for ( $col = 0; $col <= $matcolmax; $col++ ) 
	    {
		$item = $matrixP->[$row]->[$col];
		
		if ( $att{'ELEMS_BOLD'}[$col] ) 
		{
		    $bbeg = "<STRONG>";
		    $bend = "</STRONG>";
		} 
		else 
		{
		    $bbeg = "";
		    $bend = "";
		}

		push @output, qq (<TD ALIGN="$att{'ELEMS_JUSTIFY'}[$col]" BGCOLOR="$att{'ELEMS_BGCOLOR'}[$col]")
		            . qq ( $att{'ELEMS_WRAP'}[$col]>$bbeg<FONT COLOR="$att{'ELEMS_FGCOLOR'}[$col]">$item</FONT>$bend</TD>\n);
	    }
	} 
	else 
	{
	    if ( $att{'ELEMS_BOLD'}[$row] ) 
	    {
		$bbeg = "<STRONG>";
		$bend = "</STRONG>";
	    } 
	    else 
	    {
		$bbeg = "";
		$bend = "";
	    }

	    for ( $col = 0; $col <= $matcolmax; $col++ ) 
	    {
		$item = $matrixP->[$row]->[$col];
		
		push @output, qq (<TD ALIGN="$att{'ELEMS_JUSTIFY'}[$row]" BGCOLOR="$att{'ELEMS_BGCOLOR'}[$row]")
		            . qq ( $att{'ELEMS_WRAP'}[$row]>$bbeg<FONT COLOR="$att{'ELEMS_FGCOLOR'}[$row]">$item</FONT>$bend</TD>\n);
	    }
	}	    

        push @output, "</TR>\n";
    }

    push @output, "</TABLE>\n";

    return [ @output ];
} 


sub html_table_prot
{
# Returns a table specific to prot.cgi

    my ( $row_lblP, $col_lblP, $matrixP ) = @_;

    my @output;
    my $row;
    my $label;

    my $title_bgcolor = "#555555";
    my $title_fgcolor = "#FFFFFF";
    my $head_bgcolor  = "#999999";
    my $head_fgcolor  = "#FFFFFF";
    my $items_bgcolor = "#CCCCCC";
    my $items_fgcolor = "#000000";

    @output = qq (
<TABLE CELLPADDING=1 CELLSPACING=0 BORDER=1 BGCOLOR="$title_bgcolor">
<TR>
);

    $label = shift @{$col_lblP};
    push @output, qq (
<TD COLSPAN=4><TABLE WIDTH="100%" CELLPADDING=2>
    <TR><TD ALIGN="center" NOWRAP BGCOLOR="$title_bgcolor">
        <FONT COLOR="$title_fgcolor"><STRONG>$label</STRONG></FONT></TD>
    </TR></TABLE>
</TD>
);

    foreach $label (@{$col_lblP}) {
	push @output, qq (
<TD><TABLE WIDTH="100%" CELLPADDING=2>
    <TR><TD ALIGN="center" NOWRAP BGCOLOR="$title_bgcolor">
        <FONT COLOR="$title_fgcolor"><STRONG>$label</STRONG></FONT></TD>
    </TR></TABLE>
</TD>
);
    }

    push @output, "</TR>\n";

    for ( $row=0; $row < @{ $matrixP }; $row++ )
    {
	push @output, "<TR>\n";

        foreach $elem ( @{ $matrixP->[$row] } ) {
	    push @output, qq 
         (<TD Align="center" BGCOLOR="$items_bgcolor"><FONT COLOR="$items_fgcolor">$elem</FONT></TD>\n);
	}
	
	push @output, "</TR>\n";
    }

    push @output, "</TABLE>\n";

    return \@output;
} 

sub html_table
{
    use strict;

# Generic table routine for most of WIT's tables. Three inputs,

    my @user_tab = @{ shift() }; 

# @user_tab is a table of data items to be shown as HTML. It may be
# a list or a list of lists. Elements may be empty or undefined, 
# and there may be an uneven number of columns per row. HTML-tagged 
# text will be passed through and shown as HTML with its features 
# preserved. 
 
    my %user_tab_atts = %{ shift() };     

# %user_tab_atts is a hash of global table attributes (BORDER, 
# PADDING, SPACING, and more .. (update)
 
    my @user_elem_atts = @{ shift() };      

# @user_elem_atts is a table of hashes, where each hash contains 
# attributes for an element. This table will be "superimposed" on 
# the data table above (@user_tab), so the number of rows and columns 
# should match in the two. Elements may be empty and undefined, in 
# which case the attributes fall back on the defaults listed below. 
# If all elements in a row contains the same attribute, it is 
# written as a row tag where defined for rows, just to limit the 
# size of the HTML output. 

# One output, an array with HTML code, 

    my @output;  

# ------------------------------------------------------------------

    my ( @elem_atts, @row_atts, %tab_atts, $atts );
    my ( $row, $col );
    my ( $att, $att_val, @att_vals, %att_vals, %atts_used );
    my ( $tab_row, $att_row, $user_att_row );
    my ( @tags, $text, $val, $defined_tab_cols );
    my ( $i, $j );
    my ( $coltag, $siztag );

# To get the browser to display faster, find the highest number of
# columns in any row,
# 
# EW, THIS MAKES NETSCAPE STRETCH THE TABLE ALL ACROSS .. why .. 
#
#    my $cols = 0;
#
#    foreach $row ( @user_tab ) 
#    {
#	if ( @{ $row } > $cols )
#	{
#	    $cols = @{ $row };
#	}
#    }

# Default table attributes (there are more, but not supported by
# both HTML 3.2, Netscape and Explorer),

    my %tab_att_defs = (
			"ALIGN"       => "left",
			"BGCOLOR"     => "#BBBBBB",
			"BORDER"      => 1,
			"CELLPADDING" => 1,
			"CELLSPACING" => 0,
			"COLS"        => undef,  # $cols,
			"HSPACE"      => undef,
			"VSPACE"      => undef,
			"WIDTH"       => undef,
			"GRIDWIDTH"   => undef,
			"GRIDCOLOR"   => undef,
			"FRAMEWIDTH"  => undef,
			"FRAMECOLOR"  => undef,
			"FRAMEBORDER" => undef,
		       );

    # Default row attributes, 
    
    my %row_att_defs = (
			"ALIGN"    => "center",
			"VALIGN"   => "middle",
			"BGCOLOR"  => undef,
#			"WRAP"     => 1,      
		       );

    # Default element attributes, 

    my %elem_att_defs = (
			 "HEIGHT"    => undef,
			 "WIDTH"     => undef,
			 "ROWSPAN"   => undef, 
			 "COLSPAN"   => undef, 
			 "ALIGN"     => "center",
			 "VALIGN"    => undef,
			 "FGCOLOR"   => undef,
			 "BGCOLOR"   => undef,
			 "WRAP"      => 1,
			 "BLINK"     => undef,
			 "BOLD"      => undef,
			 "UNDERLINE" => undef, 
			 "FONTSIZE"  => undef,
			 "ITALICS"   => undef,
			 "STRIKE"    => undef,
			 "NBSP"      => 0,
			);

    # Remove undefined element defaults, 
    
    foreach $att ( keys %elem_att_defs )
    {
	delete $elem_att_defs{"$att"} if not defined $elem_att_defs{"$att"};
    }

    # ---------------- INITIALIZE TD AND TR ATTRIBUTES -------------------

    # Initialize element attributes (@elem_atts) with defaults above, but
    # put those that are defined for rows as well in @row_atts, 

    foreach $att ( keys %elem_att_defs )
    {
	next if not defined $elem_att_defs{"$att"};

	$att_val = $elem_att_defs{"$att"};

	if ( exists $row_att_defs{"$att"} )
	{
	    for ( $row = 0; $row < @user_tab; $row++ )
	    {
		$row_atts[$row]{"$att"} = $att_val;
	    }
	}
	else 
	{
	    for ( $row = 0; $row < @user_tab; $row++ )
	    {
		$tab_row = $user_tab[$row];
		$tab_row = [ $tab_row ] if not ref $tab_row;

		for ( $col = 0; $col < @{ $tab_row }; $col++ ) 
		{
		    if ( defined $tab_row->[$col] )
		    {
			$elem_atts[$row][$col]{"$att"} = $att_val;
		    }
		}
	    }
	}  
    }
    
    # ---------------- OVERRIDE/ADD USER TD ATTRIBUTES ------------------

    undef %atts_used;

    # First find which attributes the user has specified (usually just 
    # a few.) 

    for ( $row = 0; $row < @user_elem_atts; $row++ ) 
    {
	next if not defined $user_elem_atts[$row];

	$user_att_row = $user_elem_atts[$row];
	$user_att_row = [ $user_att_row ] if not ref $user_att_row;
	
	for ( $col = 0; $col < @{ $user_att_row }; $col++ )
	{ 
	    next if not defined $user_att_row->[$col];
	    next if not defined $user_tab[$row][$col];
	    
	    foreach $att ( keys %{ $user_att_row->[$col] } )
	    {
		$atts_used{"$att"} = 1;
	    }
	}
    }

    # After that, look through @user_elem_atts for each of these
    # and move those to @row_atts that are shared for all elements in a
    # row. Speeds up routine and produces much less HTML output,

    for ( $row = 0; $row < @user_elem_atts; $row++ )
    {
	next if not defined $user_elem_atts[$row];

	$tab_row = $user_tab[$row];
	$att_row = $user_elem_atts[$row];
	$att_row = [ $att_row ] if not ref $att_row;

	foreach $att ( keys %atts_used ) 
	{
	    undef @att_vals;
	    $defined_tab_cols = 0;

	    for ( $col = 0; $col < @{ $tab_row }; $col++ ) 
	    {
		if ( defined $tab_row->[$col] ) 
		{
		    $defined_tab_cols++;

		    if ( defined $att_row->[$col]{"$att"} )
		    { 
			push @att_vals, $att_row->[$col]{"$att"};
		    }
		}
	    }

	    %att_vals = map { $_, 1 } @att_vals;

#  	    my $var1 = scalar @att_vals;
#	    my $var2 = scalar @{ $att_row };
#	    my $var3 = keys %att_vals;

#	    print STDERR "att = $att, defined = $defined_tab_cols, \@att = $var1, \@row = $var2, keys = $var3\n";
    
	    if ( $defined_tab_cols == @att_vals     and 
		 keys %att_vals == 1                and
		 exists $row_att_defs{"$att"} ) 
	    {
		$row_atts[$row]{"$att"} = ( keys %att_vals )[0];
	    }
	    else 
	    {
		for ( $col = 0; $col < @{ $att_row }; $col++ ) 
		{
		    if ( defined $att_row->[$col]{"$att"} and 
		         defined $user_tab[$row][$col] )
		    {
			$elem_atts[$row][$col]{"$att"} = $att_row->[$col]{"$att"};
		    }
		}
	    }
	}
    }

    # ------------------- SET TABLE ATTRIBUTES ------------------------

    foreach $att ( keys %tab_att_defs )
    {
	if ( defined $user_tab_atts{"$att"} )
	{
	    $tab_atts{"$att"} = $user_tab_atts{"$att"};
	} 
	elsif ( defined $tab_att_defs{"$att"} )
	{
	    $tab_atts{"$att"} = $tab_att_defs{"$att"};
	}
    }
	    
    # ---------------- MOVE COMMON TR ATTRIBUTES TO TABLE ---------------

    # well, except those that the user has already set, and those with 
    # different meaning: WIDTH and ALIGN .. 

    if ( defined @row_atts ) 
    {
	foreach $att ( keys %row_att_defs )
	{
	    next if defined $user_tab_atts{"$att"};

	    next if $att eq "WIDTH";
	    next if $att eq "ALIGN";

	    undef %att_vals;

	    for ( $i=0; $i < @row_atts; $i++ )
	    {
		if ( defined $row_atts[$i]{"$att"} 
		         and $row_atts[$i]{"$att"} ne "" )
		{
		    $att_vals{ $row_atts[$i]{"$att"} }++;
		}
	    }

	    if ( (keys %att_vals) == 1  and 
		 $att_vals{ (keys %att_vals)[0] } == @row_atts and
		 exists $tab_att_defs{"$att"} ) 
	    {
		$tab_atts{"$att"} = (keys %att_vals)[0];
		
		for ( $i=0; $i < @row_atts; $i++ ) 
		{
		    delete $row_atts[$i]{"$att"};
		}
	    }
	}
    }

    # -------------------- PRINT TABLE ATTRIBUTES -----------------------

    my @trail_tags;
    my $border_tag = "";

    if ( defined $tab_atts{"FRAMEWIDTH"} and
	 defined $tab_atts{"FRAMECOLOR"} )
    {
	$border_tag = qq ( BORDER="$tab_atts{'FRAMEBORDER'}") if $tab_atts{'FRAMEBORDER'};

	push @output, qq (<TABLE CELLSPACING="0" CELLPADDING="$tab_atts{'FRAMEWIDTH'}" BGCOLOR="$tab_atts{'FRAMECOLOR'}"$border_tag><TR><TD>\n);
	push @trail_tags, "</TD></TR></TABLE>\n";
    }

    if ( defined $tab_atts{"GRIDWIDTH"} and
	 defined $tab_atts{"GRIDCOLOR"} )
    {
	push @output, qq (<TABLE CELLSPACING=0 CELLPADDING=0 BGCOLOR="$tab_atts{'GRIDCOLOR'}">\n);
	push @output, qq (<TR><TD>\n\n);

	unshift @trail_tags, qq (\n</TD></TR></TABLE>\n);

	$tab_atts{"CELLSPACING"} = $tab_atts{"GRIDWIDTH"};
	$tab_atts{"WIDTH"} = "100%";

	delete $tab_atts{"GRIDWIDTH"};
	delete $tab_atts{"GRIDCOLOR"};
	delete $tab_atts{"BORDER"};
    }

    @tags = ();

    foreach $att ( keys %tab_atts )
    {
	$val = $tab_atts{"$att"};
	push @tags, qq ( $att="$val");
    }

    push @output, "<TABLE" . (join "", @tags) . ">\n";

    # Print table, 

    for ( $row = 0; $row < @user_tab; $row++ ) 
    {
	# ------------------------ PRINT TR ATTRIBUTES -----------------

	@tags = ();

	if ( defined $row_atts[$row] )
	{
	    foreach $att ( keys %{ $row_atts[$row] } )
	    {
		if ( $att eq "WRAP" )
		{
		    push @tags, " NOWRAP" if $row_atts[$row]{"$att"} eq "0";
		} 
		else 
		{
		    $val = $row_atts[$row]{"$att"};
		    push @tags, qq ( $att="$val");
		}
	    }
	}
	
	push @output, "<TR" . (join "", @tags) . ">\n";

	# -------------------------- SET TD ATTRIBUTES ----------------

	$tab_row = $user_tab[$row];

	if ( not ref $tab_row ) 
	{
	    $tab_row = [ $tab_row ];
	}

	for ( $col = 0; $col < @{ $tab_row }; $col++ ) 
	{
	    @tags = ();
	    $atts = $elem_atts[$row][$col];

	    foreach $att ( "HEIGHT", "WIDTH", "ROWSPAN", "COLSPAN", "BGCOLOR" )
	    {
		if ( defined $atts->{"$att"} )
		{
		    $val = $atts->{"$att"};
		    push @tags, qq ( $att="$val");
		}
	    }
	    
#	    if ( defined $atts->{"ALIGN"} and 
#	  	         $atts->{"ALIGN"} !~ /^left$/i ) 
	    if ( defined $atts->{"ALIGN"} ) 
	    {
		push @tags, qq ( ALIGN="$atts->{'ALIGN'}");
	    }
	    
	    if ( defined $atts->{"VALIGN"} and
	                 $atts->{"VALIGN"} !~ /^center$/i )
	    {
		push @tags, qq ( VALIGN="$atts->{'VALIGN'}");
	    }
	    
	    if ( defined $atts->{"WRAP"} and
		         $atts->{"WRAP"} != 1 )
	    {
		push @tags, qq ( NOWRAP);
	    }
	    
	    # ------------------ SET TEXT ELEMENT TAGS -------------------

	    if ( $tab_row->[$col] )
	    {
		$text = $tab_row->[$col];
	    } 
	    else 
	    {
		$text = "&nbsp;";
	    }
	    
	    if ( defined $atts->{"FGCOLOR"} ) {
		$coltag = " COLOR=$atts->{'FGCOLOR'}";
	    } else {
		$coltag = "";
	    }

	    if ( defined $atts->{"FONTSIZE"} ) {
		$siztag = " SIZE=$atts->{'FONTSIZE'}";
	    } else {
		$siztag = "";
	    }

	    if ( $coltag or $siztag ) 
	    {
		$text = qq (<FONT$coltag$siztag>$text</FONT>);
	    }
	    
	    if ( defined $atts->{"BLINK"} )
	    {
		$text = qq (<BLINK>$text</BLINK>);
	    }

	    if ( defined $atts->{"BOLD"} )
	    {
		$text = qq (<STRONG>$text</STRONG>);
	    }
	    
	    if ( defined $atts->{"UNDERLINE"} )
	    {
		$text = qq (<U>$text</U>);
	    }
	    
	    if ( defined $atts->{"ITALICS"} )
	    {
		$text = qq (<I>$text</I>);
	    }

	    if ( defined $atts->{"STRIKE"} )
	    {
		$text = qq (<STRIKE>$text</STRIKE>);
	    }
	    
	    if ( defined $atts->{"NBSP"} )
	    {
		$text = "&nbsp;" x $atts->{"NBSP"} .
		        $text .
			"&nbsp;" x $atts->{"NBSP"};
	    }
	    push @output, "   <TD" . (join "", @tags) . "> $text </TD>\n";
	}
	push @output, "</TR>\n";
    }

    push @output, "</TABLE>\n";

    if ( @trail_tags )
    {
	push @output, @trail_tags;
    }

    return @output;
}

sub prot_link {
    my ( $prot, $user, $display_id ) = @_;
    my ( $id, $url, $org );

    if ( ! $display_id ) 
    { 
	$display_id = $prot;
    }

    (defined($prot) && defined($user)) || confess "prot=$prot user=$user";

    if (( &ORF::typeO($prot) eq "rid" ) &&
	($org = substr($prot,1,2)) && 
	(-d "$orgs_dir/$org"))
    {
	if (&ORF::relevant_sp_with_structures($prot))
	{
	    return qq (<A HREF="$cgi_url/prot.cgi?prot=$prot&user=$user">$display_id\*</A>);
	}
	else
	{
	    return qq (<A HREF="$cgi_url/prot.cgi?prot=$prot&user=$user">$display_id</A>);
	}
    } 
#    elsif (( &ORF::typeO($prot) eq "gi" ) && ($prot =~ /gi\|(\d+)/))
#    {
#	$id  = $1;
#	$url = "http://www.ncbi.nlm.nih.gov/htbin-post/Entrez/query?" . 
#	       "uid=$id&form=6&db=p&Dopt=g";
#
#	return "<a href=$url> $prot </a>";
#    }
    elsif (( &ORF::typeO($prot) eq "sp" ) or 
	   ( &ORF::typeO($prot) eq "gi" ))
    {
#	$url = "http://www.expasy.ch/cgi-bin/sprot-search-ac?$1";
	$url = "$cgi_url/prot.cgi?prot=" . &url_encode($prot);
	return qq (<a href="$url">$display_id</a>);
    }

    if ($display_id =~ /\w/) {
	return $display_id;
    } else {
	return $prot;
    }
}

sub put_in_links {
    my($seq,$user) = @_;

    $seq = &put_in_rid_links($seq,$user);
    $seq = &put_in_sp_links($seq);
    return $seq;
}

sub put_in_rid_links {
    my($seq,$user) = @_;
    my($before,$rid,$after);

    if ($$seq =~ /^(.*)(R[A-Z]{2}\d{5})(.*)$/)
    {
	$before = $1; $rid = $2; $after = $3;
	return \( ${&put_in_rid_links(\$before,$user)}
	        . &prot_link($rid,$user)
		. ${&put_in_rid_links(\$after,$user)}
                );
    }
    return $seq;
}

sub put_in_sp_links {
    my($seq) = @_;
    my($before,$sp,$after);

    if ($$seq =~ /^(.*)(sp\|[OPQ]\d[0-9A-Z]{3}\d)(.*)$/)
    {
	$before = $1; $sp = $2; $after = $3;
	return \( ${&put_in_sp_links(\$before)}
                . &prot_link($sp,"")
		. ${&put_in_sp_links(\$after)}
                );
    }
    return $seq;
}

sub put_in_real_sp_links {
    my($seq) = @_;
    my($x,@lines);

    @lines = split(/\n/,$$seq);
    foreach $x (@lines)
    {
	$x = &put_in_real_sp_linksR($x);
    }
    $x = (join("\n",@lines) . "\n");
    return \$x;
}

sub put_in_real_sp_linksR {
    my($seq) = @_;
    my($before,$sp,$after);

    if ($seq =~ /^(.*)([OPQ]\d[0-9A-Z]{3}\d)(.*)$/)
    {
	$before = $1; $sp = $2; $after = $3;
	return (&put_in_real_sp_linksR($before) . 
	        "<a href=http://www.expasy.ch/cgi-bin/sprot-search-ac?$sp>$sp</a>" .
		&put_in_real_sp_linksR($after)
                );
    }
    return $seq;
}

sub invoke_bcm 
{
    my($prot,$program) = @_;
    my($seq,@output);

    $seq = &ORF::prot_seq_ofO($prot);
    if ($$seq)
    {
	@output = &get_html("http://dot.imgen.bcm.tmc.edu:9331/cgi-bin/seq-search/protein-search.pl",
			    "GET",
			    [["seq_data",$$seq],
			     ["seq_name",$prot],
			     ["program_name",$program],
			     ["submit","Perform Search"]
			    ]);

	$seq = join("",@output);
	return \$seq;
    }
    else
    {
	return \"Content-type: text/html\n\n<h1>Could not get seq of $prot</h1>\n";
    }
}


sub invoke_psort {
    my($prot) = @_;
    my($seq,@lines,$x);

    $seq = &ORF::prot_seq_ofO($prot);
    if ($$seq)
    {
	@lines = `cat psort.form`;
	foreach $x (@lines)
	{
	    $x =~ s/\$cds_to_psort/$prot/;
	    $x =~ s/\$seq/$$seq/;
	}
	splice(@lines,0,0,("Content-type: text/html\n","\n"));
	$seq = join("",@lines);
	return \$seq;
    }
    return \"";
}

sub invoke_pfscan 
{
    my($prot) = @_;
    my($seq,@lines,$x);

    $seq = &ORF::prot_seq_ofO($prot);
    if ($$seq)
    {
	@output = &get_html(  "http://www.isrec.isb-sib.ch/cgi-bin/PFSCAN_form_parser"
			    , "POST"
			    , [  ["database", "all"]
			       , ["machine",  "isrec-insect"]
			       , ["sensitivity", "default"]
			       , ["sortby",   "score"]
			       , ["comm",     $prot]
			       , ["format",   "plain_text"]
			       , ["seq",      $$seq]
			       , ["submit",   "Run ProfileScan"]
			       ]
			    , "http://www.isrec.isb-sib.ch"
			    );
	
#	print STDERR @output;
	
	$x = join("",@output);
	return \$x;
    }
    return \"";
}

sub cog_url {
    my($prot) = @_;
    my($seq);

    $seq = &ORF::prot_seq_ofO($prot);
    if ($$seq)
    {
	return &url_for("http://www.ncbi.nlm.nih.gov/cgi-bin/COG/nph-cognitor"
			    , [["submit",   "compare to COGs"]
			       , ["seq",      $$seq]
			       ]);
    }
    return "";
}

sub url_for {
    my( $url, $kv_pairs ) = @_;
    my( $x,@args);
    
    @args = ();
    foreach $x (@$kv_pairs)
    {
	$x->[0] = &url_encode( $x->[0] );
	$x->[1] = &url_encode( $x->[1] );
	push( @args, "$x->[0]=$x->[1]" );
    }
	
    if (@args > 0)
    {
	return ($url . "?" . join("&",@args));
    }
    return $url;
}
    
sub invoke_emotif
{
    my($cgiQuery,$prot) = @_;
    my($url,$seq);

    if ($seq = &ORF::prot_seq_ofO($prot))
    {
	$url = "http://dna.stanford.edu/cgi-bin/emotif/nph-identify?sequence=$$seq";
	print $cgiQuery->redirect($url);
    }
    else
    {
	print $cgiQuery->header;
	print "<h1>Cannot get to Stanford IDENTIFY, sorry</h1>\n";
    }
    exit;
}


sub invoke_cog
{
    my($cgiQuery,$prot) = @_;
    my($url);

    if ($url = &cog_url($prot))
    {
	print $cgiQuery->redirect($url);
    }
    else
    {
	print $cgiQuery->header;
	print "<h1>Cannot get to NCBI COGs, sorry</h1>\n";
    }
    exit;
}

sub invoke_protscale {
    my($prot) = @_;
    my($seq,@output);

    $seq = &ORF::prot_seq_ofO($prot);
    if ($$seq)
    {
#	print STDERR "seq=$$seq\n\n";
	
	$BASE = 'http://expasy.hcuge.ch/';
	$url  = $BASE . 'cgi-bin/protscale.pl?1';
	
	@output = &get_html($url, "POST"
			    , [  ["prot_id"      => ""]
			       , ["sequence"     => $$seq]
			       , ["scale"        => "Hphob.+/+Kyte+&+Doolittle"]
			       , ["window"       => "21"]
			       , ["weight_edges" => "100"]
			       , ["weight_var"   => "linear"]
			       , ["norm"         => "no"]]
			    , $BASE);
	
	$seq = join("",@output);
	return \$seq;
    }
    else
    {
	return \"Content-type: text/html\n\n<h1>Could not get seq of $prot</h1>\n";
    }
}


sub invoke_tmpred 
{
    my($prot) = @_;
    my($seq,@output);

    $seq = &ORF::prot_seq_ofO($prot);
    if ($$seq)
    {
	@output = &get_html("http://www.isrec.isb-sib.ch/cgi-bin/TMPRED_form_parser",
			    "POST",
			    [["min",14],
			     ["max",41],
			     ["outmode","html"],
			     ["comm",$prot],
			     ["format","plain_text"],
			     ["seq",$$seq]
			    ], "http://www.isrec.isb-sib.ch/" );
	
	$seq = join("",@output);
	return \$seq;
    }
    else
    {
	return \"Content-type: text/html\n\n<h1>Could not get seq of $prot</h1>\n";
    }
}



sub gi_entries {
    my(@gids) = @_;
    my($id,$ids,$n,@out,$x);

    foreach $id (@gids)
    {
	$id =~ s/gi\|(\d+)/$1/g;
    }
    $ids = join(",",@gids);
    $n   = @gids;
    @out = &get_html($entrez_url,$entrez_type,
		     [["form",6],
		      ["db","p"],
		      ["dispmax",$n],
		      ["term",""],
		      ["cachestart",0],
		      ["cachestop",$n],
		      ["totaluids",$n],
		      ["uidstring",$ids],
		      ["field","Sequence ID"],
		      ["submit","Display"],
		      ["DisplayOption","GenPept Report"],
		      ["doclistpage",1]]);
     return \@out;
}


sub fasta_sequence_of_gids {
    my(@gids) = @_;
    my(@answer) = ();
    my($id,$ids,$n,$out,$gid,$de,$os);

    $out = &gi_entries(@gids);

    while (($_ = shift @$out) && ($_ !~ /<pre>LOCUS\s+\S+/)) {}
    while ($_ && ($_ =~ /<pre>LOCUS\s+(\S+)/))
    {
#       print STDERR "locus $_";
	$gid = $1; $de = ""; $os = "";
	$_ = shift @$out;
	while ($_ && ($_ !~ /^\/\//))
	{
#	    print STDERR "processing $_";
	    if ($_ =~ /^DEFINITION\s+(\S.*\S)\s*/)
	    {
		$de = $1;
#	        print STDERR "de=$de\n";
		$_ = shift @$out;
	    }
	    elsif ($_ =~ /^\s*ORGANISM.*>([^<]+)<\/a>/)
	    {
		$os = $1;
#	        print STDERR "os=$os\n";
		$_ = shift @$out;
	    }
	    elsif ($_ =~ /^\s*ORIGIN/)
	    {
#	        print STDERR "origin\n";
		$line = ">gi|$gid";
		if ($de) { $line .= " $de"; }
		if ($os) { $line .= " \[$os\]";}
		push(@answer,"$line\n");

		while (($_ = shift @$out) && ($_ !~ /^\/\//))
		{
		    $_ =~ s/^\s*\d+\s*//;
		    $_ =~ s/\s//g;
		    push(@answer, "$_\n");
		}
	    }
	    else
	    {
		$_ = shift @$out;
	    }
	}
	while ($_ && ($_ = shift @$out) && ($_ !~ /<pre>LOCUS\s+\S+/)) {}
    }
    return \@answer;
}

sub filter_special_chars
{
    # Replace character entities of the form &\w+; with the corresponding
    # images from ~www/docroot/bitmaps/char_ents/ .
    # Absence of an image in char_ents/ directory means it is OK to let
    # browser display it.

    my $str = shift @_;
    my %available_ents;
    foreach ( split "\n", `ls ~www/docroot/bitmaps/char_ents/` ) {
	($ent) = split /\./;
	$available_ents{$ent}++;
    }

    $str =~ s%&(\w+);% ($available_ents{$1} ? "<img src=\"/bitmaps/char_ents/$1.gif\" align=bottom>" : "&$1;") %eg;

    # fix some chars with descenders:
    $str =~ s%/([bgm])gr.gif" align=bottom%/$1gr.gif" align=texttop%g;

    $str;
}

sub enzyme_link {
    my($role,$org,$user) = @_;

    if ($role =~ /(\d+\.\d+\.\d+\.\d+)/)
    {
	return "<A HREF=$cgi_url/fr.cgi?fr=$1\&org=$org\&user=$user\&where=local_enz>$role</a>";
    }
    else
    {
	return $role;
    }
}

sub kegg_map_link {
    my($map) = @_;
    
    return "<a href=http://kegg.genome.ad.jp/dbget-bin/show_pathway?$map>$map</a>";
}

sub kegg_enzyme_link {
    my($role) = @_;

    if ($role =~ /(\d+\.\d+\.\d+\.\d+)/)
    {
	return qq(<A HREF="http://www.genome.ad.jp/dbget-bin/www_bget?ec:) . 
	    $1 . qq(">) . $role . qq(</A>);
    }
    else
    {
	return $role;
    }
}

sub enz_forms_link {
    my($role) = @_;

    if (($role =~ /(\d\.\d+\.\d+\.\d+)/) && (-s "$html_dir/EnzymeForms/$1.html"))
    {
	return "<a href=$html_url/EnzymeForms/$1.html>Forms of Sequenced Versions of $1</a>";
    }
    return "";
}

sub pw_link {
    my($pw,$org,$user,$window) = @_;

    if ( $window )
    {
	return \(qq (<A HREF="$cgi_url/pw.cgi?pw=$pw&org=$org&user=$user">$pw</A>));
    }
    else 
    {
	return \(qq (<A HREF="$cgi_url/pw.cgi?pw=$pw&org=$org&user=$user">$pw</A>));
    }	
}

sub red_blue_ramp {

# Fits a scale of colors to a scale of integers. 
#
# IN: Upper and lower bound integers
# OUT: hash where key is integer, and value is color.

    my $int_beg = shift;
    my $int_end = shift;

    my %col_ramp;

    $scale = abs 255.0 / ( $int_end - $int_beg );

    if ($int_beg < $int_end) 
    {
	for ( $i = $int_beg; $i <= $int_end; $i++) {
	    
	    $red_col = ($i - $int_beg) * $scale;
	    $blu_col = 255 - $red_col;
	    
	    $color = "#" . (sprintf "%2.2X",$red_col) . "00" 
	                 . (sprintf "%2.2X",$blu_col);
	    
	    $col_ramp{$i} = $color;
	}
    }
    else 
    {
	for ( $i = $int_beg; $i >= $int_end; $i--) {
	    
	    $red_col = ($int_beg - $i) * $scale;
	    $blu_col = 255 - $red_col;
	    
	    $color = "#" . (sprintf "%2.2X",$red_col) . "00" 
	                 . (sprintf "%2.2X",$blu_col);
	    
	    $col_ramp{$i} = $color;
	}
    }	
    return \%col_ramp;
}

sub results_headbar {

    ( $left_text, $right_text, $bgcolor ) = @_;

    if ( not $bgcolor )
    {
	$bgcolor = "#229922";
    }
    
    return qq (
<TABLE BORDER=0 BGCOLOR="#000000" WIDTH=100% CELLPADDING=2 CELLSPACING=0>

<TR><TD WIDTH="100%">
    <TABLE BGCOLOR="$bgcolor" BORDER=0 WIDTH=100% CELLSPACING=0 CELLPADDING=5>
    <TR>
       <TD WIDTH="40%" ALIGN="left" NOWRAP>
            <FONT FACE="Arial,Helvetica" COLOR="#FFFFFF">
            <STRONG><BIG><BIG><BIG>&nbsp;WIT&nbsp;&nbsp;</BIG></BIG>$left_text</BIG></STRONG>
	    </FONT></TD>
       <TD ALIGN="right" NOWRAP>
            <FONT FACE="Arial,Helvetica" COLOR="#FFFFFF">
            <STRONG><BIG>$right_text</BIG></STRONG></FONT></TD>
    </TR>
    </TABLE>
</TD>
</TR>
</TABLE>
);
}

sub rid_number {
    my $rid = shift;
    
    if ( not $rid )
    {
	return "";
    } 
    elsif ( $rid =~ /^R[A-Z][A-Z]0*([0-9]+)/ ) 
    {
	return $1;
    } 
    elsif ( $rid =~ /^[A-Z][A-Z]RNA0*([0-9]+)/ )
    { 
	return qq ((<FONT COLOR="#FF0000">$1</FONT>));
    }
    else 
    {
	return "";
    }
}

sub set_role_link {
    my($func,$org,$user) = @_;

    (defined($func) && defined($org) && defined($user))
	|| confess "org=$org user=$user func=$func";
    return \("<a href=$cgi_url/fr.cgi?org=$org\&user=$user\&fr=" . &url_encode($func) . ">$func</a>");
}

sub set_role_links {
    my($funcL,$org,$user) = @_;
    my($func,$links);

    foreach $func (@$funcL) 
    {
	$func = ${&set_role_link($func,$org,$user)};
    }
    $links =  join("; ",@$funcL);
    return \$links;
}

sub set_attrib_col {

# Writes the values of @{ $newcolP } to the column in @{ $tableP } 
# at position $newpos. If the column does not exist, then it will.
# If $newcolP is a reference to an array, that array will replace 
# what was there before. If it is anything else, then this will be
# written to all rows of the $newpos column; however, if $maskP 
# is given as a reference to a list of 0 or 1's, then this mask 
# is used to decide which rows are set. If given as a number, then 
# all rows are set. The routine works directly with the incoming 
# data, there is no return value; if you dont like this, supply a 
# copy of the table instead. 

    use strict;
    use Storable qw( dclone );

    my ( $newpos, $tableP, $newcolP, $maskP ) = @_;

    # ----------------------- CHECK ARGUMENTS ----------------------------
    
    if ( defined $maskP ) 
    {
	if ( ref $maskP and ref $maskP ne "ARRAY" ) 
	{
	    wit_die "Input mask is not an array";
	} 
	elsif ( not ref $maskP ) 
	{
	    if ( $maskP !~ /^\d+$/ )
	    {
		wit_die "Input mask is not a number";
	    }
	}
    }

    if ( not defined $newcolP ) { wit_die "No input column given"; }
    if ( not defined $newpos )  { wit_die "No input column number given"; }
    if ( not defined $tableP )  { wit_die "No input table given"; }

    # --------------------- COMPOSE COLUMN TO WRITE ----------------------

    my ( $row );
    my ( @insert_col );

    if ( ref $newcolP eq "ARRAY" )
    {
	@insert_col = @{ $newcolP };
    }
    elsif ( defined $maskP )
    {
	if ( ref $maskP eq "ARRAY" )
	{
	    for ( $row = 0; $row < @{ $maskP }; $row++ )
	    {
		if ( $maskP->[$row] ) 
		{
		    $insert_col[$row] = $newcolP;
		}
	    }
	}
	elsif ( not ref $maskP )
	{
	    for ( $row = 0; $row < $maskP; $row++ )
	    {
		push @insert_col, $newcolP;
	    }
	}
    }
    else 
    {
	for ( $row = 0; $row < @{ $tableP }; $row++ )
	{
	    push @insert_col, $newcolP;
	}
    }

    # -------------------- WRITE THE COLUMN --------------------------

    my $attrib;
    my $elem;
    my %attribs; 

    for ( $row = 0; $row < @insert_col; $row++ )
    {
	if ( ref $insert_col[$row] eq "HASH" )
	{
	    %attribs = %{ dclone $insert_col[$row] };

	    if ( ref $tableP->[$row] eq "ARRAY" ) 
	    {
		foreach $attrib ( keys %attribs ) 
		{
		    $tableP->[$row]->[$newpos]->{"$attrib"} = $attribs{"$attrib"}; 
		}
	    }
	    else
	    {
		# if the row is not an array reference, take whatever else is there
		# and make it column 0,

		$elem = $tableP->[$row];
		undef $tableP->[$row];
		$tableP->[$row]->[0] = $elem;

		foreach $attrib ( keys %attribs ) 
		{
		    $tableP->[$row]->[$newpos]->{"$attrib"} = $attribs{"$attrib"};
		}
	    } 
	}
    }
}


sub set_attrib_row { 

# Writes the values of @{ $newrowP } to the row in @{ $tableP } 
# at position $newpos. If the row does not exist, then it will.
# If $newrowP is a reference to an array, that array will replace 
# what was there before. If it is anything else, then this will be
# written to all columns of the $newpos row; however, if $maskP 
# is given as a reference to a list of 0 or 1's, then this mask 
# is used to decide which columns are set. If given as a number,
# then all columns are set. The routine works directly with the 
# incoming data, there is no return value; if you dont like this,
# supply a copy of the table instead. 

    use strict;
    use Storable qw( dclone );

    my ( $newpos, $tableP, $newrowP, $maskP ) = @_;

    # ----------------------- CHECK ARGUMENTS ----------------------------
    
    if ( defined $maskP ) 
    {
	if ( ref $maskP and ref $maskP ne "ARRAY" ) 
	{
	    wit_die "Input mask is not an array";
	} 
	elsif ( not ref $maskP ) 
	{
	    if ( $maskP !~ /^\d+$/ )
	    {
		wit_die "Input mask is not a number";
	    }
	}
    }

    if ( not defined $newrowP ) { wit_die "No input row given"; }
    if ( not defined $newpos )  { wit_die "No input row number given"; }
    if ( not defined $tableP )  { wit_die "No input table given"; }

    # ---------------------- FIND MAX COLUMN INDEX -----------------------

    my $tab_row;

    my $maxcol = 1;

    foreach $tab_row ( @{ $tableP } )
    {
	if ( ref $tab_row eq "ARRAY" and @{ $tab_row } > $maxcol )
	{
	    $maxcol = @{ $tab_row };
	}
    }

    # --------------------- COMPOSE ROW TO WRITE -------------------------

    my ( $col );
    my ( @insert_row );

    if ( ref $newrowP eq "ARRAY" )
    {
	@insert_row = @{ $newrowP };
    }
    elsif ( defined $maskP )
    {
	if ( ref $maskP eq "ARRAY" )
	{
	    for ( $col = 0; $col < @{ $maskP }; $col++ )
	    {
		$insert_row[$col] = $newrowP if $maskP->[$col];
	    }
	}
	elsif ( not ref $maskP )
	{
	    for ( $col = 0; $col < $maskP; $col++ )
	    {
		push @insert_row, $newrowP;
	    }
	}
    }
    else 
    {
	for ( $col = 0; $col < $maxcol; $col++ )
	{
	    push @insert_row, $newrowP;
	}
    }

    # ---------------------- OVERWRITE EXISTING ROW ----------------------

    my ( $elem );
    my ( %attribs, $attrib );

    if ( ref $newrowP eq "ARRAY" )
    {
	$tableP->[$newpos] = dclone $newrowP;
    }
    else 
    {
	# if the row is not an array reference, take whatever else is there
	# and make it column 0,

	if ( ref $tableP->[$newpos] ne "ARRAY" )
	{
	    $elem = $tableP->[$newpos];
	    undef $tableP->[$newpos];
	    $tableP->[$newpos]->[0] = $elem;
	}

	for ( $col = 0; $col < @insert_row; $col++ )
	{
	    if ( ref $insert_row[$col] eq "HASH" )
	    {
		%attribs = %{ dclone $insert_row[$col] };

		foreach $attrib ( keys %attribs )
		{
		    $tableP->[$newpos]->[$col]->{"$attrib"} = $attribs{"$attrib"};
		}
	    }
	}
    }
}

sub set_table_col {

# Writes the values of @{ $newcolP } to the column in @{ $tableP } 
# at position $newpos. If the column does not exist, then it will.
# If $newcolP is a reference to an array, that array will replace 
# what was there before. If it is anything else, then this will be
# written to all rows of the $newpos column; however, if $maskP 
# is given as a reference to a list of 0 or 1's, then this mask 
# is used to decide which rows are set. If given as a number, then 
# all rows are set. The routine works directly with the incoming 
# data, there is no return value; if you dont like this, supply a 
# copy of the table instead. 

    use strict;
    use Storable qw( dclone );

    my ( $newpos, $tableP, $newcolP, $maskP ) = @_;

    # ----------------------- CHECK ARGUMENTS ----------------------------
    
    if ( defined $maskP ) 
    {
	if ( ref $maskP and ref $maskP ne "ARRAY" ) 
	{
	    wit_die "Input mask is not an array";
	} 
	elsif ( not ref $maskP ) 
	{
	    if ( $maskP !~ /^\d+$/ )
	    {
		wit_die "Input mask is not a number";
	    }
	}
    }

    if ( not defined $newcolP ) { wit_die "No input column given"; }
    if ( not defined $newpos )  { wit_die "No input column number given"; }
    if ( not defined $tableP )  { wit_die "No input table given"; }

    # --------------------- COMPOSE COLUMN TO WRITE ----------------------

    my ( $row );
    my ( @insert_col );

    if ( ref $newcolP eq "ARRAY" )
    {
	@insert_col = @{ $newcolP };
    }
    elsif ( defined $maskP )
    {
	if ( ref $maskP eq "ARRAY" )
	{
	    for ( $row = 0; $row < @{ $maskP }; $row++ )
	    {
		if ( $maskP->[$row] ) 
		{
		    $insert_col[$row] = $newcolP;
		}
	    }
	}
	elsif ( not ref $maskP )
	{
	    for ( $row = 0; $row < $maskP; $row++ )
	    {
		push @insert_col, $newcolP;
	    }
	}
    }
    else 
    {
	for ( $row = 0; $row < @{ $tableP }; $row++ )
	{
	    push @insert_col, $newcolP;
	}
    }

    # -------------------- WRITE THE COLUMN --------------------------

    my $tab_row;
    my $elem;

    for ( $row = 0; $row < @insert_col; $row++ )
    {
	if ( defined $insert_col[$row] )
	{
	    if ( ref $insert_col[$row] )
	    {
		$insert_col[$row] = dclone $insert_col[$row];
	    }

	    $tab_row = $tableP->[$row];

	    if ( ref $tab_row ne "ARRAY" )
	    {
		$elem = $tab_row;
		undef $tab_row;
		$tab_row->[0] = $elem;
	    }

	    $tab_row->[$newpos] = $insert_col[$row];

	    $tableP->[$row] = $tab_row;
	}
    }
}

sub set_table_row { 

# Writes the values of @{ $newrowP } to the row in @{ $tableP } 
# at position $newpos. If the row does not exist, then it will.
# If $newrowP is a reference to an array, that array will replace 
# what was there before. If it is anything else, then this will be
# written to all columns of the $newpos row; however, if $maskP 
# is given as a reference to a list of 0 or 1's, then this mask 
# is used to decide which columns are set. If given as a number,
# then all columns are set. The routine works directly with the 
# incoming data, there is no return value; if you dont like this,
# supply a copy of the table instead. 

    use strict;
    use Storable qw( dclone );

    my ( $newpos, $tableP, $newrowP, $maskP ) = @_;

    # ----------------------- CHECK ARGUMENTS ----------------------------
    
    if ( defined $maskP ) 
    {
	if ( ref $maskP and ref $maskP ne "ARRAY" ) 
	{
	    wit_die "Input mask is not an array";
	} 
	elsif ( not ref $maskP ) 
	{
	    if ( $maskP !~ /^\d+$/ )
	    {
		wit_die "Input mask is not a number";
	    }
	}
    }

    if ( not defined $newrowP ) { wit_die "No input row given"; }
    if ( not defined $newpos )  { wit_die "No input row number given"; }
    if ( not defined $tableP )  { wit_die "No input table given"; }

    # ---------------------- FIND MAX COLUMN INDEX -----------------------

    my $tab_row;
    my $maxcol = 1;

    foreach $tab_row ( @{ $tableP } )
    {
	if ( ref $tab_row eq "ARRAY" and @{ $tab_row } > $maxcol )
	{
	    $maxcol = @{ $tab_row };
	}
    }

    # --------------------- COMPOSE ROW TO WRITE -------------------------

    my ( $col );
    my ( @insert_row );

    if ( ref $newrowP eq "ARRAY" )
    {
	@insert_row = @{ $newrowP };
    }
    elsif ( defined $maskP )
    {
	if ( ref $maskP eq "ARRAY" )
	{
	    for ( $col = 0; $col < @{ $maskP }; $col++ )
	    {
		$insert_row[$col] = $newrowP if $maskP->[$col];
	    }
	}
	elsif ( not ref $maskP )
	{
	    for ( $col = 0; $col < $maskP; $col++ )
	    {
		push @insert_row, $newrowP;
	    }
	}
    }
    else 
    {
	for ( $col = 0; $col < $maxcol; $col++ )
	{
	    push @insert_row, $newrowP;
	}
    }

    # ---------------------- OVERWRITE EXISTING ROW ----------------------

    my ( $elem );

    if ( ref $newrowP eq "ARRAY" )
    {
	$tableP->[$newpos] = dclone $newrowP;
    }
    else 
    {
	# if the row is not an array reference, take whatever else is there
	# and make it column 0,

	if ( ref $tableP->[$newpos] ne "ARRAY" )
	{
	    $elem = $tableP->[$newpos];
	    undef $tableP->[$newpos];
	    $tableP->[$newpos]->[0] = $elem;
	}

	for ( $col = 0; $col < @insert_row; $col++ )
	{
	    if ( defined $insert_row[$col] )
	    {
		if ( ref $insert_row[$col] )
		{
		    $tableP->[$newpos]->[$col] = dclone $insert_row[$col];
		}
		else
		{
		    $tableP->[$newpos]->[$col] = $insert_row[$col];
		}
	    }
	}
    }
}

sub explore_possible_connections {

    my ($to_show,$org,$user,$title,$asserted) = @_;

    my ($row_hdrs,$table,$x,$rid,$funcL,@pws,$pws);
    my (%a_pw,$pw,@pws1,$func_ref,$i,$name, @output);

    $row_hdrs = [ "ORF ID", "Organism", "Function", "Pathway(s)" ];
    $table    = [];

#    use Data::Dumper;
#    push @output, "<PRE>";
#    push @output, Data::Dumper->Dump([ $to_show ], [ "*to_show"]);

    foreach $x (@$to_show)
    {
	($rid,$funcL) = @$x;
	@pws = ();
	undef %a_pw;

	foreach $x (@$funcL)
	{
	    $pws = &FR::function_to_pwsR($x);

#	    push @output, Data::Dumper->Dump([ $x, $pws ], [ "*x", "*pws"]);

	    foreach $pw (@$pws)
	    {
		if ($asserted)
		{
		    $a_pw{$pw} = 1 if &PW::asserted_pathwayP($org,$user,$pw);
		}
		else
		{
		    $a_pw{$pw} = 1 if not &PW::asserted_pathwayP($org,$user,$pw);
		}
	    }
	}
	@pws  = sort keys(%a_pw);
	@pws1 = ();

	foreach $pw (@pws)
	{
	    if (! &PW::suppressed_pathwayP($org,$user,$pw))
	    {
		push(@pws1,$pw);
	    }
	}
	
	@pws = @pws1;

	$funcL     = &ORF::assigned_functionsO($rid,$user);
	$func_ref  = &set_role_links($funcL,$org,$user);

	undef @pw_links;

	for ($i=0; ($i <= $#pws); $i++)
	{
	    $name = &PW::name_ofP($pws[$i]);

	    if (! $$name) { $name = \$pws[$i]; }

	    push @pw_links, qq (<A HREF="$cgi_url/pw.cgi)
                          . qq (?pw=$pws[$i]&org=$org&user=$user&where=show_diagram">)
                          . qq (<IMG SRC="$gifs_url/sqbulorg.gif" BORDER=0></A>)
                          . qq (&nbsp;<A HREF="$cgi_url/pw.cgi?pw=$pws[$i]&org=$org&user=$user">$$name</A>);
	}

        $table = [
		   [ &prot_link($rid,$user) ],
		   [ &genome_bioname($org) ],
		   [ $$func_ref ],
		   [ (join "<BR>", @pw_links) ."\n" ],
		 ];
		  
#	push @output, qq (<IMG SRC="$gifs_url/sqbulorg.gif" BORDER=0> shows diagrams.<BR>);

	push @output, @{ &html_table_default ( 
					      $row_hdrs, 
					      [], 
					      $table,
					      {
					       "RHEAD_WRAP"    => "nowrap",
					       "RHEAD_JUSTIFY" => "right",
					       "RHEAD_BGCOLOR" => "#777777",
					       "RHEAD_FGCOLOR" => "#FFFFFF",
					       "RHEAD_BOLD"    => 1,
					       "ELEMS_JUSTIFY" => "left",
					       "ELEMS_BGCOLOR" => "#CCCCCC",
 					       "ELEMS_PADDING" => 3,
					       "ELEMS_WRAP"    => "nowrap"
					      }
					     ) }, "<P>\n";
    }
    return @output;
}

sub get_medline_for_ec
{
    my ( $ec_num ) = @_;
    my ( @output, $output );
    
    @output = &get_html( $entrez_url, "POST"
			, [["db","m"], ["form","4"], ["field","ECNO"], ["term","$ec_num"]] );
    $output = join( "", @output );
    
    if ( $output )
    {
	return \$output;
    } else {
	return \"Content-type: text/html\n\n<H1> Sorry, no references found for E.C. $ec_num </H1>\n";
    }
}

sub get_prodom_analysis {
    my($prot) = @_;
    my(@output,$seq,$ac,$x);

    $prot = &Util::reduce_prot_id($prot);

    if ($prot =~ /^sp\|([OPQ][A-Z0-9]{5})/)
    {
	$ac = $1;
	@output = &get_html("http://protein.toulouse.inra.fr/cgi-bin/ReqProdomII.pl","POST",
			    [["prodom_release","p34.1"],
			     ["acc_seq0",$ac]]);
    }
    else
    {
        $seq = &ORF::prot_seq_ofO($prot);

        @output = &get_html("http://protein.toulouse.inra.fr/cgi-bin/NewBlastProdomII.pl","POST",
			[["typebd","multiple alignments"],
			 ["program","wu-blastp"],
			 ["matrice","BLOSUM62"],
			 ["expect","0.01"],
			 ["filtre","seg"],
			 ["nom_seq",$prot],
			 ["sequence",$$seq],
			 ["submit","Submit Query"]]);
    }

    if ($output[0] !~ /^Content/) 
    { 
	splice(@output,0,0,("Content-type: text/html\n","\n")); 
    }
    foreach $x (@output)
    {
	$x =~ s#"/#"http://protein.toulouse.inra.fr/#g;
    }
    return @output;
}

sub save_prodom_analysis {

    my($prot) = @_;
    my(@output,$i,$url,$seq,$x,$ac);

    $prot = &Util::reduce_id($prot);

    if (! -d "$pdom_dir") { mkdir("$pdom_dir",0777); }
    if (-s "$pdom_dir/$prot.gif") { return 1; }
    
    @output = ();

    if ((&ORF::typeO($prot) eq "rid") && ($x = &ORF::equivalent_sp($prot)))
    {
	&save_prodom_analysis($$x);

	system "cp \"$pdom_dir/$$x\.gif\" $pdom_dir/$prot.gif";

	return 1;
    }

    if ($prot =~ /^sp\|([OPQ][A-Z0-9]{5})/)
    {
	$ac = $1;
	@output = &get_html("http://protein.toulouse.inra.fr/cgi-bin/ReqProdomII.pl","POST",
			    [["prodom_release","p34.1"],
			     ["acc_seq0",$ac]]);
	for ($i=0; ($i <= $#output) && ($output[$i] !~ /BORDER/); $i++) {}
	if (($i <= $#output) && ($output[$i] =~ /\"([^"]+gif)\"\s+ISMAP\s+BORDER/))
        {
            $url = "http://protein.toulouse.inra.fr" . $1;
            @output = &get_html($url,"GET",[]);
        }
    }
    else
    {
        $seq = &ORF::prot_seq_ofO($prot);

        @output = &get_html("http://protein.toulouse.inra.fr/cgi-bin/NewBlastProdomII.pl","POST",
			[["matrice","BLOSUM62"],
                         ["typebd","multiple alignments"],
			 ["program","wu-blastp"],
			 ["expect","0.01"],
			 ["filtre","seg"],
			 ["nom_seq",$prot],
			 ["sequence",$$seq],
			 ["submit","Submit Query"]]);

        for ($i=0; ($i <= $#output) && ($output[$i] !~ /<A NAME\s*=\s*\"Graphic\"/); $i++) {}
        for ($i++; ($i <= $#output) && ($output[$i] !~ /BORDER/); $i++) {}

        if (($i <= $#output) && ($output[$i] =~ /\"([^"]+gif)\" BORDER/))
        {
            $url = $1;
            @output = &get_html($url,"GET",[]);
        }
    }

    if (@output > 0)
    {
	for ($i=0; ($i <= $#output) && ($output[$i] !~ /^GIF/); $i++) {}

	if ($i <= $#output)
	{
	    open(TMP,">$pdom_dir/$prot.gif") || return 0;

	    while ($i <= $#output)
	    {
		print TMP $output[$i];
		$i++;
	    }
	    close(TMP);
	}
    }
}

sub select_menu 
{ 
    my $name = shift;
    my $show = shift;
    my @menu = @_;

    my $item;
    my @output;

    @output = "\n<SELECT NAME=$name>";

    if ( @menu ) 
    {
	foreach $item ( @menu ) 
	{
	    if ( $item->[0] =~ /^$show$/i ) {
		push @output, qq (<OPTION VALUE="$item->[0]" SELECTED>$item->[1]);
	    } else {
		push @output, qq (<OPTION VALUE="$item->[0]">$item->[1]);
	    }
	}
    }
    else 
    {
	push @output, qq (<OPTION VALUE="">&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;);
    }

    push @output, "</SELECT>";

    return join "\n", @output;
}


sub topnavbar 
{
    my ( $name,      # Display name of the system, "WIT", "WIT-Pro", ... 
	 $cgi,       # CGI.pm object with parameters
	 $script,    # Name of cgi script 
	 $orgs       # Number of organisms (optional)
       ) = @_;
    
    my @orgids;
    my @output;

    my $argonne_mpw_url = "http://wit.mcs.anl.gov/MPW";
    my $argonne_emp_url = "http://wit.mcs.anl.gov/EMP";

    my ( $value, $param );
    my $user;
    my $user_str;
    my $url_user;
    my $curate_link;

    if ( not $name )
    {
	$name = "WIT";
    }

    $user = $cgi->param('user');

    use Env;

    my %wit_env = %{ &restore_wit_env ($conf_dir, $user) };

# Link if there is a user, 

    if ( defined $user and $user =~ /\w/ ) 
    {
#	$user_str = ", User $user";
	$url_user = &url_encode ($user);
	$curate_link = qq (<A HREF="$cgi_url/org.cgi?user=$url_user">Curate Models</A>);
    } 
    else 
    {
	$user     = "";
#	$user_str = "";
	$url_user = "";
	$curate_link = "Curate Models";
    }

    $emp_link = qq (<A HREF="$WIT2_Config::emp_url">Query Enzymes</A>);

# Link if EMP is there,
#
#    if ( $wit_env{'EMP'}{'BASE_URL'} eq "$argonne_emp_url" )
#    {
#	$emp_link = qq (<A HREF="$argonne_emp_url">Query Enzymes</A>);
#    }
#    else 
#    {
#	if ( defined $emp_url and -d $emp_dir )
#	{
#	    $emp_link = qq (<A HREF="$wit_env{'EMP'}{'BASE_URL'}">Query Enzymes</A>);
#	}
#	else 
#	{
#	    $emp_link = "Query Enzymes";
#	}
#    } 

     $mpw_link = qq (<A HREF="$WIT2_Config::mpw_url">Query Pathways);

# Link if MPW is there, 

#    if ( $wit_env{'MPW'}{'BASE_URL'} eq "$argonne_mpw_url" )
#    {
#	$mpw_link = qq (<A HREF="$argonne_mpw_url">Query Pathways</A>);
#    }
#    else 
#    {
#	if ( defined $mpw_url and -d $mpw_dir )
#	{
#	    $mpw_link = qq (<A HREF="$wit_env{'MPW'}{'BASE_URL'}">Query Pathways</A>);
#	}
#	else 
#	{
#	    $mpw_link = "Query Pathways";
#	}
#    } 

    # With 200 organisms, this takes 0.1 seconds, so its fine to run it 
    # with every script.. 
    
    if ( not $orgs ) 
    {
	if ( opendir ORGDIR, "$orgs_dir" )
	{
	    @orgids = grep { /^[A-Z][A-Z]$/ and -d "$orgs_dir/$_" } readdir ORGDIR;
	    closedir ORGDIR;
	    $orgs = @orgids; 
	} 
	else 
	{
	    print &system_error (
				 "Read Error",
				 "Could not read the directory $orgs_dir",
				 "",
				 "HEAD"
				);
	}
    }

    push @output, qq (<FORM METHOD="POST" ACTION="$script">);

    foreach $param ( $cgi->param )
    {
	$value = $cgi->param($param);

	if ( $param ne "user" )
	{
	    push @output, qq (<INPUT TYPE="hidden" NAME="$param" VALUE="$value">\n);
	}
    }

    push @output, qq (
<P>
<TABLE BORDER=4 BGCOLOR="#888888" WIDTH=100% CELLPADDING=0 CELLSPACING=0>
<TR><TD WIDTH="100%">

    <TABLE BGCOLOR="#555555" BORDER=0 WIDTH=100% CELLSPACING=0 CELLPADDING=0>
    <TR>
       <TD WIDTH="50%" ALIGN="left" VALIGN="middle" NOWRAP>
           <TABLE BGCOLOR="#555555" CELLSPACING=0 CELLPADDING=0>
              <TR>
              <TD><FONT FACE="Arial,Helvetica" COLOR="#FFFFFF">
                  <STRONG><BIG><BIG><BIG>&nbsp;&nbsp;$name&nbsp;&nbsp;</BIG></BIG></BIG></STRONG></FONT>
	          <FONT FACE="Arial,Helvetica" COLOR="#FFFFFF">$orgs Genomes</FONT>
              </TD>
              </TR>
           </TABLE>
       </TD>
       <TD ALIGN="center" VALIGN="middle" NOWRAP>
	   <TABLE BGCOLOR="#FFDD77" CELLSPACING=0 CELLPADDING=5 BORDER=2>
              <TR>
              <TD><FONT FACE="Arial,Helvetica"><STRONG>
                 <A HREF="$cgi_url/index.cgi?user=$url_user">Home</A></STRONG></FONT></TD>
              <TD><FONT FACE="Arial,Helvetica"><STRONG>
                 <A HREF="$cgi_url/faq.cgi?user=$url_user">FAQ</A></STRONG></FONT></TD>
              <TD><FONT FACE="Arial,Helvetica"><STRONG>
                 <A HREF="$cgi_url/credits.cgi?user=$url_user">Credits</A></STRONG></FONT></TD>
              </TR>
           </TABLE>
        </TD>
       <TD ALIGN="right" VALIGN="bottom" NOWRAP>
           <TABLE BGCOLOR="#555555" CELLSPACING=0 CELLPADDING=0>
              <TR><TD BGCOLOR="#555555"><INPUT TYPE="text" NAME="user" VALUE="$user" SIZE="14"></TD>
  	          <TD BGCOLOR="#555555"><FONT COLOR="#FFFFFF">
                      <INPUT TYPE="submit" VALUE="Set User">&nbsp;</FONT>
                  </TD>
              </TR>
           </TABLE>
       </TD>
     </TR>
     </TABLE>

<TABLE BORDER=0 WIDTH=100% CELLPADDING=2 CELLSPACING=0>
<TR><TD WIDTH="100%">

    <TABLE BGCOLOR="#DDDDDD" BORDER=1 WIDTH=100% CELLSPACING=2 CELLPADDING=3>
    <TR ALIGN="middle">
      <TD><A HREF="$cgi_url/func_ov.cgi?user=$url_user&compare=&org&max_sz=50">General Overview</A></TD>
      <TD><A HREF="$cgi_url/examine.cgi?user=$url_user">View Models</A></TD>
      <TD>$curate_link</TD>
      <TD><A HREF="$cgi_url/configure.cgi?user=$url_user">User Settings</A></TD>
      <TD>$mpw_link</TD>
    </TR>

    <TR ALIGN="middle">
      <TD><A HREF="$cgi_url/search.cgi?user=$url_user">General Search</A></TD>
      <TD><A HREF="$cgi_url/sim_search.cgi?user=$url_user">Similarity Search</A></TD>
      <TD><A HREF="$cgi_url/clusters.cgi?user=$url_user">Ortholog Clusters</A></TD>
      <TD><A HREF="$cgi_url/operons.cgi?user=$url_user">Operons</A></TD>
      <TD>$emp_link</TD>
    </TR>
    </TABLE>

</TD></TR></TABLE>

</TD></TR></TABLE>
);
    push @output, "</FORM>\n";

    return @output;
}

sub url_decode {
    my($arg) = @_;
    
    $arg =~ s/\%28/\(/g;
    $arg =~ s/\%29/\)/g;
    $arg =~ s/\%27/\'/g;
    $arg =~ s/\%60/\`/g;
    $arg =~ s/\%2B/\+/g;
    $arg =~ s/\%3A/\:/g;
    $arg =~ s/\%09/\t/g;
    $arg =~ s/\%20/ /g;
    $arg =~ s/\%22/\"/g;
    $arg =~ s/\%3E/>/g;
    $arg =~ s/\%5C/\\/g;
    $arg =~ s/\%5E/\^/g;
    $arg =~ s/\%7B/\{/g;
    $arg =~ s/\%7D/\}/g;
    $arg =~ s/\%7C/\|/g;
    $arg =~ s/%5B/\[/g;
    $arg =~ s/%5D/\]/g;
    $arg =~ s/\%25/\%/g;
    return $arg;
}

sub url_encode {
    my($arg) = @_;
    
    defined($arg) || confess "bad arg to url_encode";

    $arg =~ s/\%/%25/g;
    $arg =~ s/\(/%28/g;
    $arg =~ s/\)/%29/g;
    $arg =~ s/\'/%27/g;
    $arg =~ s/\`/%60/g;
    $arg =~ s/\+/%2B/g;
    $arg =~ s/\:/%3A/g;
    $arg =~ s/\t/%09/g;
    $arg =~ s/ /%20/g;
    $arg =~ s/\"/%22/g;
    $arg =~ s/>/%3E/g;
    $arg =~ s/\\/%5C/g;
    $arg =~ s/\^/%5E/g;
    $arg =~ s/\{/%7B/g;
    $arg =~ s/\}/%7D/g;
    $arg =~ s/\|/%7C/g;
    $arg =~ s/\[/%5B/g;
    $arg =~ s/\]/%5D/g;
#    $arg =~ s/\&/%26/g;
    $arg =~ s/\//%2F/g;
    $arg =~ s/\=/%3D/g;
    $arg =~ s/\n/%0A/g;
    $arg =~ s/\r/%0D/g;
    return $arg;
}

1;
