package ORF;			# -*- perl -*-

use Feature;
use HashDB;
use Annotation;
use Util;
use Carp;
use Html;
use FR;
use Messages;

use Exporter ();

@ISA    = ( Exporter );

@EXPORT = qw(
	     aliasesO
	     all_orfsO
	     all_assigned_functionsO
	     assign_functionO
	     assigned_func_and_conf
	     assigned_functionsO
	     dna_seq_ofO
	     downstreamO
	     equivalent_sp
	     extended_assigned_functionsO
             func_ofO
	     genome_ofO
	     get_cannonicalO
	     get_locationO
	     lengthO
	     org_ofO
	     prot_seq_ofO
	     principal_synonym_ofO
	     printable_genome_ofO
	     relevant_sp_with_structures
	     set_cannonicalO
	     set_locationO
	     solid_assignment
	     swiss_prot_entry
	     synonyms_ofO
	     typeO
	     update_assignmentsO
	     upstreamO
);

my ( %org_dna_loaded,  %dna_seq  );
my ( %org_prot_loaded, %prot_seq );

sub relevant_sp_with_structures {
    my($rid) = @_;
    my($sps,@sps);

    if (! $opened_relevant_sps_with_structures)
    {
	($opened_relevant_sps_with_structures = &opendbH("rid_to_sps_with_pdb"))
	    || wit_die "could not open rid_to_sps_with_pdb";
    }
    
    if ($sps = &accessdbH($opened_relevant_sps_with_structures,$rid))
    {
	@sps = split(/\t/,$sps);
	return \@sps;
    }
    return "";
}

sub genome_ofO {
    my($prot) = @_;

    if ( $prot =~ /^R([A-Z]{2})\d{5}$/       )  { return $1;   }
    if ( $prot =~ /sp\|[OPQ]\d[0-9A-Z]{3}\d/ )  { return "SP"; }
    if ( $prot =~ /tr\|\w+/                  )  { return "SP"; }
    if ( $prot =~ /tn\|\w+/                  )  { return "SP"; }
    if ( $prot =~ /gi\|\d+/                  )  { return "SP"; }
    
    return "";
}

sub aliasesO {
    my( $id ) = @_;

    if (&typeO($id) eq "rid")
    {
	return &aliasesF( $id );
    }
    return "";
}

sub all_orfsO {
    my($org) = @_;

    return &all_featuresF($org,"orf");
}

sub assigned_functionsO {
    my($prot,$user) = @_;
    my($type,$func_conf,@funcs,$func);

    @funcs = ();
    $type = &typeO($prot);
#   print STDERR "ORF::assigned_functionsO --- $prot is an $type\n";
    if ( $type eq "rid" )
    {
	if ($user && defined($func_conf = &user_functions($user,$prot)))
	{
	    return [$func_conf->[0]];
	}

	if (! $opened_assigned_functions)
	{
	    ($opened_assigned_functions = &opendbH("assigned_functions"))
		|| wit_die "could not open assigned_functions";
	}

	if ($func_conf = &accessdbH($opened_assigned_functions,$prot))
	{
	    ($func) = split(/\t/,$func_conf);
	    return [$func];
	}
	else
	{
	    return [];
	}
    }
    elsif ( ($type eq "sp" ) || ($type eq "gi") || ($type eq "tr") || ($type eq "tn") )
    {
        $prot =~ s/^(sp|tr|tn|gi)\|//;
	if (! $opened_assigned_functions_sp)
	{
	    ($opened_assigned_functions_sp = &opendbH("sp_func"))
		|| wit_die "could not open sp_func";
	}

	if ($funcs = &accessdbH($opened_assigned_functions_sp,$prot))
	{
	    ($func) = split(/\t/,$funcs);
	    return [$func];
	}
	else
	{
	    return [];
	}
    }
}

sub solid_assignment {
    my($func) = @_;

    return (($func !~ /hypothetical/i) &&
#	    ($func !~ /putative/i) &&
#	    ($func !~ /predicted/i) &&
	    ($func !~ /similar/i) &&
	    ($func !~ /unknown/i) &&
	    ($func !~ /cosmid/) &&
	    ($func !~ /membrane/) &&
	    ($func !~ /homolog/i) &&
	    ($func !~ /family/i) &&
	    ($func !~ /probable/i));
}

sub func_ofO {
    my($rid,$user) = @_;
    my($funcL,$func);

    $funcL  = &assigned_functionsO($rid,$user);
    $func   = join("; ",@$funcL);
    if (! $func) { $func = ""; }
    return $func;
}

sub assigned_func_and_conf {
    my($prot,$user) = @_;
    my($type,$func_conf,@funcs,$func,$conf);

    $type = &typeO($prot);

    if ( $type eq "rid" )
    {
	if ($user && defined($func_conf = &user_functions($user,$prot)))
	{
	    ($func,$conf) = @$func_conf;
            $user_func_of{$prot} = [$func,$conf]
	}
	else
	{
	    if (! $opened_assigned_functions)
	    {
		($opened_assigned_functions = &opendbH("assigned_functions"))
		    || wit_die "could not open assigned_functions";
	    }

	    if ($_ = &accessdbH($opened_assigned_functions,$prot))
	    {
		($func,$conf) = split(/\t/,$_);
	    
	    }
	    else
	    {
		$func = $conf = "";
	    }
	}
    }
    else
    {
	$func = &func_ofO($prot,$user);
	$conf = "";
    }

    defined($func) || ($func = "");
    defined($conf) || ($conf = "");
    return [$func,$conf];
}


sub dna_seq_ofO {
    my($id,$upstream,$downstream) = @_;

    return &dna_seq_ofF($id,$upstream,$downstream);
}

sub downstreamO {
    my($id,$downstream) = @_;

    return &downstreamF($id,$downstream);
}

sub lengthO {
    my($prot) = @_;
    my($seq);

    $seq = &prot_seq_ofO($prot);
    return length($$seq);
}

sub get_locationO {
    my($rid) = @_;

    if ( &typeO($rid) eq "rid" )
    {
	return &get_locationF($rid);
    } else {
	return "";
    }
}

sub get_cannonicalO {
    my($rid) = @_;

    return &get_cannonicalF($rid);
}

sub printable_genome_ofO {
    my($rid) = @_;
    my($genomes)   = &genomes();
    my($i,$cd);

    $cd = substr($rid,1,2);
    for ($i=0; ($i < $#{$genomes}) && ($genomes->[$i] !~ /^$cd/); $i++) {}
    if ($i <= $#{$genomes})
    {
	return split(/\t/,$genomes->[$i]);
    }
    else
    {
	return "";
    }
}



sub prot_seq_ofO
{
    my( $id ) = @_;
    my( $ref, $principal, $key, $seek, $fh, $old_eor, @lines, $seq );
    
    if (! $opened_nr)
    {
	open(NR,"<$WIT2_Config::global/nr") &&
	    ($ref = &opendbH("nr_index")) 
	    || wit_die "could not open nr_index";
	$opened_nr = [$ref,\*NR];
    }
    
    $principal = &reduce_id( $id );
    if ( &typeO($principal) ne "rid" )
    {
	$principal =  &principal_synonym_ofO($id);
    }
    
    $key =  $principal;
    $key =~ s/^(sp|tr|tn|gi)\|//;
    if (defined(($seek = &accessdbH($opened_nr->[0],$key))))
    {
	$fh = $opened_nr->[1];
	seek($fh,$seek,0);
	$old_eor = $/;
	$/ = "\n";
	$_ = <$fh>;
	@lines  = ();
	$seq = "";
	while ( defined($_ = <$fh>) && ($_ !~ /^>/) )
	{
	    $_ =~ s/\s//g;
	    push( @lines, $_ );
	}
	$seq = join( "", @lines );
	$/ = $old_eor;
	return \$seq;    
    } 
    else 
    { 
	$failed_seeks++;
	if ( $failed_seeks <= 100 )
	{
	    my ( $synonyms );
	    $synonyms = &synonyms_ofO( $principal );
	    $synonyms = join( ", ", @$synonyms );
	    warn "ORF::prot_seq_ofO --- seek failed for $id --> $principal ($synonyms)";
#	    confess "aborted";
	    if ( $failed_seeks == 100 )
	    {
		print STDERR "ORF::prot_seq_ofO --- more than 100 failed seeks; going into silent mode ...\n";
	    }
	}
	return \"";
    }
}



sub prot_seq_of_gi 
{
    my($gi_ac) = @_;
    my($ref,$seek,$fh,$old_eor,@lines,$gi_seq);
    
    if (! $opened_nr)
    {
	open(NR,"<$WIT2_Config::global/nr") &&
	    ($ref = &opendbH("nr_index")) 
	    || wit_die "could not open nr_index";
	$opened_nr = [$ref,\*NR];
    }
    
    $gi_ac =~ s/^(sp|tr|tn|gi)\|//;
    if (defined(($seek = &accessdbH($opened_nr->[0],$gi_ac))))
    {
	$fh = $opened_nr->[1];
	seek($fh,$seek,0);
	$old_eor = $/;
	$/ = "\n";
	$_ = <$fh>;
	@lines  = ();
	$gi_seq = "";
	while ( defined($_ = <$fh>) && ($_ !~ /^>/) )
	{
	    $_ =~ s/\s//g;
	    push( @lines, $_ );
	}
	$gi_seq = join( "", @lines );
	$/ = $old_eor;
	return \$gi_seq;    
    } 
    else 
    { 
	$failed_gi_seeks++;
	if ( $failed_gi_seeks <= 100 )
	{
	    print STDERR "ORF::prot_seq_of_gi --- seek failed for gi\|$gi_ac\n";
	    if ( $failed_gi_seeks == 100 )
	    {
		print STDERR "ORF::prot_seq_of_gi --- more than 100 failed seeks; going into silent mode ....\n";
	    }
	}
	return \"";
    }
}


sub prot_seq_of_sp {
    my($spac) = @_;
    my($sp_entry,@lines,$seq);

    $spac =~ s/^sp\|([OPQ]\d[0-9A-Z]{3}\d).*$/$1/;
    $spac =~ s/^tr\|([OPQ]\d[0-9A-Z]{3}\d).*$/$1/;
    $spac =~ s/^tn\|([A-Z0-9]o+).*$/$1/;

    $sp_entry = &swiss_prot_entry($spac);

    if (($x = index($$sp_entry,"\nSQ   SEQUENCE")) > 0)
    {
	chomp $$sp_entry;
	substr($$sp_entry,0,$x+1) = "";
	@lines = split(/\n/,$$sp_entry);
	splice(@lines,0,1);
	splice(@lines,$#lines,1);
	$seq = join("",@lines);
	$seq =~ s/\s//g;
	return \$seq;
    }
    return \"";
}

sub old_prot_seq_ofO 
{
    my ( $prot ) = @_;
    my ( $type, $x );
    
    $type = &typeO($prot);
    if ($type eq "rid")
    {
	return &prot_seq_of_rid($prot);
    }
    elsif ( ($type eq "sp") || ($type eq "tr") || ($type eq "tn") || ($type eq "gi") )
    {
	return &prot_seq_of_gi($prot);
    }
    else
    {
	return \"";
    }
}



sub synonyms_ofO
{
    my ( $prot ) = @_;
    my ( $principal, $synonyms, @synonyms );
    
    $principal = &principal_synonym_ofO( $prot );
    unless (defined($principal) && $principal)
    {
	warn "ORF::synonyms_ofO --- failed to get principal synonym of $prot; returning [ $prot ]";
	return [ $prot ];
    }
    
    if (! $opened_synonymous_ids)
    {
	($opened_synonymous_ids = &opendbH("synonymous_ids"))
	         || confess "could not open synonymous_ids";
    }
    
    $principal =~ s/^(sp|tr|tn|gi)\|//;
    $synonyms  =  &accessdbH( $opened_synonymous_ids, $principal );
    @synonyms  =  split( /\t/, $synonyms );
    
    return [ @synonyms ];
}

sub principal_synonym_ofO
{
    my ( $prot ) = @_;
    my ( $reduced_id, $principal );
    
    $reduced_id = &reduce_id( $prot );
    
    if (! $opened_principal_synonym)
    {
	($opened_principal_synonym = &opendbH("principal_synonym"))
	            || confess "could not open principal_synonym";
    }
    
    $reduced_id =~ s/^(sp|tr|tn|gi)\|//;
    unless ( defined($principal = &accessdbH( $opened_principal_synonym, $reduced_id)) && $principal )
    {
	$principal = $reduced_id;
	warn "ORF::principal_synonym_ofO --- no principal synonym for $prot";
    }
    
    return $principal;
}



sub swiss_prot_entry {
    my($sp_ac) = @_;
    my($ref,$fh,$old_eor,$sp_entry,$seek);

    if (! $opened_sp_dat)
    {
	open(SP,"<$WIT2_Config::global/sp.dat") &&
	    ($ref = &opendbH("swiss_prot_index")) 
	    || wit_die "could not open swiss_prot_index";
	$opened_sp_dat = [$ref,\*SP];
    }
    
#   print STDERR "$sp_ac --> ";
    $sp_ac =~ s/^(sp|tr|tn)\|//;
#   print STDERR "$sp_ac\n";
    
    if (defined(($seek = &accessdbH($opened_sp_dat->[0],$sp_ac))))
    {
	$fh = $opened_sp_dat->[1];
	seek($fh,$seek,0);
	
	$old_eor = $/;
	$/ = "\n//\n";
	$sp_entry = <$fh>;
	$/ = $old_eor;
	return \$sp_entry;
    }
    else
    {
	$failed_sp_seeks++;
	if ( $failed_sp_seeks <= 100 )
	{
	    print STDERR "ORF::swiss_prot_entry --- seek failed for sp\|$sp_ac\n";
	    if ( $failed_sp_seeks == 100 )
	    {
		print STDERR "ORF::swiss_prot_entry --- more than 100 failed seeks; going into silent mode ...n";
	    }
	}
	return \"";
    }
}



sub prot_seq_of_rid 
{
    my ( $rid ) = @_;
#++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# returns sequence from fasta file $organisms/Org/Features/orf/fasta
#
# as with contigs, the first time it occurs for a genome, load the
# whole file into an associative array of pointers to the sequence.
#------------------------------------------------------------------------
    my ( @lines, $orgID, $tmp, @tmp, $seq, $numseq, $x );
#========================================================================

    $orgID = substr( $rid, 1, 2 );
    unless ( $tmp = $org_prot_loaded{ $orgID } )
    {
	if (-s "$WIT2_Config::organisms/$orgID/Features/orf/fasta" )
	{
	    $numseq = 0;
	    $org_prot_loaded{ $orgID } = 1;
	    if ( open( ORFS, "<$WIT2_Config::organisms/$orgID/Features/orf/fasta" ) )
	    {
                
		$_ = <ORFS>;
		while (defined($_) && $_ && ($_ =~ /^>(\S+)/ ))
		{
		    $numseq ++ ;
		    $tmp = $1;
                    
		    @lines = ();
		    while ( defined($_ = <ORFS>) && ($_ !~ /^>/) )
		    {
			$_ =~ s/\s//g;
			push( @lines, $_ );
		    }
		    $seq = join( "", @lines );
		    $x   = [$seq];
		    $prot_seq{ $tmp } = \($x->[0]);
		}
		close( ORFS );
	    } else { print STDERR "open failed for org $orgID\n"; }
	} else { print STDERR "No FASTA ORFs file for org $orgID\n"; }
    }

    if ( $tmp = $prot_seq{ $rid } )
    {
	return $tmp;
    } else {
	return \"";
    }
}



sub user_functions {
    my($user,$rid) = @_;
    my($org,$x,$func,$rid1,$func,$funcs,$conf,@assignments);

    
    $org = substr($rid,1,2);
    if ((! $loaded_user_assignments{"$org\t$user"}) &&
	(-s "$WIT2_Config::organisms/$org/UserModels/$user/assigned_functions"))
    {
	if (! $loaded_user_assignments{"$org\t$user"})
	{
#	    print STDERR "loading $org/$user\n";
	    $loaded_user_assignments{"$org\t$user"} = 1;
	    @assignments = grep(/^R/,`cat $WIT2_Config::organisms/$org/UserModels/$user/assigned_functions`);
	    chop @assignments;
	    foreach $x (@assignments)
	    {
		($rid1,$func,$conf) = split(/\t/,$x);
		if (! defined($conf)) { $conf = ""; }
		if (! defined($func)) { $func = ""; }

		if ($func)
		{
		    $user_func_of{"$rid1\t$user"} = [$func,$conf];
		}
		else
		{
#		    print STDERR "$rid1 is deleted\n";
		    delete $user_func_of{"$rid1\t$user"};
		}
	    }
	}
    }
    return $user_func_of{"$rid\t$user"};
}

sub assign_functionO {
    my($rid,$user,$func,$no_annotations) = @_;

    defined($no_annotations) || ($no_annotations = 0);

    $org = substr($rid,1,2);
    (-d "$WIT2_Config::organisms/$org") || return 0;

    if (&master_user($org,$user))
    {
	&assign_function_by_master($rid,$user,$func,$no_annotations);
    }
    else
    {
	&assign_function_by_non_master($rid,$user,$func);
    }
}

sub assign_function_by_master {
    my($rid,$user,$function,$no_annotations) = @_;
    my(@old,@keep,@annotations,$func,@pitch,$x,$y,$rid1,$i,@rids);
    my($just_function);

    $just_function = $function;
    $just_function =~ s/\t.*$//;
    if (! $opened_assigned_functions)
    {
	($opened_assigned_functions = &opendbH("assigned_functions"))
	    || wit_die "could not open assigned_functions";
    }

    if (! $opened_function_to_rids)
    {
	($opened_function_to_rids = &opendbH("function_to_rids"))
	    || wit_die "could not open function_to_rids";

    }
    @old   = `cat $WIT2_Config::organisms/$org/assigned_functions`;

    @keep  = grep(!/^$rid/,@old);
    @pitch = grep(/^$rid/,@old);

    foreach $x (@pitch)
    {
	chop $x;
	($rid1,$func) = split(/\t/,$x);
	($rid1 eq $rid) || wit_die "BAD: rid=$rid rid1=$rid1";
	if ($y = &accessdbH($opened_function_to_rids,$func))
	{
	    @rids = split(/\t/,$y);
	    for ($i=0; ($i <= $#rids) && ($rids[$i] ne $rid); $i++) {}
	    if ($i <= $#rids)
	    {
		splice(@rids,$i,1);
		$y = join("\t",@rids);
		&replacedbH($opened_function_to_rids,$func,$y);
	    }
	    else
	    {
		warn "failed to delete $rid from $func in function_to_rids\n";
	    }
	}
    }

    if (! $function)
    {
	&write_file("$WIT2_Config::organisms/$org/assigned_functions",\@keep);
	if (! $no_annotations)
	{
	    &add_annotation($rid,$user,"\#~\n$~\n",1);
	}
	&deletedbH($opened_assigned_functions,$rid);
    }
    else
    {
	@annotations = ();
	push(@keep,"$rid\t$function\n");
	push(@annotations,"#~ $function\n");
	if ($y = &accessdbH($opened_function_to_rids,$just_function))
	{
	    @rids = split(/\t/,$y);
	    if (! grep(/$rid/,@rids))
	    {
		&replacedbH($opened_function_to_rids,$just_function,"$y\t$rid");
	    }
	}
	else
	{
	    &insertdbH($opened_function_to_rids,$just_function,"$just_function\t$rid");
	}
	&write_file("$WIT2_Config::organisms/$org/assigned_functions",\@keep);

	if (! $no_annotations)
	{
	    $annotation = join("",@annotations);
	    &add_annotation($rid,$user,$annotation,1);
	}

	&replacedbH($opened_assigned_functions,$rid,$function);
    }

    if ($opened_function_to_rids)
    {
	&sync_dbH($opened_function_to_rids);
    }

    if ($opened_assigned_functions)
    {
	&sync_dbH($opened_assigned_functions);
    }
}

sub assign_function_by_non_master {
    my($rid,$user,$function) = @_;
    my(@old,@keep,$func,@pitch);

    $user || return "";
    $org = substr($rid,1,2);

    delete $loaded_user_assignments{"$org\t$user"};

    &ok_directory("$WIT2_Config::organisms/$org/UserModels");
    &ok_directory("$WIT2_Config::organisms/$org/UserModels/$user");

    if (-s "$WIT2_Config::organisms/$org/UserModels/$user/assigned_functions")
    {
	@old = `cat $WIT2_Config::organisms/$org/UserModels/$user/assigned_functions`;
    }
    else
    {
	@old = ();
    }
    @keep  = grep(!/^$rid/,@old);
    @pitch = grep(/^$rid/,@old);

    if ((! $function) && (! @pitch))
    {
	push(@keep,"$rid\t");
    }
    elsif ($function)
    {
	push(@keep,"$rid\t$function\n");
    }

    if (! &write_file("$WIT2_Config::organisms/$org/UserModels/$user/assigned_functions",\@keep))
    {
	wit_die "could not make an assignment to $WIT2_Config::organisms/$org/UserModels/$user/assigned_functions";
    }
    if ($function =~ /^(.*)\t(\S)/)
    {
	$user_func_of{$rid} = [$1,$2];
    }
    else
    {
	$user_func_of{$rid} = [$function,""];
    }
}

sub set_locationO {
    my($rid,$loc) = @_;

    return &set_locationF($rid,$loc);
}

sub set_cannonicalO {
    my($rid,$cannonical) = @_;

    return &set_cannonicalF($rid,$cannonical);
}

sub typeO {
    my($prot) = @_;
#   $prot || confess "prot=$prot";
    
    $prot = &reduce_id( $prot );
    
    if ( $prot =~ /^R[A-Z]{2}\d{5}$/          )  { return "rid"; }
    if ( $prot =~ /^sp\|[OPQ]\d[0-9A-Z]{3}\d/ )  { return "sp";  }
    if ( $prot =~ /^tr\|\w+/                  )  { return "sp";  }
    if ( $prot =~ /^tn\|\w+/                  )  { return "sp";  }
    if ( $prot =~ /^gi\|\d+/                  )  { return "gi";  }
    
    return "";
}

sub update_assignmentsO {
    my($org,$rid,$human,$machine,$user) = @_;

    if ($human)
    {
	&assign_functionO($rid,$user,$human,1);
    }

    if ($machine)
    {
	if (($machine =~ /^((sp|tr|tn)\|[A-Za-z0-9]+)/) && ($x = &assigned_functionsO($1,$user)) && $x->[0])
	{
	    &assign_functionO($rid,$user,$x->[0],1);
	}
	elsif (($machine =~ /^(gi\|\d+)/) && ($x = &assigned_functionsO($1,$user)) && $x->[0])
	{
	    &assign_functionO($rid,$user,$x->[0],1);
	}
	elsif (($machine =~ /^(R[A-Z]{2}\d{5})/) && 
	       ($x = &assigned_functionsO($1,$user)) && 
	       $x->[0])
	{
	    &assign_functionO($rid,$user,$x->[0],1);
	}
    }
}    

sub upstreamO {
    my($id,$upstream) = @_;

    return &upstreamF($id,$upstream);
}

sub all_assigned_functionsO {
    my($org,$user) = @_;
    my(@funcs,$x,$y,$rid,$func,%user,$user_assignments,$standard);

    @funcs = ();
    if ($user_assignments = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/assigned_functions"))
    {
	foreach $x (@$user_assignments)
	{
	    ($rid,$func) = split(/\t/,$x);
	    if (! ($y = $user{$rid}))
	    {
		$y = ($user{$rid} = []);
	    }
	    push(@$y,$func);
	}

	foreach $rid (keys(%user))
	{
	    if ($y = $user{$rid})
	    {
		foreach $func (@$y)
		{
		    push(@funcs,"$rid\t$func");
		}
	    }
	}
    }

    if ($standard = &fast_read("$WIT2_Config::organisms/$org/assigned_functions"))
    {
	foreach $x (@$standard)
	{
	    if (($x =~ /^(\S+)/) && (! defined($user{$1})))
	    {
		push(@funcs,$x);
	    }
	}
    }
    return \@funcs;
}

sub extended_assigned_functionsO {
    my($prot,$user) = @_;
    my($funcL,$func,$higher,$x,%function,@all);

    $funcL  = &assigned_functionsO($prot,$user);
    foreach $func (@$funcL)
    {
	$function{$func} = 1;
	$higher    = &higher_level_functional_rolesR($func);	
	foreach $x (@$higher)
	{
	    $function{$x} = 1;
	}
    }
    @all = sort keys(%function);
    return \@all;
}

sub org_ofO {
    my($prot) = @_;
    my($type);
    
    $prot = &reduce_prot_id($prot);
    
    $type = &typeO( $prot );
#   print STDERR "ORF::org_ofO --- prot=$prot, type=$type\n";
    if ($type eq "rid")
    {
        $prot =~ m/^R([A-Z]{2})/;
	return &genome_bioname($1);
    }
    elsif ( ($type eq "sp") || ($type eq "tr") || ($type eq "tn") || ($type eq "gi") )
    {
	if (! $opened_sp_org)
	{
	    ($opened_sp_org = &opendbH("sp_org"))  ||  wit_die "could not open sp_org";
	}
	
        $prot =~ s/^(sp|tr|tn|gi)\|//;
	return &accessdbH( $opened_sp_org, $prot );
    }
    
    return "";
}

sub sp_org {
    my($ac) = @_;
    my($sp,@lines,@os,$genus1,$genusR,$species);

    $sp = &swiss_prot_entry($ac);
    @lines = split(/\n/,$$sp);
    @os = grep(/^OS   /,@lines);
#   print STDERR "ORF::sp_org --- $ac\n", join( "\n", @os ), "\n";
    if ((@os  > 0) && ($os[0] =~ /^OS   (\S)(\S+)\s+(\S+)/))
    {
	$genus1 = $1; $genusR = $2; $species = $3;
#	print STDERR "genus1=$genus1\ngenusR=$genusR\tspecies=$species\n";
	$genusR =~ tr/A-Z/a-z/; $species =~ tr/A-Z/a-z/;
	if ($species ne "sp.") { $species =~ s/\.$//; }
	return "$genus1$genusR $species";
    }
    return "";
}

sub gi_org {
    my($gi) = @_;
    my($out,@os);

    if (! $opened_assigned_functions_nr)
    {
	($opened_assigned_functions_nr = &opendbH("nr_func"))
	    || wit_die "could not open nr_func";
    }
    
    if ($funcs = &accessdbH($opened_assigned_functions_nr,"gi\|$gi"))
    {
	if ($funcs =~ /\[([^]]+)\]/)
        {
	    return $1;
	}
    }

    $out = &gi_entries($gi);
    @os = grep(/^\s*ORGANISM.*>([^<]+)<\/a>/,@$out);
    if ((@os > 0) && ($os[0] =~ /^\s*ORGANISM.*>([^<]+)<\/a>/))
    {
	return $1;
    }
    return "";
}

sub equivalent_sp {
    my($prot) = @_;
    my($x,$i, $id);

    if ((&typeO($prot) eq "rid") && ($x = &aliasesO($prot)))
    {
	for ($i=0; ($i <= $#{$x}) && ($x->[$i] !~ /^sp\|/); $i++) {}
	if ($i <= $#{$x})
	{
	    $id = &reduce_id ($x->[$i]);
	    return \$id;
	}
    }
    return "";
}

1;

