package PW;			# -*- perl -*-

use Util;
use HashDB;
use Annotation;
use Carp;
use Messages;

require Exporter;
@ISA = (Exporter);
@EXPORT = qw(
	        add_pathway_to_collectionP
	        all_asserted_pathwaysP
	        all_categoriesP
	        all_suppressed_pathwaysP
	        asserted_pathwayP
	        assert_pathwaysP
	        category_pwsP
	        delete_pathway_from_collectionP
	        delete_pathwaysP
		functional_rolesP
		name_ofP
		names_ofP
	        organismsP
	        pw_categoriesP
	        replace_pathway_in_collectionP
	        suppressed_pathwayP
	        suppress_pathwaysP
	        unsuppress_pathwaysP
);

sub all_categoriesP {
    return
	[
	 ["Amino Acid Metabolism",                     "aac.txt"],
	 ["Aromatic Hydrocarbons",                     "aro.txt"],
	 ["Carbohydrate Metabolism",                   "car.txt"],
	 ["Coenzymes and Vitamins",                    "coe.txt"],
	 ["Electron Transport",                        "elt.txt"],
	 ["Enzyme Metabolism",                         "enz.txt"],
	 ["Exobiotics",                                "exo.txt"],
	 ["Halide Metabolism",                         "hal.txt"],
	 ["Hydrogen Metabolism",                       "hyd.txt"],
	 ["Hydrocarbon Metabolism",                    "hyc.txt"],
	 ["Intracellular Transport",                   "int.txt"],
	 ["Lipid Metabolism",                          "lip.txt"],
	 ["Membrane Transport",                        "mtr.txt"],
         ["Metabolism of Oxygen and Its Radicals",     "oxy.txt"],
	 ["Nitrogen Metabolism",                       "nit.txt"],
	 ["Nucleic Acids Metabolism",                  "nuc.txt"],
	 ["One-Carbon Metabolism",                     "c1.txt"],
	 ["Phosphate Metabolism",                      "pho.txt"],
	 ["Protein Metabolism",                        "pro.txt"],
	 ["Purine Metabolism",                         "pur.txt"],
	 ["Pyrimidine Metabolism",                     "pyr.txt"],
	 ["Signal Transduction",                       "sig.txt"],
	 ["Sulfur Metabolism",                         "sul.txt"]
	];
}

sub delete_pathway_from_collectionP {
    my($pw,$user) = @_;
    my(@hits);
    
    &delete_pw_category($pw,$user);
    &delete_pw_names($pw,$user);
    &delete_pw_functions($pw,$user);
    
    return 1;
}

##########################################
sub delete_pw_category {
    my($pw,$user) = @_;

    if (! $user)
    {
	&delete_pw_category_master($pw);
    }
    else
    {
	&delete_pw_category_user($pw,$user);
    }
}

sub delete_pw_category_user {
    my($pw,$user) = @_;

    if (-s "$WIT2_Config::global/UserModels/$user/pw_category.table")
    {
	@tmp = grep(!/^$pw\t/,@{ &fast_read("$WIT2_Config::global/UserModels/$user/pw_category.table")});
	$tmp = join("\n",@tmp) . "\n";
	if (rename("$WIT2_Config::global/UserModels/$user/pw_category.table",
		   "$WIT2_Config::global/UserModels/$user/pw_category.table~") &&
	    open(TMP,">$WIT2_Config::global/UserModels/$user/pw_category.table"))
	{
	    print TMP $tmp;
	    close(TMP);
	}
    }
}

sub delete_pw_category_master {
    my($pw) = @_;
    my($ref1,$ref2,@pw_cats,$i,@pws,$x,$category);

    (($ref1 = &opendbH("pw_categories",1)) &&
     ($ref2 = &opendbH("category_pws",1)))
	|| wit_die "category dbs are messed up";

    if ($category = &accessdbH($ref1,$pw))
    {
	&insertdbH($ref1,$pw,"");
    }
    &closedbH($ref1);

    if (! $category)
    {
	return 0;
    }

    if ($x = &accessdbH($ref2,$category))
    {
	@pws = split(/\t/,$x);
	for ($i=0; ($i <= $#pws) && ($pws[$i] ne $pw); $i++) {}
	if ($i <= $#pws)
	{
	    splice(@pws,$i,1);
	    $x = join("\t",@pws);
	    &insertdbH($ref2,$category,$x);
	}
    }
    &closedbH($ref2);

    @pw_cats = `cat $WIT2_Config::global/pw_category.table`;
    $i = 0;
    while ($i <= $#pw_cats)
    {
	if ($pw_cats[$i] =~ /^$pw\t/)
	{
	    splice(@pw_cats,$i,1);
	}
	else
	{
	    $i++;
	}
    }
    @pw_cats = sort @pw_cats;
    &write_file("$WIT2_Config::global/pw_category.table",\@pw_cats);
}

##########################################

sub delete_pw_names {
    my($pw,$user) = @_;

    if (! $user)
    {
	&delete_pw_names_master($pw);
    }
    else
    {
	&delete_pw_names_user($pw,$user);
    }
}

sub delete_pw_names_user {
    my($pw,$user) = @_;

    if (-s "$WIT2_Config::global/UserModels/$user/pw_names.table")
    {
	@tmp = grep(!/^$pw\t/,@{ &fast_read("$WIT2_Config::global/UserModels/$user/pw_names.table")});
	$tmp = join("\n",@tmp) . "\n";
	if (rename("$WIT2_Config::global/UserModels/$user/pw_names.table",
		   "$WIT2_Config::global/UserModels/$user/pw_names.table~") &&
	    open(TMP,">$WIT2_Config::global/UserModels/$user/pw_names.table"))
	{
	    print TMP $tmp;
	    close(TMP);
	}
    }
}

sub delete_pw_names_master {
    my($pw) = @_;
    my($ref,@entries,$i);
    
    ($ref = &opendbH("pw_names",1)) || wit_die "could not open pw_names";
    &insertdbH($ref,$pw,"");
    &closedbH($ref);

    @entries = `cat $WIT2_Config::global/pw_names.table`;
    for ($i=0; ($i <= $#entries) && ($entries[$i] !~ /^$pw/); $i++) {}
    if ($i <= $#entries)
    {
	splice(@entries,$i,1);
    }
    &write_file("$WIT2_Config::global/pw_names.table",\@entries);
}

##########################################

sub delete_pw_functions {
    my($pw,$user) = @_;

    if (! $user)
    {
	&delete_pw_functions_master($pw);
    }
    else
    {
	&delete_pw_functions_user($pw,$user);
    }
}

sub delete_pw_functions_user {
    my($pw,$user) = @_;

    if (-s "$WIT2_Config::global/UserModels/$user/pw_func.table")
    {
	@tmp = grep(!/^$pw\t/,@{ &fast_read("$WIT2_Config::global/UserModels/$user/pw_func.table")});
	$tmp = join("\n",@tmp) . "\n";
	if (rename("$WIT2_Config::global/UserModels/$user/pw_func.table",
		   "$WIT2_Config::global/UserModels/$user/pw_func.table~") &&
	    open(TMP,">$WIT2_Config::global/UserModels/$user/pw_func.table"))
	{
	    print TMP $tmp;
	    close(TMP);
	}
    }
}

sub delete_pw_functions_master {
    my($pw) = @_;
    my($ref1,$ref2,@pws,$role,@entries,$i,$x,@roles);

    @entries = `cat $WIT2_Config::global/pw_func.table`;
    for ($i=$#entries; ($i >= 0); $i--)
    {
	if ($entries[$i] =~ /^$pw\b/)
	{
	    splice(@entries,$i,1);
	}
    }
    &write_file("$WIT2_Config::global/pw_func.table",\@entries);

    ($ref1 = &opendbH("pw_func",1)) || wit_die "could not open pw_func";
    $roles = &accessdbH($ref1,$pw);
    @roles = split(/\t/,$roles);

    &insertdbH($ref1,$pw,"");
    &closedbH($ref1);

    ($ref2 = &opendbH("func_pw",1)) || wit_die "could not open func_pw";
    foreach $role (@roles)
    {
	if ($x = &accessdbH($ref2,$role))
	{
	    @pws = split(/\n/,$x);
	    for ($i=0; ($i <= $#pws) && ($pws[$i] ne $pw); $i++) {}
	    if ($i <= $#pws)
	    {
		splice(@pws,$i,1);
		&insertdbH($ref2,$role,join("\n",@pws));
	    }
	}
    }
    &closedbH($ref2);
}

##########################################

sub add_pathway_to_collectionP {
    my($pw,$mpw,$spn,$apn,$category,$roles,$user) = @_;

    if (! $user)
    {
	&add_pathway_to_collectionP_master($pw,$mpw,$spn,$apn,$category,$roles);
    }
    else
    {
	&add_pathway_to_collectionP_user($pw,$mpw,$spn,$apn,$category,$roles,$user,$user);
    }
}

sub add_pathway_to_collectionP_master {
    my($pw,$mpw,$spn,$apn,$category,$roles) = @_;
    my(@hits);
    
    @hits = grep(/^$pw\b/,`cut -f1 $WIT2_Config::global/pw_names.table`);
    if (@hits > 0)
    {
	return 0;
    }

    if ($category)
    {
	&set_pw_category_master($pw,$category);
    }
    &set_pw_name_entry_master($pw,$mpw,$spn,$apn);
    &set_pw_func_entries_master($pw,$roles);
    return 1;
}

sub add_pathway_to_collectionP_user {
    my($pw,$mpw,$spn,$apn,$category,$roles,$user) = @_;
    my(@hits);

    if (-s "$WIT2_Config::global/UserModels/$user/pw_names.table")
    {

	@hits = grep(/^$pw\b/,`cut -f1 $WIT2_Config::global/UserModels/$user/pw_names.table`);
	if (@hits > 0)
	{
	    return 0;
	}
    }

    if (!(-d "$WIT2_Config::global/UserModels"))
    {
	mkdir("$WIT2_Config::global/UserModels",0777)
	    || wit_die "could not make $WIT2_Config::global/UserModels";
    }

    if (!(-d "$WIT2_Config::global/UserModels/$user"))
    {
	mkdir("$WIT2_Config::global/UserModels/$user",0777)
	    || wit_die "could not make $WIT2_Config::global/UserModels/$user";
    }
    
    if ($category)
    {
	&set_pw_category_user($pw,$category,$user);
    }
    else
    {
	&set_pw_category_user($pw,"",$user);
    }

    &set_pw_name_entry_user($pw,$mpw,$spn,$apn,$user);
    &set_pw_func_entries_user($pw,$roles,$user);
    return 1;
}

sub replace_pathway_in_collectionP {
    my($pw,$mpw,$spn,$apn,$category,$roles,$user) = @_;

    if (! $user)
    {
	&replace_pathway_in_collectionP_master($pw,$mpw,$spn,$apn,$category,$roles);
    }
    else
    {
	&replace_pathway_in_collectionP_user($pw,$mpw,$spn,$apn,$category,$roles,$user,$user);
    }
}

sub replace_pathway_in_collectionP_master {
    my($pw,$mpw,$spn,$apn,$category,$roles) = @_;
    my(@hits);
    
    @hits = grep(/^$pw\b/,`cut -f1 $WIT2_Config::global/pw_names.table`);
    if (@hits == 0)
    {
	return 0;
    }

    if ($category)
    {
	&set_pw_category_master($pw,$category);
    }
    else
    {
	&set_pw_category_master($pw,"");
    }
    &set_pw_name_entry_master($pw,$mpw,$spn,$apn);
    &set_pw_func_entries_master($pw,$roles);
    return 1;
}

sub replace_pathway_in_collectionP_user {
    my($pw,$mpw,$spn,$apn,$category,$roles,$user) = @_;
    my(@hits);
    
    (-s "$WIT2_Config::global/UserModels/$user/pw_names.table")
	|| (return 0);

    @hits = grep(/^$pw\b/,`cut -f1 $WIT2_Config::global/UserModels/$user/pw_names.table`);
    (@hits > 0) || (return 0);

    if ($category)
    {
	&set_pw_category_user($pw,$category,$user);
    }
    else
    {
	&set_pw_category_user($pw,"",$user);
    }
    &set_pw_name_entry_user($pw,$mpw,$spn,$apn,$user);
    &set_pw_func_entries_user($pw,$roles,$user);
    return 1;
}

sub set_pw_name_entry_master {
    my($pw,$mpw,$spn,$apn) = @_;
    my($ref,@entries,$i);
    
    ($ref = &opendbH("pw_names",1)) || wit_die "could not open pw_names";
    &insertdbH($ref,$pw,"$spn\t$mpw\t$apn");
    &closedbH($ref);

    @entries = `cat $WIT2_Config::global/pw_names.table`;
    for ($i=0; ($i <= $#entries) && ($entries[$i] !~ /^$pw/); $i++) {}
    $entries[$i] = "$pw\t$spn\t$mpw\t$apn\n";
    &write_file("$WIT2_Config::global/pw_names.table",\@entries);
}

sub set_pw_name_entry_user {
    my($pw,$mpw,$spn,$apn,$user) = @_;
    my($ref,@entries,$i);
    
    if (-s "$WIT2_Config::global/UserModels/$user/pw_names.table")
    {
	@entries = `cat $WIT2_Config::global/UserModels/$user/pw_names.table`;
    }
    else
    {
	@entries = ();
    }

    for ($i=0; ($i <= $#entries) && ($entries[$i] !~ /^$pw/); $i++) {}
    $entries[$i] = "$pw\t$spn\t$mpw\t$apn\n";
    &write_file("$WIT2_Config::global/UserModels/$user/pw_names.table",\@entries);
}

sub set_pw_func_entries_master {
    my($pw,$roles) = @_;
    my($ref1,$ref2,@pws,$role,@entries,$i,$x);

    @entries = `cat $WIT2_Config::global/pw_func.table`;
    for ($i=$#entries; ($i >= 0); $i--)
    {
	if ($entries[$i] =~ /^$pw\b/)
	{
	    splice(@entries,$i,1);
	}
    }
    foreach $role (@$roles)
    {
	push(@entries,"$pw\t$role\n");
    }
    &write_file("$WIT2_Config::global/pw_func.table",\@entries);

    ($ref1 = &opendbH("pw_func",1)) || wit_die "could not open pw_func";
    &insertdbH($ref1,$pw,join("\t",@$roles));
    &closedbH($ref1);

    ($ref2 = &opendbH("func_pw",1)) || wit_die "could not open func_pw";
    foreach $role (@$roles)
    {
	if ($x = &accessdbH($ref2,$role))
	{
	    @pws = split(/\n/,$x);
	    for ($i=0; ($i <= $#pws) && ($pws[$i] ne $pw); $i++) {}
	    if ($i > $#pws)
	    {
		&insertdbH($ref2,$role,"$x\n$pw");
	    }
	}
	else
	{
	    &insertdbH($ref2,$role,$pw);
	}
    }
    &closedbH($ref2);
}

sub set_pw_func_entries_user {
    my($pw,$roles,$user) = @_;
    my($ref1,$ref2,@pws,$role,@entries,$i,$x);

    if (-s "$WIT2_Config::global/UserModels/$user/pw_func.table")
    {
	@entries = `cat $WIT2_Config::global/UserModels/$user/pw_func.table`;
    }
    else
    {
	@entries = ();
    }

    for ($i=$#entries; ($i >= 0); $i--)
    {
	if ($entries[$i] =~ /^$pw\b/)
	{
	    splice(@entries,$i,1);
	}
    }
    foreach $role (@$roles)
    {
	push(@entries,"$pw\t$role\n");
    }
    &write_file("$WIT2_Config::global/UserModels/$user/pw_func.table",\@entries);
}

sub set_pw_category_master {
    my($pw,$category) = @_;
    my($ref1,$ref2,@pw_cats,$i,@pws,$x,$old_category);

    (($ref1 = &opendbH("pw_categories",1)) &&
     ($ref2 = &opendbH("category_pws",1)))
	|| wit_die "category dbs are messed up";

    $old_category = &accessdbH($ref1,$pw);
    &insertdbH($ref1,$pw,$category);
    &closedbH($ref1);

    if ($old_category && ($x = &accessdbH($ref2,$old_category)))
    {
	@pws = split(/\t/,$x);
	for ($i=0; ($i <= $#pws) && ($pws[$i] ne $pw); $i++) {}
	if ($i <= $#pws)
	{
	    splice(@pws,$i,1);
	    $x = join("\t",@pws);
	    &insertdbH($ref2,$old_category,$x);
	}
    }

    if ($category && ($x = &accessdbH($ref2,$category)))
    {
	@pws = split(/\t/,$x);
	for ($i=0; ($i <= $#pws) && ($pws[$i] ne $pw); $i++) {}
	if ($i > $#pws)
	{
	    &insertdbH($ref2,$category,"$x\t$pw");
	}
    }
    else
    {
	&insertdbH($ref2,$category,$pw);
    }
    &closedbH($ref2);

    @pw_cats = `cat $WIT2_Config::global/pw_category.table`;
    $i = 0;
    while ($i <= $#pw_cats)
    {
	if ($pw_cats[$i] =~ /^$pw\t/)
	{
	    splice(@pw_cats,$i,1);
	}
	else
	{
	    $i++;
	}
    }
    push(@pw_cats,"$pw\t$category\n");
    @pw_cats = sort @pw_cats;
    &write_file("$WIT2_Config::global/pw_category.table",\@pw_cats);
}

sub set_pw_category_user {
    my($pw,$category,$user) = @_;
    my($ref1,$ref2,@pw_cats,$i,@pws,$x,$old_category);

    $category || return;

    if (-s "$WIT2_Config::global/UserModels/$user/pw_category.table")
    {
	@pw_cats = `cat $WIT2_Config::global/UserModels/$user/pw_category.table`;
    }
    else
    {
	@pw_cates = ();
    }

    $i = 0;
    while ($i <= $#pw_cats)
    {
	if ($pw_cats[$i] =~ /^$pw\t/)
	{
	    splice(@pw_cats,$i,1);
	}
	else
	{
	    $i++;
	}
    }
    push(@pw_cats,"$pw\t$category\n");
    @pw_cats = sort @pw_cats;
    &write_file("$WIT2_Config::global/UserModels/$user/pw_category.table",\@pw_cats);
}

##########################################

sub pw_categoriesP {
    my($pw,$user) = @_;
    my($x,@cats);

    if (! $pw_cats_ref)
    {
	$pw_cats_ref = &opendbH("pw_categories");
    }

    if ($x = &accessdbH($pw_cats_ref,$pw))
    {
	@cats = split(/\t/,$x);
	return \@cats;
    }
    elsif ($user && (-s "$WIT2_Config::global/UserModels/$user/pw_category.table"))
    {
	@cats = map { $_ =~ /^$pw\t(\S.*\S)\s*$/; $1; } 
	          `grep \"$pw\" $WIT2_Config::global/UserModels/$user/pw_category.table` ;
	return \@cats;
    }
    else
    {
	return [];
    }
}

sub category_pwsP {
    my($cat,$user) = @_;
    my($x,@pws);

    if (! $cat_pws_ref)
    {
	$cat_pws_ref = &opendbH("category_pws");
    }

    if ($x = &accessdbH($cat_pws_ref,$cat))
    {
	@pws = split(/\t/,$x);
    }
    else
    {
	@pws = ();
    }

    if ($user && (-s "$WIT2_Config::global/UserModels/$user/pw_category.table"))
    {
	@tmp = map { $_ =~ /^(\S+)\t$cat\s*$/; $1; } 
	          `grep \"$pw\" $WIT2_Config::global/UserModels/$user/pw_category.table` ;
	push(@pws,@tmp);
    }

    if (@pws > 0)
    {
	return \@pws;
    }
    else
    {
	return [];
    }
}

sub name_ofP {
    my($pw,$user) = @_;
    my($x,$spn,$mpw,$apns,@tmp);

    $spn = $mpw = $apns = "";

    if (! $pw_names_ref)
    {
	$pw_names_ref = &opendbH("pw_names");
    }

    if ($x = &accessdbH($pw_names_ref,$pw))
    {
	($spn,$mpw,$apns) = split(/\t/,$x);

    }
    elsif ($user && (-s "$WIT2_Config::global/UserModels/$user/pw_names.table"))
    {
	@tmp = `grep \"^$pw\" $WIT2_Config::global/UserModels/$user/pw_names.table` ;
	if (@tmp > 0)
	{
	    ($spn,$mpw,$apns) = split(/\t/,$tmp[0]);
	}
    }
	
    if ($mpw)
    {
	return \$mpw;
    }
    elsif ($spn)
    {
	return \$spn;
    }
    elsif ($apns)
    {
	$apns =~ s/;.*$//;
	return \$apns;
    }
    else
    {
	return \"";
    }
}

sub names_ofP {
    my($pw,$user) = @_;
    my($x,$spn,$mpw,$apns,@tmp);

    $spn = $mpw = $apns = "";

    if (! $pw_names_ref)
    {
	$pw_names_ref = &opendbH("pw_names");
    }

    if ($x = &accessdbH($pw_names_ref,$pw))
    {
	($spn,$mpw,$apns) = split(/\t/,$x);

    }
    elsif ($user && (-s "$WIT2_Config::global/UserModels/$user/pw_names.table"))
    {
	@tmp = `grep \"^$pw\" $WIT2_Config::global/UserModels/$user/pw_names.table` ;
	if (@tmp > 0)
	{
	    ($spn,$mpw,$apns) = split(/\t/,$tmp[0]);
	}
    }
    return ($spn,$mpw,$apns);
}

sub functional_rolesP {
    my($pw,$user) = @_;
    my($x,@funcs);

    if (! $pw_func_ref)
    {
	$pw_func_ref = &opendbH("pw_func");
    }

    if ($x = &accessdbH($pw_func_ref,$pw))
    {
	@funcs = split(/\t/,$x);
	return \@funcs;
    }
    elsif ($user &&
	   (-s "$WIT2_Config::global/UserModels/$user/pw_func.table"))
    {
	@funcs = `grep \"^$pw\" $WIT2_Config::global/UserModels/$user/pw_func.table` ;
	if (@funcs = map { $_ =~ /^\S+\t(\S.*\S)\s*$/; $1; } @funcs)
	{
	    return \@funcs;
	}
    }
    return [];
}

sub organismsP {
    my($pw,$user) = @_;
    my($x,@tmp,%hits);

    foreach $x (`grep $pw $WIT2_Config::organisms/*/asserted_pathways`)
    {
	if ($x =~ /^.*([A-Z]{2})\/asserted_pathways.*$/)
	{
	    $hits{$1} = 1;
	}
    }

    if ($user)
    {
	foreach $x (`grep $pw $WIT2_Config::organisms/*/UserModels/$user/asserted_pathways`)
	{
	    if ($x =~ /^.*([A-Z]{2})\/asserted_pathways.*$/)
	    {
		$hits{$1} = 1;
	    }
	}
    }
    @tmp = sort keys(%hits);
    return \@tmp;
}

sub asserted_pathwayP {
    my($org,$user,$pw) = @_;
    my($x,$pws);

    if ($user && ($pws = &user_asserted_pathways($org,$user,$pw)) && (defined($x = $pws->{$pw})))
    {
        return $x;
    }
    return (($pws = &standard_asserted_pathways($org)) && (defined ($x = $pws->{$pw})));
}

sub user_asserted_pathways {
    my($org,$user,$pw) = @_;
    my($line,$x,$pws);

    if (! defined($pws = $user_asserted_pathways{"$org\t$user"}))
    {
	if (defined ($x = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/asserted_pathways")))
	{
	    $user_asserted_pathways{"$org\t$user"} = ($pws = {});
	    foreach $line (@$x)
	    {
		if ($line =~ /^(\S+)\s+([01])\s*$/)
		{
		    $pws->{$1} = $2;
		}
	    }
	}
	else
	{
	    $pws = ($user_asserted_pathways{"$org\t$user"} = {});
	}
    }
    return $pws;
}

sub standard_asserted_pathways {
    my($org) = @_;
    my($x,$pws);

    if (! defined($pws = $standard_asserted_pathways{$org}))
    {
	if ($x = &fast_read("$WIT2_Config::organisms/$org/asserted_pathways"))
	{
	    $standard_asserted_pathways{$org} = ($pws = {});
	    foreach $pw (@$x)
	    {
		$pws->{$pw} = 1;
	    }
	}
	else
	{
	    $pws = ($standard_asserted_pathways{$org} = {});
	}
    }
    return $pws;
}

sub assert_pathwaysP {
    my($org,$user,@pws) = @_;

    (-d "$WIT2_Config::organisms/$org") || return 0;

    if (&master_user($org,$user))
    {
	return &assert_pws_by_master($org,$user,@pws);
    }
    else
    {
	return &assert_pws_by_non_master($org,$user,@pws);
    }
}

sub assert_pws_by_master {
    my($org,$user,@pws) = @_;
    my(@hits,$x,$y,$pw,%in);

    if (! ($x = &fast_read("$WIT2_Config::organisms/$org/asserted_pathways")))
    {
	$x = [];
    }
    
    foreach $pw (@pws)
    {
	$in{$pw} = 1;
	&add_annotation($org,$pw,$user,"asserted",1);
    }

    @hits = ();
    foreach $pw (@$x)
    {
	if (! $in{$pw})
	{
	    push(@hits,"$pw\n");
	}
    }
    foreach $pw (keys(%in))
    {
	push(@hits,"$pw\n");
    }
    return &write_file("$WIT2_Config::organisms/$org/asserted_pathways",\@hits);
}

sub assert_pws_by_non_master {
    my($org,$user,@pws) = @_;
    my(@hits,$x,$y,$pw,%in);

    if (! -d "$WIT2_Config::organisms/$org/UserModels")
    {
	mkdir("$WIT2_Config::organisms/$org/UserModels",0777) 
	    || wit_die "could not make $WIT2_Config::organisms/$org/UserModels";
	system "chmod 777 $WIT2_Config::organisms/$org/UserModels";
    }

    if (! -d "$WIT2_Config::organisms/$org/UserModels/$user")
    {
	mkdir("$WIT2_Config::organisms/$org/UserModels/$user",0777) 
	    || wit_die "could not make $WIT2_Config::organisms/$org/UserModels/$user";
	system "chmod 777 $WIT2_Config::organisms/$org/UserModels/$user";
    }

    if (! ($x = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/asserted_pathways")))
    {
	$x = [];
    }

    foreach $pw (@pws)
    {
	$in{$pw} = 1;
    }

    @hits = ();
    foreach $x (@$x)
    {
	if (!(($x =~ /^(\S+)/) && $in{$1}))
	{
	    push(@hits,"$x\n");
	}
    }

    foreach $pw (keys(%in))
    {
	push(@hits,"$pw\t1\n");
    }
    return &write_file("$WIT2_Config::organisms/$org/UserModels/$user/asserted_pathways",\@hits);
}

sub delete_pathwaysP {
    my($org,$user,@pws) = @_;

    (-d "$WIT2_Config::organisms/$org") || return 0;

    if (&master_user($org,$user))
    {
	return &delete_pws_by_master($org,@pws);
    }
    else
    {
	return &delete_pws_by_non_master($org,$user,@pws);
    }

}

sub delete_pws_by_master {
    my($org,@pws) = @_;
    my(@hits,$x,$y,$pw,%in);

    if (! ($x = &fast_read("$WIT2_Config::organisms/$org/asserted_pathways")))
    {
	$x = [];
    }
    
    foreach $pw (@pws)
    {
	$in{$pw} = 1;
	&add_annotation($org,$pw,$user,"deleted",1);
    }
    @hits = ();
    foreach $y (@$x)
    {
	if (!(($y =~ /^(\S+)/) && $in{$1}))
	{
	    push(@hits,"$y\n");
	}
    }
    return &write_file("$WIT2_Config::organisms/$org/asserted_pathways",\@hits);
}

sub delete_pws_by_non_master {
    my($org,$user,@pws) = @_;
    my(@hits,$x,$y,$pw,%in);

    if (! -d "$WIT2_Config::organisms/$org/UserModels/$user")
    {
	mkdir("$WIT2_Config::organisms/$org/UserModels/$user",0777)
	    || wit_die "could not make $WIT2_Config::organisms/$org/UserModels/$user";
	system "chmod 777 $WIT2_Config::organisms/$org/UserModels/$user";
    }

    if (! ($x = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/asserted_pathways")))
    {
	$x = [];
    }

    foreach $pw (@pws)
    {
	$in{$pw} = 1;
    }
    
    @hits = ();
    foreach $y (@$x)
    {
	if (! (($y =~ /^(\S+)/) && $in{$1}))
	{
	    push(@hits,"$y\n");
	}
    }

    foreach $y (keys(%in))
    {
	push(@hits,"$y\t0\n");
    }
    return &write_file("$WIT2_Config::organisms/$org/UserModels/$user/asserted_pathways",\@hits);
}

sub all_asserted_pathwaysP {
    my($org,$user) = @_;
    my(@pws,$line,$pw,$status,%seen,$x);

    @pws = ();
    $org || confess "org=$org user=$user";

    if ($x = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/asserted_pathways"))
    {
	foreach $line (@$x)
	{
	    ($pw,$status) = split(/\t/,$line);
	    if ($status)
	    {
		push(@pws,$pw);
	    }
	    $seen{$pw} = $status;
	}
    }

    if ($x = &fast_read("$WIT2_Config::organisms/$org/asserted_pathways"))
    {
	foreach $pw (@$x)
	{
	    if (! defined($seen{$pw}))
	    {
		push(@pws,$pw);
	    }
	    $seen{$pw} = 1;
	}
    }
    return \@pws;
}

sub suppressed_pathwayP {
    my($org,$user,$pw) = @_;
    my($x,$pws);

    if ($user && ($pws = &user_suppressed_pathways($org,$user,$pw)) && (defined($x = $pws->{$pw})))
    {
	return $x;
    }
    return (($pws = &standard_suppressed_pathways($org)) && (defined ($x = $pws->{$pw})));
}

sub user_suppressed_pathways {
    my($org,$user,$pw) = @_;
    my($line,$x,$pws);

    if (! defined($pws = $user_suppressed_pathways{"$org\t$user"}))
    {
	if ($x = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/suppressed_pathways"))
	{
	    $user_suppressed_pathways{"$org\t$user"} = ($pws = {});
	    foreach $line (@$x)
	    {
		if ($line =~ /^(\S+)\s+([01])\s*$/)
		{
		    $pws->{$1} = $2;
		}
	    }
	}
	else
	{
	    $pws = ($user_suppressed_pathways{"$org\t$user"} = {});
	}
    }
    return $pws;
}

sub standard_suppressed_pathways {
    my($org) = @_;
    my($x,$pws);

    if (! defined($pws = $standard_suppressed_pathways{$org}))
    {
	if ($x = &fast_read("$WIT2_Config::organisms/$org/suppressed_pathways"))
	{
	    $standard_suppressed_pathways{$org} = ($pws = {});
	    foreach $pw (@$x)
	    {
		$pws->{$pw} = 1;
	    }
	}
	else
	{
	    $pws = ($standard_suppressed_pathways{$org} = {});
	}
    }
    return $pws;
}

sub all_suppressed_pathwaysP {
    my($org,$user) = @_;
    my(@pws,$line,$pw,$status,%seen,$x);

    @pws = ();
    if ($x = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/suppressed_pathways"))
    {
	foreach $line (@$x)
	{
	    ($pw,$status) = split(/\t/,$line);
	    if ($status)
	    {
		push(@pws,$pw);
	    }
	    $seen{$pw} = $status;
	}
    }

    if ($x = &fast_read("$WIT2_Config::organisms/$org/suppressed_pathways"))
    {
	foreach $pw (@$x)
	{
	    if (! defined($seen{$pw}))
	    {
		push(@pws,$pw);
	    }
	    $seen{$pw} = 1;
	}
    }
    return \@pws;
}

sub suppress_pathwaysP {
    my($org,$user,@pws) = @_;

    (-d "$WIT2_Config::organisms/$org") || return 0;

    if (&master_user($org,$user))
    {
	return &suppress_pws_by_master($org,@pws);
    }
    else
    {
	return &suppress_pws_by_non_master($org,$user,@pws);
    }
}

sub suppress_pws_by_master {
    my($org,@pws) = @_;
    my(@hits,$x,$y,$pw,%in);

    if (! ($x = &fast_read("$WIT2_Config::organisms/$org/suppressed_pathways")))
    {
	$x = [];
    }
    
    foreach $pw (@pws)
    {
	$in{$pw} = 1;
    }

    @hits = ();
    foreach $pw (@$x)
    {
	if (! $in{$pw})
	{
	    push(@hits,"$pw\n");
	}
    }
    foreach $pw (keys(%in))
    {
	push(@hits,"$pw\n");
    }
    return &write_file("$WIT2_Config::organisms/$org/suppressed_pathways",\@hits);
}

sub suppress_pws_by_non_master {
    my($org,$user,@pws) = @_;
    my(@hits,$x,$y,$pw,%in);

    if (! -d "$WIT2_Config::organisms/$org/UserModels/$user")
    {
	mkdir("$WIT2_Config::organisms/$org/UserModels/$user",0777) 
	    || wit_die "could not make $WIT2_Config::organisms/$org/UserModels/$user";
	system "chmod 777 $WIT2_Config::organisms/$org/UserModels/$user";
    }

    if (! ($x = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/suppressed_pathways")))
    {
	$x = [];
    }

    foreach $pw (@pws)
    {
	$in{$pw} = 1;
    }

    @hits = ();
    foreach $x (@$x)
    {
	if (!(($x =~ /^(\S+)/) && $in{$1}))
	{
	    push(@hits,"$x\n");
	}
    }

    foreach $pw (keys(%in))
    {
	push(@hits,"$pw\t1\n");
    }
    return &write_file("$WIT2_Config::organisms/$org/UserModels/$user/suppressed_pathways",\@hits);
}

sub unsuppress_pathwaysP {
    my($org,$user,@pws) = @_;

    (-d "$WIT2_Config::organisms/$org") || return 0;

    if (&master_user($org,$user))
    {
	return &unsuppress_pws_by_master($org,@pws);
    }
    else
    {
	return &unsuppress_pws_by_non_master($org,$user,@pws);
    }

}

sub unsuppress_pws_by_master {
    my($org,@pws) = @_;
    my(@hits,$x,$y,$pw,%in);

    if (! ($x = &fast_read("$WIT2_Config::organisms/$org/suppressed_pathways")))
    {
	$x = [];
    }
    
    foreach $pw (@pws)
    {
	$in{$pw} = 1;
    }
    @hits = ();
    foreach $y (@$x)
    {
	if (!(($y =~ /^(\S+)/) && $in{$1}))
	{
	    push(@hits,"$y\n");
	}
    }
    foreach $pw (keys(%in))
    {
	push(@hits,"$y\t0\n");
    }
    return &write_file("$WIT2_Config::organisms/$org/suppressed_pathways",\@hits);
}


sub unsuppress_pws_by_non_master {
    my($org,$user,@pws) = @_;
    my(@hits,$x,$y,$pw,%in);

    if (! -d "$WIT2_Config::organisms/$org/UserModels/$user")
    {
	mkdir("$WIT2_Config::organisms/$org/UserModels/$user",0777)
	    || wit_die "could not make $WIT2_Config::organisms/$org/UserModels/$user";
	system "chmod 777 $WIT2_Config::organisms/$org/UserModels/$user";
    }

    if (! ($x = &fast_read("$WIT2_Config::organisms/$org/UserModels/$user/suppressed_pathways")))
    {
	$x = [];
    }

    foreach $pw (@pws)
    {
	$in{$pw} = 1;
    }
    
    @hits = ();
    foreach $y (@$x)
    {
	if (! (($y =~ /^(\S+)/) && $in{$1}))
	{
	    push(@hits,"$y\n");
	}
    }

    foreach $y (keys(%in))
    {
	push(@hits,"$y\t0\n");
    }
    return &write_file("$WIT2_Config::organisms/$org/UserModels/$user/suppressed_pathways",\@hits);
}

1;
