package Sim;			# -*- perl -*-

use Carp;
use Util;
use Messages;

require Exporter;
@ISA = (Exporter);
@EXPORT = qw(
	     best_against_id2S
	     best_hitS
	     clusterS
	     connect_by_ortholog
	     connect_by_ortholog_constrained_to_nearest
	     connect_by_sim
	     display_sims
	     extract_sim_table_entries
	     get_simS
	     homologS
	     id1S
             id2S
	     ln1S
             ln2S
	     orthologS
	     ortholog_ridS
	     ortholog_rids_and_scoreS
             region1S
             region2S
	     sim_matrix
	     swsc_matrix
	     scoreS
	     toolS
	     trim_by_linkage
);

use HashDB;

#      4. current_f3_nr_sims(Id) -> ListOfSims
#      5. current_blast2_nr_sims(Id) -> ListOfSims

sub get_simS {
    my ( $protid, $max, $cutoff ) = @_;
    my ( $sims );
    
#   print STDERR "Sims::get_simS --- getting $max sims for $protid better than $cutoff\n";
    $protid = &reduce_prot_id($protid);
    $sims   = &get_sims_prot($protid,$max,$cutoff);
#   print STDERR "...got ", (scalar @$sims), " sims\n\n";
    
    return $sims;
}

sub toolS {
    my($sim) = @_;

    return $sim->[0];
}

sub id1S {
    my( $sim ) = @_;

    return $sim->[1];
}

sub ln1S {
    my( $sim ) = @_;

    return $sim->[2];
}

sub id2S {
    my( $sim ) = @_;

    ($sim && $sim->[3]) || confess "id2S: sim=@$sim";
    return $sim->[3];
}

sub ln2S {
    my( $sim ) = @_;

    return $sim->[4];
}

sub scoreS {
    my( $sim ) = @_;

    return [ "swsc,psc", $sim->[5], $sim->[6] ];
}

sub region1S {
    my( $sim ) = @_;

    if ( $sim->[2] )
    {
	return [ $sim->[7], $sim->[8] ];
    }
    else
    {
	return "";
    }
}

sub region2S {
    my( $sim ) = @_;

    if ( $sim->[2] )
    {
	return [ $sim->[9], $sim->[10] ];
    }
    else
    {
	return "";
    }
}

sub best_against_id2S {
    my ( $sim ) = @_;

    if (defined($sim->[11]))
    {
	return [$sim->[11],$sim->[12]];
    }
    else
    {
	return "";
    }
}

sub clusterS {
    my($prot,$get_cons,$genomes,$cutoff,$max,$seen) = @_;
    my(%used,$i,$x,$sims,$sim,$id1,$score,$id2);
    my($sim_set,@args,$rtn);
    local($cluster);
    if (! defined($seen))
    {
	$seen = {};
    }

    $sim_set = [];
    $cluster = [$prot];
    $used{$prot} = 1;
    for ($i=0; ($i < $max) && ($i <= $#{$cluster}); $i++)
    {
	if (ref($get_cons) eq "CODE")
	{
	    $sims   = &{$get_cons}($cluster->[$i],$cutoff);
	}
	else
	{
	    ($rtn,@args) = @$get_cons;
	    $sims   = &{$rtn}($cluster->[$i],$cutoff,@args);
	}
	foreach $sim (@$sims)
	{
	    $id1   = &id1S($sim);
	    $id2   = &id2S($sim);
	    $score = &scoreS($sim);
#	    print STDERR "checking $id2\n";
	    if (&ok_genome($id2,$genomes))
	    {
#		print STDERR "keeping $id2 from $id1 @$score\n";
		if (! $used{$id2})
		{
		    if ($seen->{$id2})
		    {
			$i = $max;
		    }
		    else
		    {
			push(@$cluster,$id2);
			$used{$id2} = 1;
		    }
		}
		push(@$sim_set,$sim);
	    }
	}
    }

    while (($id2,$x) = each %used)
    {
	$seen->{$id2} = 1;
    }

    if ($i >= $max)
    {
	return "";
    }
    return [$cluster,$sim_set];
}

sub connect_by_sim {
    my($prot,$cutoff) = @_;
    my($id2,$x,@sims,$sims);

    $sims = &get_simS($prot,1000,$cutoff);
    @sims = ();
    foreach $x (@$sims)
    {
	$id2          = &id2S($x);
	if ($id2 =~ /^R[A-Z]{2}\d{5}/)
	{
	    push(@sims,$x);
	}
    }
    return \@sims;
}

sub connect_by_ortholog {
    my($prot,$cutoff) = @_;

    return &orthologS(&get_simS($prot,1000,$cutoff),"");
}

sub best_hitS {
    my($sims,$genomes) = @_;
    my($set,$score,$id2,$sim);

    $set = [];
    foreach $sim (@$sims)
    {
	$id1   = &id1S($sim);
	$id2   = &id2S($sim);
	$score = &scoreS($sim);
	if ((! (($id1 =~ /^(R[A-Z]{2})\d{5}$/) && (index($id2,$1) == 0))) &&
	    &ok_genome($id2,$genomes) && (! &best_against_id2S($sim)))
	{
	    push(@$set,$sim);
	}
    }
    return $set;
}

sub ortholog_ridS {
    my($rid,$cutoff) = @_;
    my(@orthos,$rid1,@pairs,$pair);

    @orthos = ();
    if (! $opened_orthologs)
    {
	($opened_orthologs = &opendbH("orthologs")) 
	    || wit_die "could not open orthologs";
    }

    if ($_ = &accessdbH($opened_orthologs,$rid))
    {
	@pairs = split(/\t/,$_);
	foreach $pair (@pairs)
	{
	    ($rid1,$sc) = split(/,/,$pair);
	    (defined($sc) && defined($cutoff)) || confess "sc=$sc cutoff=$cutoff";
	    if ($sc <= $cutoff)
	    {
		push(@orthos,$rid1);
	    }
	}
    }
    return @orthos;
}
       
sub ortholog_rids_and_scoreS {
    my($rid,$cutoff) = @_;
    my(@orthos,$x);

    @orthos = ();
    foreach $x (@{&orthologS(&get_simS($rid,1000,$cutoff),"")})
    {
	
	push(@orthos,[&id2S($x),(&scoreS($x))->[2]]);
    }
    return @orthos;
}
       
sub orthologS {
    my($sims,$genomes) = @_;
    my($set,$sim,$id1,$id2,%seen,$genome);

    $set = [];
    if (@$sims == 0) { return $set; }

    $sim           = $sims->[0];
    $id1           = &id1S($sim);
    $genome        = substr($id1,1,2);
    $seen{$genome} = 1;

    foreach $sim (@$sims)
    {
	$id2   = &id2S($sim);
	if (&ok_genome($id2,$genomes) &&
	    ($id2 =~ /^R([A-Z]{2})\d{5}/) &&
	    ($genome = $1) &&
	    (! $seen{$genome}))
	{
	    $seen{$genome} = 1;
	    if (! &best_against_id2S($sim))
	    {
		push(@$set,$sim);
	    }
	}
    }
    return $set;
}

sub sim_matrix
{
    my ( $prots ) = @_;
    my ( $index, $matrix, $tmp, $i, $j, $k, $x );
    
    for ( $i=0; ($i <= $#{$prots}); $i++ )
    {
	$index{ $prots->[$i] } = $i;
    }
    
    $matrix = [];
    for ( $i=0; ($i <= $#{$prots}); $i++ )
    {
	$matrix->[$i] = [];
	for ( $j=0; ($j <= $#{$prots}); $j++ ) 
	{ 
	    $matrix->[$i]->[$j] = ""; 
	}

	$tmp = &get_simS( $prots->[$i], 1000, 1.0 );
	for ( $k=0; ($k <= $#{$tmp}); $k++ )
	{
	    $x = &id2S( $tmp->[$k] );
	    if ( $x && defined($j = $index{$x}) )
	    {
		if ( $i != $j )
		{
		    $x = &scoreS( $tmp->[$k] );
#		    print STDERR "$i,$j -- $x->[1], $x->[2]\n";
		    $matrix->[$i]->[$j] = $x->[2];
		} else {
		    $matrix->[$i]->[$j] = " ";
		}
	    }
	}
    }
    return $matrix;
}

sub swsc_matrix
{
    my ( $prots ) = @_;
    my ( $index, $matrix, $tmp, $i, $j, $k, $x );
    
    for ( $i=0; ($i <= $#{$prots}); $i++ )
    {
	$index{ $prots->[$i] } = $i;
    }
    
    $matrix = [];
    for ( $i=0; ($i <= $#{$prots}); $i++ )
    {
	$matrix->[$i] = [];
	for ( $j=0; ($j <= $#{$prots}); $j++ ) 
	{ 
	    $matrix->[$i]->[$j] = ""; 
	}

	$tmp = &get_simS( $prots->[$i], 1000, 1.0 );
	for ( $k=0; ($k <= $#{$tmp}); $k++ )
	{
	    $x = &id2S( $tmp->[$k] );
	    if ( $x && defined($j = $index{$x}) )
	    {
		if ( $i != $j )
		{
		    $x = &scoreS( $tmp->[$k] );
#		    print STDERR "$i,$j -- $x->[1], $x->[2]\n";
		    $matrix->[$i]->[$j] = $x->[1];
		} else {
		    $matrix->[$i]->[$j] = 0;
		}
	    }
	}
    }
    return $matrix;
}

sub extract_sim_table_entries
{
    my ( $prots, $sims ) = @_;
    my ( $i, $j, %index, $matrix, $x, $y);

    $matrix = [];

    for ($i=0; ($i <= $#{$prots}); $i++)
    {
	$index{$prots->[$i]} = $i;
    }
    
    for ($i=0; ($i <= $#{$prots}); $i++)
    {
	for ($j=0; ($j <= $#{$prots}); $j++)
	{
	    $matrix->[$i]->[$j] = "";
	}
    }

    foreach $sim (@$sims)
    {
	$id1   = &id1S($sim);
	$id2   = &id2S($sim);
	$score = &scoreS($sim);
	print STDERR "$id1 $id2 @$score\n";
	if (defined($x = $index{$id1}) && defined($y = $index{$id2}))
	{
	    $matrix->[$x]->[$y] = $score->[2];
	}
	else
	{
	    if ($x == $y )
	    {
		$matrix->[$x]->[$y] = " ";
	    } else {
		$matrix->[$x]->[$y] = "";
	    }
	}
    }
    return $matrix;
}

####################################################
$global = $main::global;

sub get_sims_prot {
    my($prot,$max,$cutoff) = @_;
    my($seek,$ref,$sims,@fields,$fh,$old);

    $old = $/;
    $/   = "\n";

    if (! $opened_sims)
    {
#	print STDERR "opening $WIT2_Config::global/all_sims\n";
	
	open(SIMS,"<$WIT2_Config::global/all_sims") &&
	($ref = &opendbH("sim_index")) 
	    || wit_die "could not open sim_index";
	$opened_sims = [$ref,\*SIMS];
    }

    if (defined(($seek = &accessdbH($opened_sims->[0],$prot))))
    {
	$fh = $opened_sims->[1];
	seek($fh,$seek,0);
#	print STDERR "seek=$seek\n";
	
	$sims = [];
	while (defined($_ = <$fh>) && chomp &&
	       (@fields = split(/,/,$_)) &&
	       ($fields[1] eq $prot))
	{
	    if (defined($fields[6]) && ($fields[6] <= $cutoff))
	    {
		if ($max--)
		{
		    if ($fields[6] == 0)
		    {
			$fields[6] = 0;  # clean up the 0.0e+0.0 nonsense
		    }

		    if ((@fields > 12) && $fields[12] == 0)
		    {
			$fields[12] = 0;  # clean up the 0.0e+0.0 nonsense
		    }

		    push(@$sims,[@fields]);
		}
		else
		{
		    $/ = $old;
		    return $sims;
		}
	    }
	}
	$/ = $old;
	return $sims;
    }
    else
    {
	$/ = $old;
	return [];
    }
}

sub ok_genome {
    my($prot,$genomes) = @_;

    if (substr($prot,1,2) eq "UA") { return 0; }
    if (! $genomes) { return 1; }
    return ((($prot =~ /^R([A-Z]{2})\d{5}$/) && $genomes->{$1}) ||
	    (($prot =~ /^sp\|/) && ($genomes->{"sp"} || $genomes->{"nr"})) ||
	    (($prot =~ /^gi\|/) && $genomes->{"nr"}));
}

sub refine_cluster {
    my($cluster,$link,$n) = @_;
    my(@weak,@mem,$done,$id,$x,$y,$id2,$i,%weak);

    @weak = ();
    @mem  = (@$cluster);
#   print STDERR "mem: @mem\n";
    $done = 0;
    while (! $done)
    {
	$done = 1;
	foreach $id (@mem)
	{
	    if (! $weak{$id})
	    {
#		print STDERR "checking $id\n";
		$x = $link->{$id};
		if ($x && ($#{$x} < ($n-1)))
		{
#		    print STDERR "$id is weak $#{$x}\n";
		    $done = 0;
		    $link->{$id} = "";
		    push(@weak,$id);
		    $weak{$id} = 1;
		    foreach $id2 (@$x)
		    {
#		        print STDERR "pulling $id2 -> $id\n";
			$y = $link->{$id2};
			for ($i=0; ($y->[$i] ne $id); $i++) 
			{
			    ($i <= $#{$y}) || wit_die "invalid bidrectional links $id $id2";
			}
			splice(@$y,$i,1);
		    }
		}
	    }
	}
    }
    return \@weak;
}

sub add_link {
    my($id1,$id2,$links) = @_;
    my($x,$i);
    
    if (! ($x = $links->{$id1}))
    {
	$x = ($links->{$id1} = []);
    }
    for ($i=0; ($i <= $#{$x}) && ($x->[$i] ne $id2); $i++) {}
    if ($i > $#{$x})
    {
	push(@$x,$id2);
    }
}
	

sub trim_by_linkage {
    my($pair,$n) = @_;
    my($links,$prots,$sims,$sim,$id1,$id2,$weak,%weak,$x);
    my(@new_prots,@new_sims);

    $links = {};
    ($prots,$sims) = @$pair;
    (@$sims > 0) || return [[],[]];

    foreach $sim (@$sims)
    {
	$id1 = &id1S($sim);
	$id2 = &id2S($sim);
	&add_link($id1,$id2,$links);
	&add_link($id2,$id1,$links);
    }
    $weak = &refine_cluster($prots,$links,$n);

    foreach $x (@$weak)
    {
	$weak{$x} = 1;
    }

    @new_prots = ();
    @new_sims  = ();

    foreach $x (@$prots)
    {
	if (! $weak{$x})
	{
	    push(@new_prots,$x);
	}
    }

    foreach $sim (@$sims)
    {
	$id1 = &id1S($sim);
	$id2 = &id2S($sim);

	if ((! $weak{$id1}) && (! $weak{$id2}))
	{
	    push(@new_sims,$sim);
	}
    }
    return [\@new_prots,\@new_sims];
}

sub near_orthologs {
    my($sims,$genomes,$cutoff) = @_;
    my($sim,$id1,$id2,%seen,$genome);
    my($n,$x,$set,@near_orthologs,$best_rid,$rid,$sims1,$i,$already_seen);

    @near_orthologs = ();
    if (@$sims == 0) { return $set; }

    $best_rid = {};
    foreach $rid (@$cluster)
    {
	$genome = substr($rid,1,2);
	$best_rid->{$genome} = $rid;
#	print STDERR "best: $genome = $rid\n";
    }

    $sim           = $sims->[0];
    $id1           = &id1S($sim);
    $genome        = substr($id1,1,2);
    $seen{$genome} = 1;

#    print STDERR "near orthologs for $id1\n";

    foreach $sim (@$sims)
    {
	$id2   = &id2S($sim);
#	print STDERR "against $id2\n";
	if (&ok_genome($id2,$genomes) &&
	    ($id2 =~ /^R([A-Z]{2})\d{5}/) &&
	    ($genome = $1) &&
	    (! $seen{$genome}))
	{
	    $seen{$genome} = 1;
	    if ((! &best_against_id2S($sim)) && 
		((! ($x = $best_rid->{$genome})) || ($x eq $id2)))
	    {
#		print STDERR "checking $id2\n";
		if (! $x)
		{
		    $sims1 = &get_simS($id2,1000,$cutoff);
		    $already_seen = {};
		    for ($i=0; ($i <= $#{$sims1}) && &ok_sim($sims1->[$i],$best_rid,$already_seen); $i++) {}
		    if ($i > $#{$sims1})
		    {
			$set    = &orthologS($sims1,"");
			$n = @$set;
			push(@near_orthologs,[$n,$sim]);
			$best_rid->{$genome} = $id2;
#			print STDERR "set best_rid for $genome to $id2\n";
		    }
		    else
		    {
			$n = &id2S($sims1->[$i]);
#			print STDERR "failed due to similarity to $n\n";
		    }
		}
		else
		{
#		    print STDERR "keeping - no best for $genome\n";
		    push(@near_orthologs,[0,$sim]);
		}
		    
	    }
	}
    }

    foreach $x (@near_orthologs)
    {
#	print STDERR "n: $x->[0] against: $x->[1]->[3]\n";
    }

    @near_orthologs = sort { $b->[0] <=> $a->[0] } @near_orthologs;
    $set = [];
    foreach $i (@near_orthologs)
    {
	push(@$set,$i->[1]);
    }
    return $set;
}

sub ok_sim {
    my($sim,$best_rid,$already_seen) = @_;
    my($id2,$x);

    $id2 = &id2S($sim);
    if ($id2 =~ /^R([A-Z]{2})\d{5}/)
    {
	if ($already_seen->{$1} || (! ($x = $best_rid->{$1})) || ($x eq $id2))
	{
	    $already_seen->{$1} = 1;
	    return 1;
	}
	else
	{
	    return 0;
	}
    }
    else
    {
	return 1;
    }
}

sub connect_by_ortholog_constrained_to_nearest {
    my($prot,$cutoff) = @_;

    return &near_orthologs(&get_simS($prot,1000,$cutoff),"",$cutoff);
}

sub display_sims {
    my($sims) = @_;
    my($x,$tool,$id1,$ln1,$id2,$ln2,$score,$region1,$region2,$best_against);
    my($funcL,$func);

    foreach $x (@$sims)
    {
	$tool         = &toolS($x);
	$id1          = &id1S($x);
	$ln1          = &ln1S($x);
	$id2          = &id2S($x);
	$ln2          = &ln2S($x);
	$score        = &scoreS($x);
	$region1      = &region1S($x);
	$region2      = &region2S($x);
	$best_against = &best_against_id2S($x);
	print "$tool,$id1,$ln1,$id2,$ln2,@$score,@$region1,@$region2,@$best_against\n";
	$funcL   = &ORF::assigned_functionsO($id2);
	$func    = join("; ",@$funcL);
	if ($func) { print "    $func\n";}
    }
    print "\n";
}

sub homologS {
    my($rid,$cutoff) = @_;
    my(@homologs) = ();
    my($sims,$id2) = @_;

    $sims = &get_simS($rid, 1000, $cutoff);
    foreach $x (@$sims)
    {
	$id2          = &id2S($x);
	if ($id2 =~ /^R[A-Z]{2}\d{5}$/)
	{
	    push(@homologs,$id2);
	}
    }
    return @homologs;
}

1;


