package Taxonomy;		# -*- perl -*-

require Exporter;
@ISA = (Exporter);
@EXPORT = qw(
	taxonomy_of_prot
	taxonomy_of_dna
	taxonomy_of_org
	taxonomy_of
        phylogenetic_distance	     
);

use tree_utilities;
use Html;
use WIT2_Config;
use Carp;
use Messages;

sub taxonomy_of_prot {
    my($id) = @_;

    return &taxonomy_of($id,"p");
}

sub taxonomy_of_dna {
    my($id) = @_;
    my($gid,$emb,$i,@report);
    
    if ($id =~ /^gi\|/)
    {
	$gid = $id;
    }
    elsif ($id =~ /embl\|(\S+)/)
    {
	$emb = $1;
	@report = &Html::get_html($Html::entrez_url,$Html::entrez_type,
			    [["field",6],["mode",0],["term",$emb],["db",2],["form",4],
			     ["choseninfo","$emb [accession]~1~1~1"],
			     ["retrievestring","$emb [accession]"],
			     ["docstoretrieve",1],
			     ["submit","Retrieve 1 Document"]]);
	for ($i=0; ($i <= $#report) && ($report[$i] !~ /\"uid\"/); $i++) {}
	if (($i <= $#report) && ($report[$i] =~ /VALUE\=\"(\d+)\"/))
	{
	    $gid = "gi|$1";
	}
	else
	{
	    return "";
	}
    }
    else
    {
	return "";
    }
    return &taxonomy_of($gid,"n");
}

$tax_url  = "http://www.ncbi.nlm.nih.gov/htbin-post/Taxonomy/wgetorg";
$tax_type = "POST";

sub taxonomy_of_org {
    my($org) = @_;
    my(@report,$i,@tax,$tax,@link);

    defined($org) || confess "undefined org";
    defined($WIT2_Config::organisms) || wit_die "bad config";

    if (($org =~ /^[A-Z]{2}$/) && (-s "$WIT2_Config::organisms/$org/TAXONOMY"))
    {
	@report = `cat $WIT2_Config::organisms/$org/TAXONOMY`;
	chop $report[0];
	return $report[0];
    }

    @report = &Html::get_html($tax_url,$tax_type,
			[["name",$org],["submit","Start Search"],["srchmode",1]]);
#   print STDERR "rep: @report";
    for ($i=0; ($i <$#report) && ($report[$i] !~ /abbreviated/); $i++) {}
    if ($i < $#report)
    {
	@tax = ();
	for ($i++; ($report[$i] !~ /^<\/DL>/); $i++)
	{
	    if ($report[$i] =~ /\">(\S+)\<\/A\>/)
	    {
		push(@tax,$1);
	    }
	}
	$tax = join("; ",@tax);
	return "$tax; $org";
    }
    else
    {
	@link = grep(/wgetorg.*\>$org\</,@report);
#	print STDERR "link: @link\n";
	if ((@link > 0) && ($link[0] =~ /wgetorg(\?[^\"]+)/))
	{
            @report = split(/\n/,&LWP::Simple::get(($tax_url . $1)));
#            print STDERR "rep: @report";
            for ($i=0; ($i <$#report) && ($report[$i] !~ /abbreviated/); $i++) {}
            if ($i < $#report)
            {
	        @tax = ();
	        for ($i++; ($report[$i] !~ /^<\/DL>/); $i++)
	        {
	            if ($report[$i] =~ /\">(\S+)\<\/A\>/)
	            {
		        push(@tax,$1);
	            }
	        }
	        $tax = join("; ",@tax);
	        return "$tax; $org";
            }
            else
            {
                return "";
            }
        }
        else
        {
            return "";
        }
    }
}
	

sub taxonomy_of {
    my($id,$type) = @_;
    my($gi,$report,$line,$org,@tax,$tax);

    if ($id =~ /^R([A-Z]{2})\d{5}$/)
    {
	return &taxonomy_of_org($1);
    }

    if ($id =~ /^(sp|tr)\|/)
    {
	return &taxonomy_of_org(&ORF::org_ofO($id));
    }

    if ($id =~ /^gi\|(\d+)/i)
    {
	$gi = $1;
	$report = &gi_entries($id);

	while (($line = shift @$report) && ($line !~ /^\s*ORGANISM/)) {}
	if ($line =~ /ORGANISM.*>([^<>]+)<\/a>/)
	{
	    $org = $1;
	    @tax = ();
	    while (($line = shift @$report) && ($line =~ /^          \s*(\S.*\S)\s*$/))
	    {
		push(@tax,$1);
	    }
	    $tax = join(" ",@tax);
	    $tax =~ s/\.\s*$//;
	    $tax .= "; $org";
	    return $tax;
	}
    }
    return "";
}

sub phylogenetic_distance {
    my($org1,$org2) = @_;

    if ((! defined(%distances)) && open(TMP,"<$WIT2_Config::global/distance.matrix"))
    {
	while (defined($_ = <TMP>))
	{
	    chop;
	    ($id1,$id2,$dist) = split(/\t/,$_);
	    $distances{"$id1,$id2"} = $distances{"$id2,$id1"} = $dist;
	}
	close(TMP);
    }
    
    if (defined($_ = $distances{"$org1,$org2"}))
    {
	return $_;
    }
    else
    {
	return undef;
    }
}

1;
