package TransmembraneDB;

use DB_File;
use Fcntl;

use HashDB;
use ORF;
use LWP;

use Boulder::Stream;
use Stone;

require Exporter;
@ISA = (Exporter);
@EXPORT = qw(
	     open_transmembrane_db
	     close_transmembrane_db
	     is_transmembraneO
	     transmembrane_regionsO
	     transmembrane_regions_boundsO
	     transmembrane_db_lengthO
	     transmembrane_db_hydrophobicity_averageO
	     transmembrane_region_list
	     is_in_transmembrane_dbO
	     calculate_transmembraneO
	     parse_transmembraneHTML
	     update_transmembrane_db
	     transmembrane_db_print
	     );

$ua=new LWP::UserAgent;
$ua->agent("AgentName/0.1 " . $ua->agent);

sub open_transmembrane_db {
    return &opendbH("transmembrane");
}

sub close_transmembrane_db {
    my ($ref)=@_;
    &closedbH($ref);
}

sub is_transmembraneO {
    my ($ref,$orf)=@_;
    $category=&accessdbH($ref,"${orf}__category");
    if ($category eq "MEMBRANE PROTEIN") {
	return 1;
    }
    else {
	return 0;
    }
}

sub transmembrane_regionsO {
    my ($ref,$orf,$type)=@_;
    $type="all" unless $type; # other types are PRIMARY and SECONDARY
    if ($type eq "all") {
	return &accessdbH($ref,"${orf}__regions");
    }
    elsif ($type eq "PRIMARY") {
	return &accessdbH($ref,"${orf}__primary_regions");
    }
    elsif ($type eq "SECONDARY") {
	return &accessdbH($ref,"${orf}__secondary_regions");
    }
    return 0;
}

sub transmembrane_regions_boundsO {
    my ($ref,$orf,$type)=@_;
    my @r=();
    $type="all" unless $type; # other types are PRIMARY and SECONDARY
    $regions=&accessdbH($ref,"${orf}__regions");
    for $i (0..($regions-1)) {
	$region_string=&accessdbH($ref,"${orf}__region_$i");
	($rtype,$n,$c)=$region_string=~/(.*)_(.*)_(.*)/;
	if ($type eq "all" || $type eq $rtype) {
	    push @r,[$n,$c];
	}
    }
    return @r;
}

sub transmembrane_db_lengthO {
    my ($ref,$orf)=@_;
    return &accessdbH($ref,"${orf}__length");
}

sub transmembrane_db_hydrophobicity_averageO {
    my ($ref,$orf)=@_;
    return &accessdbH($ref,"${orf}__hydrophobicity_average");
}

sub transmembrane_region_list {
    return 0;
}

sub is_in_transmembrane_dbO {
    my ($ref,$orf)=@_;
    if (&accessdbH($ref,"${orf}__category")) {
	return 1;
    }
    else {
	return 0;
    }
}

sub calculate_transmembraneO {
    my ($rid,$query_seq)=@_;
    $req = new HTTP::Request GET => "http://www.tuat.ac.jp/cgi/~mitaku/adv_sosui.script";
    $req->content_type('application/x-www-form-urlencoded');
    $req->content("comm=$rid\&query_seq=$$query_seq\&submit=Exec");
    # Pass request to the user agent and get a response back
    $res = $ua->request($req);
    # Check the outcome of the response
    if ($res->is_success) {
	return $res->content;
    }
}

sub parse_transmembraneHTML {
    my ($rid,$html)=@_;
    $sosui=$html;
    $stone=new Stone;
    $stone->insert(rid=>$rid);
    if ($sosui=~/is too short/) {
	$category="TOO SHORT";
	$stone->insert(category=>$category);
	$stream->write_record($stone);
	next;
    }
    $sosui=~/Total length \:\s*(\d+)/;
    $length=$1;
    $stone->insert("length" => $length);
    $sosui=~/Average of hydrophobicity \: ([\-|\d|\.]*)/;
    $hydrophobicity_average=$1;
    $stone->insert("hydrophobicity_average" => $hydrophobicity_average);
    $sosui=~/sequence is of a ([A-Z][A-Z|\s]*[A-Z])/;
    $category=$1;
    $stone->insert("category" => $category);
#    print "Category: $category\n";
#    print "$category\n";

    if ($category eq "MEMBRANE PROTEIN") {
	@values=$sosui=~/\<td align\=center\>(.*?)\<\/td\>/g;
	splice(@values,0,6);
	while (@values) {
	    ($number,$n_terminal,$region,$c_terminal,$type,$length)=splice(@values,0,6);
#	    print "region: $n_terminal, $c_terminal, $type\n";
	    $stone->insert(region=>[$type,$n_terminal,$c_terminal]);
	}
    }
#    print $sosui;
#    $stream->write_record($stone);
    return $stone;
}

sub update_transmembrane_db {
    my ($ref,$rid,$stone)=@_;
    &replacedbH($ref,"${rid}__category",$stone->get("category"));
    &replacedbH($ref,"${rid}__hydrophobicity_average",$stone->get(hydrophobicity_average));
    &replacedbH($ref,"${rid}__length",$stone->get(length));
    if ($stone->get("category") eq "MEMBRANE PROTEIN") {
	@regions=$stone->get("region");
	&replacedbH($ref,"${rid}__regions",(@regions+0)/3);
	$primary=0;
	$secondary=0;
	$count=0;
	while (@regions) {
	    ($rtype,$n,$c)=splice(@regions,0,3);
	    $primary++ if $rtype eq "PRIMARY";
	    $secondary++ if $rtype eq "SECONDARY";
	    &replacedbH($ref,"${rid}__region_${count}","${rtype}_${n}_${c}");
            $count++;
	}
	&replacedbH($ref,"${rid}__primary_regions",$primary);
	&replacedbH($ref,"${rid}__secondary_regions",$secondary);
    }
}

sub transmembrane_db_print {
    my ($ref,$orf)=@_;
    $exists=&is_in_transmembrane_dbO($ref,$orf);
    $is=&is_transmembraneO($ref,$orf);
    $all=&transmembrane_regionsO($ref,$orf);
    $primary=&transmembrane_regionsO($ref,$orf,"PRIMARY");
    $secondary=&transmembrane_regionsO($ref,$orf,"SECONDARY");
    @bounds=&transmembrane_regions_boundsO($ref,$orf);
    @primary_bounds=&transmembrane_regions_boundsO($ref,$orf,"PRIMARY");
    @secondary_bounds=&transmembrane_regions_boundsO($ref,$orf,"SECONDARY");
    $bounds=join (", ", (map {"(" . join(",",@{$_}) . ")"} &transmembrane_regions_boundsO($ref,$orf,"all")));
    $primary_bounds=join (", ", (map {"(" . join(",",@{$_}) . ")"} &transmembrane_regions_boundsO($ref,$orf,"PRIMARY")));
    $secondary_bounds=join (", ", (map {"(" . join(",",@{$_}) . ")"} &transmembrane_regions_boundsO($ref,$orf,"SECONDARY")));


    $a_string=<<END;
ORF:       $orf
EXISTS:    $exists
CATEGORY:  $category
IS:        $is
ALL:       $all
ALL_BOUNDS: $bounds
PRIMARY:   $primary
PRIMARY_BOUNDS: $primary_bounds
SECONDARY: $secondary
SECONDARY_BOUNDS: $secondary_bounds

END

    return $a_string;
}

1;
