package Util;		# -*- perl -*-

use Carp;
use set_utilities;
use HashDB;
use Messages;

require Exporter;
@ISA = (Exporter);
@EXPORT = qw(
	     add_to_function_hierarchy
	     attr
	     between
	     bring_wit_down
	     bring_wit_up
	     class_of_cluster
	     clean_function
	     code_ext_name
	     codon
	     codons
	     commify
	     common_name
	     connectable
	     default_orgset
	     display_id_and_seq
	     display_id_and_seq_struct
	     display_seq
	     display_seq_struct
	     domain_of_org
	     fast_read
	     genetic_code
	     genome_bioname
	     genome_of_id
	     genomes
	     genomes_with_bionames
	     genomes_with_bionames_and_stats
	     higher_level_functions
	     id_of_obj
	     join_locii
	     meet_locii
	     left_boundary
	     lines_in_file
	     loc_obj
	     locus_for_set
	     master_user
	     max
	     min
	     next_group
	     ok_directory
	     one_to_many
	     overlaps
	     prot_mol_weight
	     read_fasta
	     reduce_id
	     reduce_prot_id
	     rev_comp
	     right_boundary
	     set_timestamp
	     since_timestamp
	     standardize_bioname
	     start
	     stop 
	     sort_by_loc
	     sort_by_midpoint
	     sort_locii
	     split_on_bars
	     strand
	     to_id
	     toM
	     translate
	     type_of_obj
	     wit_is_down
	     write_file
);


sub add_to_function_hierarchy {
    my($function,$ref4,$ref5) = @_;
    my($higher_level,$higher,$current,$i,@existing);

#    defined($ref4) || confess "bad ref4 parameter";
    $higher_level = &higher_level_functions($function);
    foreach $higher (@$higher_level)
    {
#	defined($ref4) || confess "lost ref4 parameter";
	if ($current = &HashDB::accessdbH($ref4,$higher))
	{
	    @existing = split(/\t/,$current);
	    for ($i=0; ($i <= $#existing) && ($existing[$i] ne $function); $i++) {}
	    if ($i > $#existing)
	    {
		$current .= "\t$function";
		&HashDB::insertdbH($ref4,$higher,$current);
	    }
	}
	else
	{
	    &HashDB::insertdbH($ref4,$higher,$function);
	}

	if ($current = &HashDB::accessdbH($ref5,$function))
	{
	    @existing = split(/\t/,$current);
	    for ($i=0; ($i <= $#existing) && ($existing[$i] ne $higher); $i++) {}
	    if ($i > $#existing)
	    {
		$current .= "\t$higher";
		&HashDB::insertdbH($ref5,$function,$current);
	    }
	}
	else
	{
	    &HashDB::insertdbH($ref5,$function,$higher);
	}
    }
}



sub attr {
    my($obj,$field) = @_;
    
    if (${$obj} =~ /\b$field\t([^\n]+)/)
    {
        return $1;
    }
    else
    {
        return "";
    }
}



sub between {
    my($x,$y,$z) = @_;

    if ($x < $z)
    {
	return (($x <= $y) && ($y <= $z));
    }
    else
    {
	return (($x >= $y) && ($y >= $z));
    }
}



sub bring_wit_down {
    my ($msg) = @_;

    open(TMP,">$WIT2_Config::base/down");
    if ($msg) { print TMP $msg; }
    close(TMP);
}

sub bring_wit_up {
    unlink("$WIT2_Config::base/down");
}

sub wit_is_down {
    my ($msg);
    
    $msg = "";
    if (-e "$WIT2_Config::base/down")
    {
	if (-s "$WIT2_Config::base/down")
	{
	    $msg = `cat $WIT2_Config::base/down`;
	}
	else
	{
	    $msg = qq(
<H1>WIT is Down</H1>
We are sorry to inconvenience you, but WIT is currently down for some reason.
<BR>
We will try to get it back up as soon as possible.
		      );
	}
    }
    
    return \$msg;
}



sub class_of_cluster {
    my($rids) = @_;
    my(%classes,$rid,$x);
    
    if(! defined(%domain)) { &load_domains; }

    foreach $rid (@$rids)
    {
	$x = substr($rid,1,2);
	$classes{$domain{$x}} = 1;
#	print STDERR "class_of_cluster for $rid: $x -> $domain{$x}\n";
    }
    return join( "", sort(keys(%classes)) );
}



sub clean_function {
    my($func) = @_;

#    $func =~ s/ PRECURSOR//g;
#    $func =~ s/ FRAGMENT//g;
    $func =~ s/^\s+//;
    $func =~ s/\s+$//;
    return $func;
}


sub commify {

    use strict;

#  IN: an integer, or integer string
# OUT: an integer string with commas inserted

    my $text = reverse $_[0];

    $text =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
    
    return scalar reverse $text;
}


sub common_name {
    my($rid) = @_;
    my($genome,$x);

    if ($rid =~ /^R(..)/)
    {
	$genome = $1;
	&load_orf_tbl($genome);

	if ($x = $orf_tbl{$rid})
	{
	    if ($x =~ /^\S+\s+\S+\t(\S+)/)
	    {
		return $1;
	    }
	}
	else
	{
	    return "";
	}
    }
    else
    {
	return "";
    }
}



sub connectable {
    my($funcL,$org,$user) = @_;
    my($func,@pws);
    
    foreach $func (@$funcL)
    {
	$pws = &FR::function_to_pwsR($func,$org,$user);
	if (@$pws > 0)
	{
	    return 1;
	}
    }
    return 0;
}



sub display_id_and_seq {
    my( $id, $seq, $fh ) = @_;
    
    if (! defined($fh) )  { $fh = \*STDOUT; }
    
    print $fh ">$id\n";
    &display_seq($seq, $fh);
}

sub display_seq {
    my ( $seq, $fh ) = @_;
    my ( $i, $n, $ln );
    
    if (! defined($fh) )  { $fh = \*STDOUT; }

    $n = length($$seq);
#   confess "zero-length sequence ???" if ( (! defined($n)) || ($n == 0) );
    for ($i=0; ($i < $n); $i += 60)
    {
	if (($i + 60) <= $n)
	{
	    $ln = substr($$seq,$i,60);
	}
	else
	{
	    $ln = substr($$seq,$i,($n-$i));
	}
	print $fh "$ln\n";
    }
}


sub display_id_and_seq_struct {
    my($id,$seq,@structures) = @_;

    print ">$id\n";
    &display_seq_struct($seq,@structures);
}



sub display_seq_struct {
    my($seq,@structures) = @_;

    my $t=0; my $a="";

    my $line_length=60;
    @tags=map {$t++; $a=$_; unless ($t % 2) {$a="</FONT>";} else {$a="<FONT COLOR=red>"} $a;} @structures;
    my($i,$n,$ln);

    my %struct_hash;
    foreach $structure (@structures) {
	push @{$struct_hash{$line_length*int($structure/$line_length)}},
	       ($structure%$line_length);
	push @{$tag_hash{$line_length*int($structure/$line_length)}},
	       (shift @tags);
    }
    print "<BR>\n";
    $n = length($$seq);
    for ($i=0; ($i < $n); $i += $line_length)
    {
	if (($i + $line_length) <= $n)
	{
	    $ln = substr($$seq,$i,$line_length);
	}
	else
	{
	    $ln = substr($$seq,$i,($n-$i));
	}
	foreach $sh (reverse @{$struct_hash{$i}}) {
	    substr($ln,$sh,0)=pop (@{$tag_hash{$i}});
	}
	print "$ln\n";
    }

}



sub domain_of_org {
    my ( $orgID ) = @_;
    my ( @tmp, $short, $name, $domain );

    if (! defined(%domain)) { &load_domains; }
    
    if (defined($domain = $domain{$orgID}) && $domain)
    {
	return $domain;
    }
    else
    {
	warn "bad genome: $orgID";
	return "";
    }
}

sub load_domains
{
    my ( @tmp, $entry, $short, $name, $domain );
    
    @tmp    = `cat $WIT2_Config::global/genomes`;
    ( @tmp > 0 ) || wit_die "could not read $WIT2_Config::global/genomes";
    
    chomp( @tmp );
    foreach $entry ( @tmp )
    {
	($short, $name, $domain ) = split(/\t/,$entry);
	$domain{$short} = $domain;
#	print STDERR "load_domains: $short -> $domain\n";
    }
    return 1;
}



sub fast_read {
    my($file) = @_;
    my($input,$status,$lines);
    
    $lines = [];
    if (open(TMP,"<$file"))
    {
	$input  = "";
	$status = sysread(TMP,$input,10000000);
	close(TMP);
	
	if (defined($status) && $status)
	{
	    chomp $input;
	    @$lines = split(/\n/,$input);
	}
	else
	{
	    $lines = "";
	}
    }
    
    return $lines;
}



sub genetic_code {
    my($type) = @_;

    if ($type eq "standard")
    {
	$stop{"TGA"}  = 1;
    }
    $stop{"TAA"}  = 1;
    $stop{"TAG"}  = 1;

    $start{"ATG"} = 1;
    $start{"GTG"} = 1;
    $start{"TTG"} = 1;

    $codon{"AAA"} = "K";
    $codon{"AAC"} = "N";
    $codon{"AAG"} = "K";
    $codon{"AAT"} = "N";
    $codon{"ACA"} = "T";
    $codon{"ACC"} = "T";
    $codon{"ACG"} = "T";
    $codon{"ACT"} = "T";
    $codon{"AGA"} = "R";
    $codon{"AGC"} = "S";
    $codon{"AGG"} = "R";
    $codon{"AGT"} = "S";
    $codon{"ATA"} = "I";
    $codon{"ATC"} = "I";
    $codon{"ATG"} = "M";
    $codon{"ATT"} = "I";
    $codon{"CAA"} = "Q";
    $codon{"CAC"} = "H";
    $codon{"CAG"} = "Q";
    $codon{"CAT"} = "H";
    $codon{"CCA"} = "P";
    $codon{"CCC"} = "P";
    $codon{"CCG"} = "P";
    $codon{"CCT"} = "P";
    $codon{"CGA"} = "R";
    $codon{"CGC"} = "R";
    $codon{"CGG"} = "R";
    $codon{"CGT"} = "R";
    $codon{"CTA"} = "L";
    $codon{"CTC"} = "L";
    $codon{"CTG"} = "L";
    $codon{"CTT"} = "L";
    $codon{"GAA"} = "E";
    $codon{"GAC"} = "D";
    $codon{"GAG"} = "E";
    $codon{"GAT"} = "D";
    $codon{"GCA"} = "A";
    $codon{"GCC"} = "A";
    $codon{"GCG"} = "A";
    $codon{"GCT"} = "A";
    $codon{"GGA"} = "G";
    $codon{"GGC"} = "G";
    $codon{"GGG"} = "G";
    $codon{"GGT"} = "G";
    $codon{"GTA"} = "V";
    $codon{"GTC"} = "V";
    $codon{"GTG"} = "V";
    $codon{"GTT"} = "V";
    $codon{"TAA"} = "*";
    $codon{"TAC"} = "Y";
    $codon{"TAG"} = "*";
    $codon{"TAT"} = "Y";
    $codon{"TCA"} = "S";
    $codon{"TCC"} = "S";
    $codon{"TCG"} = "S";
    $codon{"TCT"} = "S";
    if ($type eq "standard")
    {
	$codon{"TGA"} = "*";
    }
    else
    {
	$codon{"TGA"} = "W";
    }
    $codon{"TGC"} = "C";
    $codon{"TGG"} = "W";
    $codon{"TGT"} = "C";
    $codon{"TTA"} = "L";
    $codon{"TTC"} = "F";
    $codon{"TTG"} = "L";
    $codon{"TTT"} = "F";
    
    
    $code_ext_name{"A"} = "Ala\tAlanine";
    $code_ext_name{"C"} = "Cys\tCysteine";
    $code_ext_name{"D"} = "Asp\tAspartic Acid";
    $code_ext_name{"E"} = "Glu\tGlutamic Acid";
    $code_ext_name{"F"} = "Phe\tPhenylalanine";
    $code_ext_name{"G"} = "Gly\tGlycine";
    $code_ext_name{"H"} = "His\tHistidine";
    $code_ext_name{"I"} = "Ile\tIsoleucine";
    $code_ext_name{"K"} = "Lys\tLysine";
    $code_ext_name{"L"} = "Leu\tLeucine";
    $code_ext_name{"M"} = "Met\tMethionine";
    $code_ext_name{"N"} = "Asn\tAsparagine";
    $code_ext_name{"P"} = "Pro\tProline";
    $code_ext_name{"Q"} = "Gln\tGlutamine";
    $code_ext_name{"R"} = "Arg\tArginine";
    $code_ext_name{"S"} = "Ser\tSerine";
    $code_ext_name{"T"} = "Thr\tThreonine";
    $code_ext_name{"V"} = "Val\tValine";		     
    $code_ext_name{"W"} = "Trp\tTryptophan";
    $code_ext_name{"Y"} = "Tyr\tTyrosine";
    
    $codons = 
       [
	  ["*","TAA","TAG","TGA"],
          ["A","GCA","GCC","GCG","GCT"],
          ["C","TGC","TGT"],
          ["D","GAC","GAT"],
          ["E","GAA","GAG"],
          ["F","TTC","TTT"],
          ["G","GGA","GGC","GGG","GGT"],
          ["H","CAC","CAT"],
          ["I","ATA","ATC","ATT"],
          ["K","AAA","AAG"],
          ["L","CTA","CTC","CTG","CTT","TTA","TTG"],
          ["M","ATG"],
          ["N","AAC","AAT"],
          ["P","CCA","CCC","CCG","CCT"],
          ["Q","CAA","CAG"],
          ["R","AGA","AGG","CGA","CGC","CGG","CGT"],
          ["S","AGC","AGT","TCA","TCC","TCG","TCT"],
          ["T","ACA","ACC","ACG","ACT"],
          ["V","GTA","GTC","GTG","GTT"],
          ["W","TGG"],
          ["Y","TAC","TAT"]
       ];
    
    return;
}



sub genome_bioname {
    my ( $orgID ) = @_;
    my ( @tmp, $short, $name );

    @tmp    = `cat $WIT2_Config::global/genomes`;
    @tmp    =  grep( /^$orgID/, @tmp);
    if (@tmp == 1)
    {
	chomp( $tmp[0] );
	($short,$name) = split(/\t/,$tmp[0]);
	return $name;
    }
    else
    {
#	warn "bad genome: $orgID";
	return "";
    }
}

sub standardize_bioname
{
    my ( $org ) = @_;
    my ( $tmp );
    
    $tmp =  $org;
    $org =~ s/^[\"]?([^\"]+)/$1/;
    if ( $tmp =~ m/^(\S)(\S+)/ )
    {
	$x = $1;
	$y = $2;    
	
	if ( $tmp =~ m/^\S+\s+([^\.\s]+)/ )  { $z = $1; } else { $z = "sp."; }
	
	$x =~ tr/a-z/A-Z/;
	$y =~ tr/A-Z/a-z/;
	$z =~ tr/A-Z/a-z/;
	$z =~ s/sp$/sp./;
	
	$tmp = "$x$y $z";
    }
    else
    {
	warn "\nUtil::standardize_bioname --- invalid org $org";
	$tmp = "Unparsable sp.";
    }
    
    return $tmp;
}



sub genome_of_id {
    my($id) = @_;

    if ($id =~ /^R([A-Z]+)\d{5}$/)      { return $1; }
    if ($id =~ /^([A-Z]+)RNA\d{3}$/)    { return $1; }
    if ($id =~ /^([A-Z]+)INTR\d{3}$/)   { return $1; }
    if ($id =~ /^([A-Z]+)INTE\d{3}$/)   { return $1; }
    return "";
}



sub genomes {
    my( @out );
    my(@out1,$x);

    @out = `cut -f1 $WIT2_Config::global/genomes`;
    chomp @out;

    @out1 = ();
    foreach $x (@out)
    {
	if (-d "$WIT2_Config::organisms/$x")
	{
	    push(@out1,$x);
	}
    }
    return @out1;
}

sub genomes_with_bionames {
    my($org) = @_;
    my(@genomes,@tmp,$x,$short,$real);

    @genomes = ();
    @tmp    = `cat $WIT2_Config::global/genomes`;
    chomp @tmp;
    foreach $x (@tmp)
    {
	($short,$real) = split(/\t/,$x);
	if (-d "$WIT2_Config::organisms/$short")
	{
	    push(@genomes,[$short,$real]);
	}
    }
#   @genomes = sort by_arg2 @genomes;
    return \@genomes;
}

sub genomes_with_bionames_and_stats {
    my($org) = @_;
    my(@genomes,@tmp,$x,$short,$real,$domain,$size,$orfs);

    @genomes = ();
    @tmp    = `cat $WIT2_Config::global/genomes`;
    chomp @tmp;

    foreach $x (@tmp)
    {
	($short,$real,$domain,$size,$orfs) = split(/\t/,$x);

	if (-d "$WIT2_Config::organisms/$short")
	{
	    push(@genomes,[$short,$real,$domain,$size,$orfs]);
	}
    }
#   @genomes = sort by_arg2 @genomes;
    return \@genomes;
}

sub by_arg2 {
    return ($a->[1] cmp $b->[1]);
}
	


sub higher_level_functions {
    my($function) = @_;
    my($higher_level) = [];

    if (($function !~ /^\d+\.\d+\.\d+\.\d+$/) && ($function =~ /\d+\.\d+\.\d+\.\d+/))
    {
	while ($function =~ s/(\d+\.\d+\.\d+\.\d+)//)
	{
	    push(@$higher_level,$1);
	}
    }
    elsif ($function =~ /^(.*), \S+ CHAIN$/i)
    {
	push(@$higher_level,$1);
    }
    return $higher_level;
}



sub join_locii {
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Returns the ``join'' (directed union) of two locii overlapping 
#   on the same strand of the same contig, and the nil-string otherwise...
#-----------------------------------------------------------------------
    my ( $locus1, $locus2 ) = @_;
    my ( $contig1, $beg1, $end1 );
    my ( $contig2, $beg2, $end2 );
    my ( $locus );
    
    $locus1 =~ m/^(\S+)_(\d+)_(\d+)$/;
    ( $contig1, $beg1, $end1 ) = ( $1, $2, $3 );
    
    $locus2 =~ m/^(\S+)_(\d+)_(\d+)$/;
    ( $contig2, $beg2, $end2 ) = ( $1, $2, $3 );
    
    $locus = "";
    if ( ($contig1 eq $contig2) && (($beg1 <=> $end1) == ($beg2 <=> $end2)) )
    {
	if ( &overlaps( $beg1,$end1, $beg2,$end2 ) ) 
	{ 
	    if ( $beg1 < $end1 ) 
	    {
		$beg1 = &min( $beg1, $beg2 );
		$end1 = &max( $end1, $end2 );
	    } else {
		$beg1 = &max( $beg1, $beg2 );
		$end1 = &min( $end1, $end2 );
	    }
	    $locus = "$contig1\_$beg1\_$end1";
	}
    }
    
#   print STDERR "join of $locus1 and $locus2 is: $locus.\n";
    return $locus;
}

sub meet_locii {
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...Returns the ``meet'' (directed intersection) of two locii overlapping
#   on the same strand of the same contig, and the nil-string otherwise...
#-----------------------------------------------------------------------
    my ( $locus1, $locus2 ) = @_;
    my ( $contig1, $beg1, $end1 );
    my ( $contig2, $beg2, $end2 );
    my ( $locus );
    
    $locus1 =~ m/^(\S+)_(\d+)_(\d+)$/;
    ( $contig1, $beg1, $end1 ) = ( $1, $2, $3 );
    
    $locus2 =~ m/^(\S+)_(\d+)_(\d+)$/;
    ( $contig2, $beg2, $end2 ) = ( $1, $2, $3 );
    
    $locus = "";
    if ( ($contig1 eq $contig2) && (($beg1 <=> $end1) == ($beg2 <=> $end2)) )
    {
	if ( &overlaps( $beg1,$end1, $beg2,$end2 ) ) 
	{ 
	    if ( $beg1 < $end1 ) 
	    {
		$beg1 = &max( $beg1, $beg2 );
		$end1 = &min( $end1, $end2 );
	    } else {
		$beg1 = &min( $beg1, $beg2 );
		$end1 = &max( $end1, $end2 );
	    }
	    $locus = "$contig1\_$beg1\_$end1";
	}
    }
    
#   print STDERR "meet of $locus1 and $locus2 is: $locus.\n";
    return $locus;
}



sub id_of_obj {
    my($obj) = @_;

    return &attr($obj,"id");
}

sub type_of_obj {
    my($obj) = @_;

    return &attr($obj,"type");
}



sub left_boundary {
    my($obj) = @_;
    my(@locs,$loc,$b,$e);

    $loc = &attr($obj,"location");
    if ($loc)
    {
	@locs = split(/\s*,\s*/,$loc);
	if ((@locs == 1) && ($locs[0] =~ /_(\d+)_(\d+)/))
	{

	    $b = $1; $e = $2;
	    if (($b <= $e) && (($e - $b) < 15000)) { return $b; }
	    if (($e < $b) && (($b - $e) < 15000)) { return $e; }
	}
    }
    return "";
}

sub right_boundary {
    my($obj) = @_;
    my(@locs,$loc,$b,$e);

    $loc = &attr($obj,"location");
    if ($loc)
    {
	@locs = split(/\s*,\s*/,$loc);
	if ((@locs == 1) && ($locs[0] =~ /_(\d+)_(\d+)/))
	{

	    $b = $1; $e = $2;

	    if (($b <= $e) && (($e - $b) < 15000)) { return $e; }
	    if (($e < $b) && (($b - $e) < 15000)) { return $b; }
	}
    }
    return "";
}



sub lines_in_file {
    my($file) = @_;

    my(@tmp) = `wc \"$file\"`;
    $tmp[0] =~ /^\s*(\d+)/;
    return $1;
}



sub loc_obj {
    my($obj) = @_;
    my($loc) = &attr($obj,"location");
    my($contig,$beg,$end);

    if ($loc && ($loc =~ /^(\S+)_(\d+)_(\d+)/))
    {
	$contig = $1; $beg = $2; $end = $3;
	return (&attr($obj,"genome"),$contig,$beg,$end);
    }
    else
    {
	return "";
    }
}



sub locus_for_set {
    my ( @set_of_ids ) = @_;
    my ( $genome, $genome1, $contig1, $id, $loc );
    my ( $first, $last, $min, $max, $orientation );
    my ( $beg, $end, $contig );
    
    $genome = "";
    $contig = "";
    $min = $first = 100000000;
    $max = $last  = 0;
    $orientation = 0;
    foreach $id ( @set_of_ids )
    {
	($contig1, $beg, $end) = &Feature::get_parsed_locationF( $id );
	if ( $beg < $end ) { $orientation++; } else { $orientation--; }

	$genome1 = &Feature::genome_ofF($id);
	if (! $genome)  { $genome = $genome1; }
	if (! $contig)  { $contig = $contig1; }
	
#	print STDERR "Util::locus_for_set --- feature = $feature, genome = $genome1, loc = $loc, contig = $contig1, beg = $beg, end = $end\n";
	
	if ($genome ne $genome1)  { 
	    $bad = 1;  # print STDERR "Util::locus_for_set --- for @set_of_ids, mismatched genomes $genome, $genome1\n"; 
	}
	if ($contig ne $contig1)  { 
	    $bad = 1;  # print STDERR "Util::locus_for_set --- for @set_of_ids, mismatched contigs $contig, $contig1\n"; 
	}
	
	last if $bad;
	
	$min    = &min($beg,$end);
	$max    = &max($beg,$end);
	$first  = &min($min,$first);
	$last   = &max($max,$last);
    }
    
    if ( $orientation < 0 ) { ($first, $last) = ($last, $first); }
    
    if ((! $bad) && $contig && $first && $last)
    { 
	return "$contig\_$first\_$last";
    } else {
	return "";
    }
}



sub master_user {
    my($org,$user) = @_;
    my(@masters,@hits);

    return ( ($user eq "master")
	     || 
	     ((-s "$WIT2_Config::organisms/$org/master.users")
	      && (@masters = `cat $WIT2_Config::organisms/$org/master.users`)
	      && (@hits = grep(/^$user$/,@masters))
	      && (@hits >= 1)
	     )
	   );
}



sub max {
    my($x,$y) = @_;

    if ($x >= $y) { return $x; } else { return $y; }
}

sub min {
    my($x,$y) = @_;

    if ($x <= $y) { return $x; } else { return $y; }
}

sub next_group {
    my($obj,$field) = @_;
    my($group,$ans,$i);

    if (($#{$obj} >= 0) && ($obj->[0] =~ /\b$field\t([^\n]+)/))
    {
	$group = $1;
	$ans   = [$obj->[0]];
	for ($i=1; ($i <= $#{$obj}) && ($obj->[$i] =~ /\b$field\t([^\n]+)/) && ($1 eq $group); $i++)
	{
	    push(@$ans,$obj->[$i]);
	}
	splice(@$obj,0,($#{$ans} + 1));
	return $ans;
    }
    else
    {
	return "";
    }
}




sub ok_directory {
    my($dir) = @_;

    if (! -d $dir)
    {
	mkdir($dir,0777);
	system("chmod 777 $dir");
    }
}



sub one_to_many {
    my($array, $key, $data) = @_;
    my($x,$i);

    if (! ($x = $array->{$key}))
    {
        $x = ($array->{$key} = []);
    }

    if ("ARRAY" eq ref($data))
    {
	for ($i = 0; (($i <= $#{$x}) && ($x->[$i]->[0] lt $data->[0])); $i++) { }
	
	if (($i > $#{$x}) || ($x->[$i]->[0] gt $data->[0]))
	{
	    splice (@$x, $i, 0, $data);
	}
	
    }
    else
    {
	for ($i = 0; (($i <= $#{$x}) && ($x->[$i] lt $data)); $i++) { }
	
	if (($i > $#{$x}) || ($x->[$i] gt $data))
	{
	    splice (@$x, $i, 0, $data);
	}
    }
}



sub overlaps {
    my($b1,$e1,$b2,$e2) = @_;

    return (&between($b1,$b2,$e1) || &between($b2,$b1,$e2));
}


sub prot_mol_weight {

#  IN: Protein sequence given as a string of single characters
# OUT: Molecular weight in daltons
 
    use strict;

    my $prot_seq = shift @_;

    # Taken from Schultz & Schirmer, ninth print, page 2.

    my %weights = (
		   "A" => 71,
		   "E" => 128,
		   "Q" => 128,
		   "D" => 114,
		   "N" => 114,
		   "L" => 113,
		   "G" => 57,
		   "K" => 129,
		   "S" => 87,
		   "V" => 99,
		   "R" => 157,
		   "T" => 101,
		   "P" => 97,
		   "I" => 113,
		   "M" => 131,
		   "F" => 147,
		   "Y" => 163,
		   "C" => 103,
		   "W" => 186,
		   "H" => 137
		  );    
    my $aa;

    foreach $aa ( keys %weights )
    {
	$weights{ lc $aa } = $weights{$aa};
    }

    my $weight = 0;

    foreach $aa ( split "", $prot_seq )
    {
	$weight += $weights{ $aa };
    }

    return $weight;
}


sub read_fasta {
    my ( $file ) = @_;
    my ( $input, $entry, @entry, $header, $id, $seq, $rest );
    my ( $fasta_parsed,$status );
    
    $fasta_parsed = [];
    if (open(TMP,"<$file"))
    {
	$input = "";
	$status = sysread(TMP,$input,10000000);
	close(TMP);
	
	if (defined($status) && $status)
	{
	    chomp($input);
	    if ($input !~ m/^>(\S+)/) 
	    {
		print STDERR "!!! WARNING: $file is not in FASTA format !!!\n"; 
		return "";
	    }
	    
	    $input = substr( $input, 1 );
	    @$fasta_parsed = split( /\n>/, $input );
	    foreach $entry ( @$fasta_parsed )
	    {
		@entry  = split(/\n/, $entry);
		$header = shift @entry;
		$header =~ m/^(\S+)\s*(.*)$/;
		$id     = $1;
		$rest   = $2;
		$seq    = join("", @entry );
		$entry  = [ $id, $seq, $rest ];
	    }
	}
	else
	{
	    print STDERR "Util::fasta_read --- sysread failed on $file";
	    if (defined($status)) { print STDERR ", $status bytes read"; }
	    print STDERR "\n";
	    $fasta_parsed = "";
	}
    }
    else
    {
	$fasta_parsed = "";
    }
    
    return $fasta_parsed;
}

sub reduce_id 
{
    my($id) = @_;
    my($x);
    
    return $id if ($id =~ m/^R[A-Z]{2}\d{5}$/);
    
    if    ( $id =~ m/(sp\|[OPQ]\d[0-9A-Z]{3}\d)/i )  { $id = $1; }
    elsif ( $id =~ m/(tr\|[OPQ]\d[0-9A-Z]{3}\d)/i )  { $id = $1; }
    elsif ( $id =~ m/(tn\|[^\|]+)/i               )  { $id = $1; }
    elsif ( $id =~ m/(gi\|\d+)/i                  )  { $id = $1; }
    
    return $id;
}



sub reduce_prot_id 
{
    my( $id ) = @_;
    my( $x, $tmp );

    defined($id) || confess "bad call to reduce_prot_id: id is undefined";
    return $id if ($id =~ /^R[A-Z]{2}\d{5}$/);

    $tmp =  &reduce_id( $id );
    $tmp =~ s/^(sp|tr|tn|gi)\|//;

    if (! $opened_principal_synonym)
    {
	($opened_principal_synonym = &HashDB::opendbH("principal_synonym"))
	    || wit_die "could not open principal_synonym";
    }
    
    print STDERR "looking for $tmp, " if (defined($ENV{"DEBUG"}) && ($ENV{"DEBUG"} eq "VERBOSE"));
    if (defined($x = &HashDB::accessdbH($opened_principal_synonym, $tmp)))
    {
	print STDERR "found $id -> $x\n" if (defined($ENV{"DEBUG"}) && ($ENV{"DEBUG"} eq "VERBOSE"));
	$id = $x;
    }
    else
    {
	print STDERR "found nothing, returning $id\n" if (defined($ENV{"DEBUG"}) && ($ENV{"DEBUG"} eq "VERBOSE"));
    }
    
    print STDERR "$id\n" if (defined($ENV{"DEBUG"}) && ($ENV{"DEBUG"} eq "VERBOSE"));
    return $id;
}

sub rev_comp {
    my( $seqP ) = @_;
    my( $rev  );

    $rev =  reverse( $$seqP );
    $rev =~ tr/a-z/A-Z/;
    $rev =~ tr/ACGTUMRWSYKBDHV/TGCAAKYWSRMVHDB/;
    return \$rev;
}

sub set_timestamp {
    my($org,$user) = @_;

    &ok_directory("$WIT2_Config::organisms/$org/UserModels");
    open(TMP,">$WIT2_Config::organisms/$org/UserModels/$user/timestamp");
    close(TMP);
}

sub since_timestamp {
    my($org,$user) = @_;

    if ($since = -M "$WIT2_Config::organisms/$org/UserModels/$user/timestamp")
    {
	return ($since * 24 * 60);  # return minutes since stamp
    }
    return undef;
}



sub sort_by_loc {
    my($objs) = @_;
    my(@loc,$x,$key,@tmp1,@ans);
    my(@tmp) = ();

    foreach $x (@$objs)
    {
	if (@loc = &loc_obj(\$x))
	{
	    $key = [@loc];
	    push(@tmp,[$key,$x]);
	}
    }

    @tmp1 = sort by_location @tmp;
    @ans = ();
    foreach $x (@tmp1)
    {
	if ((@ans == 0) || ($ans[$#ans] ne $x->[1]))
	{
	    push(@ans,$x->[1]);
	}
    }
    return @ans;
}

sub by_location {

   if ($a->[0]->[0] ne $b->[0]->[0])
    {
	return ($a->[0]->[0] cmp $b->[0]->[0]);
    }
    elsif ($a->[0]->[1] ne $b->[0]->[1])
    {
	return ($a->[0]->[1] cmp $b->[0]->[1]);
    }
    else
    {
	return (&min($a->[0]->[2],$a->[0]->[3]) <=> &min($b->[0]->[2],$b->[0]->[3]));
    }
}



sub sort_by_midpoint {
    my ( $features ) = @_;
    my ( $feature, $beg, $end, $genome, $contig, $genome1, $contig1 );
    my ( $to_show,$ln,@loc );
    
    $bad = "";
    $genome  = ""; 
    $contig  = "";
    $to_show = [];
    foreach $feature (@$features)
    {
	$loc     = &Feature::locationF($feature);
	$genome1 = &Feature::genome_ofF($feature);
	if (($loc =~ /^([A-Za-z0-9-_|.:]+)_(\d+)_(\d+)$/) ||
	    ($loc =~ /^([A-Za-z0-9-_|.:]+)_(\d+)_\d+,.*_(\d+)$/) )
	{
	    $contig1 = $1; $beg = $2; $end = $3;
	    @loc = split(/,/,$loc);
	    $ln = 0;
	    foreach $_ (@loc)
	    {
		if ($_ =~ /^\S+_(\d+)_(\d+)$/)
		{
		    $ln += abs($1 - $2) + 1;
		}
	    }
	    if (! $genome)  { $genome = $genome1; }
	    if (! $contig)  { $contig = $contig1; }
	}
	else
	{
	    $contig1 = ""; $beg = ""; $end = ""; $genome = $genome1;
	}
#	print STDERR "Util::sort_by_midpoint --- feature = $feature, genome = $genome1, loc = $loc, "
#	                                       ,"contig = $contig1, beg = $beg, end = $end\n";
	
	if ($genome ne $genome1)  { $bad = $feature; print STDERR "Util::sort_by_midpoint --- for feature $feature, mismatched genomes $genome, $genome1 feature=$feature\n"; }
	if ($contig ne $contig1)  { $bad = $fearure; print STDERR "Util::sort_by_midpoint --- for feature $feature, mismatched contigs $contig, $contig1\n"; }
	last if $bad;
	
	push(@$to_show,[$feature,$beg,$end,$ln]);
    }
    
    if (! $bad) { 
	@$to_show = sort by_midpoint @$to_show;
    } else { 
	print STDERR "Util::sort_by_midpoint --- bad feature-set:\n";
	foreach $feature ( @$to_show )  { print STDERR join( ", ", @$feature), "\n"; }
	$to_show = [];
    }
    return $to_show;
}

sub by_midpoint {
    return ( ($a->[1] + $a->[2]) <=> ($b->[1] + $b->[2]) );
}
	    


sub sort_locii {
    my ( @locii ) = @_;
    
    return (sort by_locus_midpoint @locii);
}

sub by_locus_midpoint {
    my ( $contig1, $beg1, $end1 );
    my ( $contig2, $beg2, $end2 );
    
    if ($a =~ /^(\S+)_(\d+)_(\d+)$/)
    {
	$contig1 = $1; $beg1 = $2; $end1 = $3;
    }
    
    if ($b =~ /^(\S+)_(\d+)_(\d+)$/)
    {
	$contig2 = $1; $beg2 = $2; $end2 = $3;
    }
    
    if ( $x = ($contig1 cmp $contig2) )  { return $x; }
    return ( ($beg1 + $end1) <=> ($beg2 + $end2) );
}
	    


sub split_on_bars
{
    my ( $ids ) = @_;
    my ( @ids, $id, $i );
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ... NOTE: handling of "\cA"s in FASTA headers is still not right... :-(
#-----------------------------------------------------------------------    
    print STDERR "WARNING: whitespace or ^A in $ids --- check results\n" if ($ids =~ m/(\s|\cA)/);
    
    @ids = split( /[\|\cA]/, $ids );
    for ( $i=0; $i < @ids; $i++ )
    {
	if ( $ids[$i] =~ m/(sp|tr|tn|gi)/ )  
	{ 
	    unless (defined($ids[$i+1])) 
	    { 
		wit_die "out of data --- \$i=$i,\n\$ids=$ids,\n\@ids=", join( ", ", @ids );
	    }
	    $ids[$i] .=  '|' . splice(@ids, ($i+1), 1);
	}
	$ids[$i] =~ s/^(\S+)/$1/;
    }
    
    return @ids;
}



sub strand {
    my ( $beg, $end ) = @_;

    if ($beg <= $end) { return "+"; }
    else              { return "-"; }
}



sub translate {
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#...NOTE: Must call '&genetic_code()' first...
#-----------------------------------------------------------------------
    my( $dnaP,$seleno ) = @_;
    my( $i,$j,$ln );
    my( $x,$y );
    my( $prot );
    my( $dna  ) = $$dnaP;

    $ln = length($dna);
    $prot = "X" x ($ln/3);
    $dna =~ tr/a-z/A-Z/;

    confess "Util::translate --- &genetic_code() has not yet been called"
	if ((scalar (keys %codon)) == 0);
    for ($i=0,$j=0; ($i < ($ln-2)); $i += 3,$j++)
    {
	$x = substr($dna,$i,3);
	if ($seleno && (($x eq "TGA") || ($x eq "UGA")))
	{
	    substr($prot,$j,1) = "U";
	}
        elsif ($y = $codon{$x})
        {
	    substr($prot,$j,1) = $y;
        }
    }
    return \$prot;
}

sub toM {
    my( $dnaP ) = @_;
    my( $x ) = substr( $$dnaP, 0,3 );
    my($trans);

    $x =~ tr/a-z/A-Z/;
    if ( $start{$x} )
    {
	return "M";
    }
    elsif ($trans = $codon{$x})
    {
	return $trans;
    }
    else
    {
	return "X";
    }
}

sub write_file {
    my($file,$list) = @_;
    my($x);

    if (-e $file)
    {
	if (rename($file,"$file~"))
	{
	    if (&basic_write_file($file,$list))
	    {
		unlink("$file~");
		return 1;
	    }
	    else
	    {
		rename("$file~",$file);
		return 0;
	    }
	}
	else
	{
	    return &basic_write_file($file,$list);
	}
    }
    else
    {
	return &basic_write_file($file,$list);
    }
}

sub basic_write_file {
    my($file,$list) = @_;
    my($i,$rc);

    if (!open(TMP,">$file"))
    {
	warn "failed to write $file\n";
	return 0;
    }

    $rc = 1;
    for ($i=0; $rc && ($i <= $#{$list}); $i++)
    {
	$rc = (print TMP $list->[$i]);
	if (! $rc)
	{
	    warn "FAILED on a print to $file\n";
	}
    }
    close(TMP);
    if ($rc) { system "chmod 777 $file"; }
    return $rc;
}

sub to_id {
    my($id) = @_;
    my($org,$id1,@possible);

    if ($id =~ /^MJ\d{4}$/) { $id = "MJ\|" . $id; }   # For Carl
    if ($id =~ /^R[A-Z]{2}\d{5}$/) { return $id; }
    if (($id =~ /^([A-Z]{2})\|(\S+)$/) && (-d "$WIT2_Config::organisms/$1/Features"))
    {
	$org = $1;
	$id1 = $2;
	@possible = grep(/\b$id1\b/,`grep $id1 $WIT2_Config::organisms/$org/Features/*/tbl`);
	if ((@possible == 1) && ($possible[0] =~ /(R[A-Z]{2}\d{5})/))
	{
	    return $1;
	}
    }
    return "";
}


sub default_orgset {

    return "AA,BB,DR,CY,PG,CT,PN,BS,CA,EF,MP,MT,CJ,NG,TP,HP,RC,PA,EC,YP,HI,AG,TH,MJ,PH,SC,CE";
}

	
1;
#####  End Util.pm  ####################################################
